#include "stdafx.h"

#include <assert.h>

#define CB3DE_CPP

#include "cb3de.h"

typedef otNode *(*calcFuncPtr)(int x, int y, int z, otNode *o);

vector dirVectorLookup[6] =
{
  {0,1,0},
  {1,0,0},
  {0,-1,0},
  {-1,0,0},
  {0,0,1},
  {0,0,-1},
};

direction opposite[6] =
{
  south,west,north,east,back,front
};

// For looking up what frontDir should be, given northDir and eastDir
direction frontDirLookup[6][6] =
{
	{none,front,none,back,west,east},
	{back,none,front,none,north,south},
	{none,back,none,front,east,west},
	{front,none,back,none,south,north},
	{east,south,west,north,none,none},
	{west,north,east,south,none,none},
};

// given a 5 bit orientation value, this table stores the northDir, eastDir and frontDir
direction dirLookup[24][3] =
{
	{north,east,front}, // 0
	{north,west,back},
	{north,front,west},
	{north,back,east},
	{east,north,back}, // 4
	{east,south,front},
	{east,front,north},
	{east,back,south},
	{south,east,back}, //8
	{south,west,front},
	{south,front,east},
	{south,back,west},
	{west,north,front}, // 12
	{west,south,back},
	{west,front,south},
	{west,back,north},
	{front,north,east}, // 16
	{front,east,south},
	{front,south,west},
	{front,west,north},
	{back,north,west}, // 20
	{back,east,north},
	{back,south,east},
	{back,west,south},
};
unsigned char makeOrientation[6][6] =
{
	{0,0,0,1,2,3},
	{4,0,5,0,6,7},
	{0,8,0,9,10,11},
	{12,0,13,0,14,15},
	{16,17,18,19,0,0},
	{20,21,22,23,0,0},
};
unsigned char rotateLookup[24][6];

otNode *eg;

#define ALLOC_TABLE_SIZE 5000000
#define GC_THRESHHOLD 100000

otNode *allocTable;

otNode *freeList;
unsigned freeCount;

// First 100 elements of hash table are empty nodes
#define HASH_SIZE 4999991
#define EMPTY_LIST_SIZE 100

otNode **hashTable;

unsigned lookupOrientationValue(direction p, direction s)
{
	for(unsigned i = 0; i<24; i++)
	{
		if (dirLookup[i][0] == p && dirLookup[i][1] == s)
		{
			return i;
		}
	}

	return 0;
}

unsigned getBlockType(block b) {return b&0x00000007;}
void setBlockType(block *b, unsigned tp) {*b = (*b & ~0x00000007)|tp;}

unsigned getOrientation(block b) {return (b&0x000000f8)>>3;}
void setOrientation(block *b, unsigned orientation) {*b = (*b & ~0x000000f8)|(orientation<<3);}

unsigned getOutput(block b) {return (b&0x00000100)>>8;}
void setOutput(block *b, unsigned output) {*b = (*b & ~0x00000100)|(output<<8);}

unsigned getConnect(block b) {return (b&0x00007e00)>>9;}
void setConnect(block *b, unsigned connect) {*b = (*b & ~0x00007e00)|(connect<<9);}

unsigned getMovedir(block b) {return (b&0x00038000)>>15;}
void setMovedir(block *b, unsigned d) {*b= (*b & ~0x00038000)|(d<<15);}

void alterConnect(block *b, unsigned d, bool c)
{
	if (c)
		*b = *b | (1<<(d+9));
	else
		*b = *b & ~(1<<(d+9));
}

void markTree(otNode *o)
{
	if (o && !(o->flags & VISITED_FLAG))
	{
		o->flags |= VISITED_FLAG;

		if (o->level != 0)
		{
			markTree(o->calc[0]);
			markTree(o->calc[1]);
			markTree(o->calc[2]);
			markTree(o->calc[3]);
			markTree(o->calc[4]);

			markTree(o->u.children[0][0][0]);
			markTree(o->u.children[0][0][1]);
			markTree(o->u.children[0][1][0]);
			markTree(o->u.children[0][1][1]);
			markTree(o->u.children[1][0][0]);
			markTree(o->u.children[1][0][1]);
			markTree(o->u.children[1][1][0]);
			markTree(o->u.children[1][1][1]);
		}
	}
}

void markRandom(void)
{
	for(unsigned i = 0; i<HASH_SIZE; i++)
	{
		otNode *ptr = hashTable[i];

		if (ptr)
		{
			markTree(ptr);
			ptr = ptr->next;
		}

		while(ptr)
		{
			if (!(ptr->flags & VISITED_FLAG))
			{
				if (rand()%2 != 0)
				{
					markTree(ptr);
				}
			}

			ptr = ptr->next;
		}
	}
}

void removeUnmarked(void)
{
	for(unsigned i = 0; i<HASH_SIZE; i++)
	{
		otNode **ptrPtr = &(hashTable[i]);
		otNode *ptr = *ptrPtr;

		while(ptr)
		{
			if (ptr->flags & VISITED_FLAG)
			{
				ptr->flags &= ~VISITED_FLAG;
				
				ptrPtr = &(ptr->next);
			}
			else
			{
				*ptrPtr = ptr->next;

				destroyNode(ptr);
			}

			ptr = *ptrPtr;
		}
	}
}

// Garbage collect by traversing the tree, including all calc nodes, marking
// everything as visited, then going through the hash table and removing anything
// that is not visited
void garbageCollect(otNode *o)
{
	markTree(o);
//	markRandom();
	removeUnmarked();
}


void initTables(void)
{
        allocTable = new otNode[ALLOC_TABLE_SIZE];
        hashTable = new otNode *[HASH_SIZE];
   
	for(unsigned i = 0; i<HASH_SIZE; i++)
	{
		hashTable[i] = NULL;
	}

	for(unsigned i = 0; i<ALLOC_TABLE_SIZE-1; i++)
	{
		allocTable[i].next = &(allocTable[i+1]);
	}

	allocTable[ALLOC_TABLE_SIZE-1].next = NULL;

	freeList = &allocTable[0];
	freeCount = ALLOC_TABLE_SIZE;

	for(unsigned i = 0; i<24; i++)
	{
		for(unsigned a = 0; a<6; a++)
		{
			if (a == dirLookup[i][0])
			{
				rotateLookup[i][a] = makeOrientation[dirLookup[i][0]][opposite[dirLookup[i][2]]];
			}
			else if (a==opposite[dirLookup[i][0]])
			{
				rotateLookup[i][a] = makeOrientation[dirLookup[i][0]][dirLookup[i][2]];
			}
			else if (a == dirLookup[i][1])
			{
				rotateLookup[i][a] = makeOrientation[dirLookup[i][2]][dirLookup[i][1]];
			}
			else if (a==opposite[dirLookup[i][1]])
			{
				rotateLookup[i][a] = makeOrientation[opposite[dirLookup[i][2]]][dirLookup[i][1]];
			}
			else if (a == dirLookup[i][2])
			{
				rotateLookup[i][a] = makeOrientation[opposite[dirLookup[i][1]]][dirLookup[i][0]];
			}
			else if (a==opposite[dirLookup[i][2]])
			{
				rotateLookup[i][a] = makeOrientation[dirLookup[i][1]][opposite[dirLookup[i][0]]];
			}
		}
	}
}

unsigned hashFunction(otNode *o)
{
	if (o->flags & EMPTY_FLAG)
	{
		return o->level;
	}
	else if (o->level == 0)
	{
		return o->u.leaf+EMPTY_LIST_SIZE+((o->flags&PROPSTATIC_FLAG)?(1<<BLOCK_BITS):0);
	}
	else
	{
		return ((((unsigned)(o->u.children[0][0][0]))<<0) +
				(((unsigned)(o->u.children[0][0][1]))<<1) +
				(((unsigned)(o->u.children[0][1][0]))<<2) +
				(((unsigned)(o->u.children[0][1][1]))<<3) +
				(((unsigned)(o->u.children[1][0][0]))<<4) +
				(((unsigned)(o->u.children[1][0][1]))<<5) +
				(((unsigned)(o->u.children[1][1][0]))<<6) +
				(((unsigned)(o->u.children[1][1][1]))<<7))
		        %(HASH_SIZE-EMPTY_LIST_SIZE-(1<<(BLOCK_BITS+1)))+EMPTY_LIST_SIZE+(1<<(BLOCK_BITS+1));
	}
}

unsigned hashFunctionBlock(block b, unsigned char flags)
{
	return b+EMPTY_LIST_SIZE+((flags&PROPSTATIC_FLAG)?(1<<BLOCK_BITS):0);
}

otNode *addToHash(otNode *o, unsigned index)
{
	o->next = hashTable[index];
	return hashTable[index] = o;
}

bool compare(otNode *o1, otNode *o2)
{
	if (o1->level == o2->level)
	{
		if (o1->level == 0)
		{
			return o1->u.leaf == o2->u.leaf && 
				((o1->flags & PROPSTATIC_FLAG) == (o2->flags & PROPSTATIC_FLAG));
		}
		else if ((o1->flags & EMPTY_FLAG) || (o2->flags & EMPTY_FLAG))
		{
			return (o1->flags & EMPTY_FLAG) && (o2->flags & EMPTY_FLAG);
		}
		else
		{
			return !memcmp(&o1->u,&o2->u,8*sizeof(otNode *));
		}
	}
	else
	{
		return false;
	}
}

bool compareBlock(block b, unsigned char flags, otNode *o)
{
	return true;
#if 0
	if (o->level == 0)
	{
		return o->u.leaf == b && 
				((o->flags & PROPSTATIC_FLAG) == (flags & PROPSTATIC_FLAG));
	}
	else
	{
		return false;
	}
#endif
}

otNode *hashLookup(otNode *o, unsigned index)
{
	otNode *ptr = hashTable[index];

	while(ptr)
	{
		if (compare(o,ptr))
		{
			return ptr;
		}

		ptr = ptr->next;
	}

	return NULL;
}

#define hashLookupBlock(b,flags,index) hashTable[index];
#if 0
otNode *hashLookupBlock(block b, unsigned char flags, unsigned index)
{
	otNode *ptr = hashTable[index];

	while(ptr)
	{
		if (compareBlock(b,flags,ptr))
		{
			return ptr;
		}

		ptr = ptr->next;
	}

	return NULL;
}
#endif
// Check whether a newly created node is already in the hash
// if so then destroy o and return the one from the hash
// otherwise add o to the hash and return it
otNode *hashCheck(otNode *o)
{
	unsigned index = hashFunction(o);

	otNode *tmp = hashLookup(o,index);

	if (tmp != NULL)
	{
		destroyNode(o);
		return tmp; 
	}
	else
	{
		return addToHash(o,index);
	}
}

otNode *hashCheckBlock(block b, unsigned char flags)
{
	unsigned index = hashFunctionBlock(b,flags);

	otNode *tmp = hashLookupBlock(b,flags,index);

	if (tmp != NULL)
	{
		return tmp; 
	}
	else
	{
		otNode *o = createNode(0,false,false);
		o->u.leaf = b;
		o->flags = flags;

		return addToHash(o,index);
	}
}

block getBlockAtLocation(otNode *o, int x, int y, int z)
{
	if (o->flags & EMPTY_FLAG)
		return 0;

	switch(o->level)
	{
	case 0:
		return o->u.leaf;	
	case 1:
		return o->u.children[x][y][z]->u.leaf;
	case 2:
		return o->u.children[x/2][y/2][z/2]->u.children[x&1][y&1][z&1]->u.leaf;
	default:
		{
			int d = (1<<(o->level-1));

			int xo = !(x<d);
			int yo = !(y<d);
			int zo = !(z<d);

			return getBlockAtLocation(
				o->u.children[xo][yo][zo],
				x%d,
				y%d,
				z%d);
		}
		break;
	}
}

#define GETBLOCK_L2(o,x,y,z) \
	(o->u.children[(x)>>1][(y)>>1][(z)>>1]->u.children[(x)&1][(y)&1][(z)&1]->u.leaf)
#define GETFLAGS_L2(o,x,y,z) \
	(o->u.children[(x)>>1][(y)>>1][(z)>>1]->u.children[(x)&1][(y)&1][(z)&1]->flags)

// Make a mutable copy of the subtree in which the block resides, then
// make the change
otNode *setBlockAtLocation(otNode *o, int x, int y, int z, block b)
{
	otNode *rval = duplicateNode(o);
	rval->calc[0] = rval->calc[1] = rval->calc[2] = rval->calc[3] = NULL;
	rval->next = NULL;
	
	switch(o->level)
	{
	case 0:
		rval->u.leaf = b;
		if (getBlockType(b)==empty)
			rval->flags = EMPTY_FLAG|NOTPROP_FLAG|PROPSTATIC_FLAG|NOTACTIVE_FLAG;
		else
			rval->flags = NOTPROP_FLAG|PROPSTATIC_FLAG|NOTACTIVE_FLAG;
		break;	
	default:
		{
			int d = (1<<(o->level-1));

			int xo = ((x<d)?0:1);
			int yo = ((y<d)?0:1);
			int zo = ((z<d)?0:1);
			
			rval->u.children[xo][yo][zo] = setBlockAtLocation(rval->u.children[xo][yo][zo],
				x-xo*d,
				y-yo*d,
				z-zo*d,
				b);

			if (getBlockType(b)!=empty)
				rval->flags &= ~EMPTY_FLAG;
		}
		break;
	}

	return hashCheck(rval);
}

void destroyNode(otNode *o)
{
	o->next = freeList;
	freeList = o;
	freeCount++;
}

otNode *getEmptyNode(unsigned level);

otNode *createNode(unsigned level, bool populated, bool lookupEmpty)
{
	otNode *rval = freeList;
	freeList = freeList->next;
	freeCount--;

	rval->level = level;
	rval->calc[0] = rval->calc[1] = rval->calc[2] = rval->calc[3] = rval->calc[4] = NULL;
	rval->flags = EMPTY_FLAG|NOTPROP_FLAG|PROPSTATIC_FLAG|NOTACTIVE_FLAG;
	rval->next = NULL;
	
	switch(level)
	{
	case 0:
		rval->u.leaf = 0;
		break;
	default:
		{
			for(int i = 0; i<2; i++)
				for(int j = 0; j<2; j++)
					for(int k = 0; k<2; k++)
					{
						if (populated)
						{
							if (lookupEmpty)
							{
								rval->u.children[i][j][k] = getEmptyNode(rval->level-1);
							}
							else
							{
								rval->u.children[i][j][k] = createNode(rval->level-1,populated,lookupEmpty);
							}
						}
						else
						{
							rval->u.children[i][j][k] = NULL;
						}
					}
		}
		break;
	}

	return rval;
}

// create a duplicate node with everything the same
otNode *duplicateNode(otNode *o)
{
	otNode *rval = createNode(o->level,false,false);

	*rval = *o;

	return rval;
}

otNode *getEmptyNode(unsigned level)
{
	if (hashTable[level] != NULL)
		return hashTable[level];
	else
	{
		otNode *rval = createNode(level,true,true);
		hashTable[level] = rval;

		return rval;
	}
}

// The PROPSTATIC_FLAG is used in a different way in leaf nodes to other flags
// In leaf nodes, it is used simply to return a value from a process function

otNode *processBlockDoMove(int x, int y, int z, otNode *o)
{
	block b = GETBLOCK_L2(o,x,y,z);
	unsigned char flags;

	if (getMovedir(b))
	{
		b = 0;
	}

	for(unsigned i = 0; i<6; i++)
	{
		vector p = dirVectorLookup[i];

		block bn = GETBLOCK_L2(o,x+p.x,y+p.y,z+p.z);

		if (getMovedir(bn))
		{
			if (getMovedir(bn)-1 == opposite[i])
			{
				b = bn;
				setMovedir(&b,0);
			}
		}
	}

	flags = NOTPROP_FLAG|NOTACTIVE_FLAG|PROPSTATIC_FLAG;

	if (b==0)
	{
		flags |= EMPTY_FLAG;
	}

	return hashCheckBlock(b,flags);
}

otNode *processBlockProp(int x, int y, int z, otNode *o)
{
	block b = GETBLOCK_L2(o,x,y,z);
	unsigned char flags;

	if (b==0)
	{
		flags = EMPTY_FLAG|NOTPROP_FLAG|NOTACTIVE_FLAG|PROPSTATIC_FLAG;
		return hashCheckBlock(b,flags);
	}

	if (getMovedir(b))
	{
		flags = NOTACTIVE_FLAG|PROPSTATIC_FLAG;
		return hashCheckBlock(b,flags);
	}

	for(unsigned i = 0; i<6; i++)
	{
		vector p = dirVectorLookup[i];

		block bn = GETBLOCK_L2(o,x+p.x,y+p.y,z+p.z);

		if (getMovedir(bn))
		{
			if (getMovedir(bn)-1 == opposite[i] || ((getConnect(b) & (1<<i)) && (getConnect(bn) & (1<<opposite[i]))) )
			{
				setMovedir(&b,getMovedir(bn));
			}
		}
	}
	
	if (getMovedir(b) == 0)
	{
		flags = NOTPROP_FLAG | NOTACTIVE_FLAG | PROPSTATIC_FLAG;
		return hashCheckBlock(b,flags);
	}
	else
	{
		flags = NOTACTIVE_FLAG;
		return hashCheckBlock(b,flags);
	}
}

#define CHECK_FUSEUNFUSE(xo,yo,zo,pri,sec) \
		{ \
			block bi = GETBLOCK_L2(o,x+xo,y+yo,z+zo); \
			unsigned orientationi = getOrientation(bi); \
			unsigned tpi = getBlockType(bi); \
			if ((tpi == fuse || tpi == unfuse) && getOutput(bi)) \
			{ \
				if (dirLookup[orientationi][0] == pri && \
					dirLookup[orientationi][1] == sec) \
				{ \
					vector pc = dirVectorLookup[opposite[sec]]; \
					block bc = GETBLOCK_L2(o,x+pc.x,y+pc.y,z+pc.z); \
					\
					if (getBlockType(bc) != empty) \
						alterConnect(&b,opposite[sec],tpi==fuse); \
				} \
				if (dirLookup[orientationi][0] == sec && \
					dirLookup[orientationi][1] == pri) \
				{ \
					vector pc = dirVectorLookup[opposite[pri]]; \
					block bc = GETBLOCK_L2(o,x+pc.x,y+pc.y,z+pc.z); \
					\
					if (getBlockType(bc) != empty) \
						alterConnect(&b,opposite[pri],tpi==fuse); \
				} \
			} \
		}

otNode *processBlockActive(int x, int y, int z , otNode *o)
{
	block b = GETBLOCK_L2(o,x,y,z);
	unsigned char flags;

	unsigned tp = getBlockType(b);
	unsigned orientation = getOrientation(b);

	if (tp != empty)
	{
		if (tp == slide || tp == fuse || tp == unfuse || tp == rotate)
		{
			setOutput(&b,0);
		}

		flags = NOTPROP_FLAG|NOTACTIVE_FLAG|PROPSTATIC_FLAG;

		for(unsigned i = 0; i<6; i++)
		{
			vector p = dirVectorLookup[i];
			block bi = GETBLOCK_L2(o,x+p.x,y+p.y,z+p.z);
	
			if (getBlockType(bi) == slide && getOutput(bi) &&
				dirLookup[getOrientation(bi)][0] == opposite[i])
			{
				setMovedir(&b,dirLookup[getOrientation(bi)][1]+1);
				flags = NOTACTIVE_FLAG;
			}

			if (getBlockType(bi) == rotate && getOutput(bi) &&
				dirLookup[getOrientation(bi)][0] == opposite[i])
			{
				setOrientation(&b,rotateLookup[orientation][opposite[i]]);
				flags = NOTACTIVE_FLAG;
			}

			if ((getBlockType(bi) == fuse || getBlockType(bi) == unfuse)
				&& getOutput(bi) && dirLookup[getOrientation(bi)][0] == opposite[i])
			{
				vector pc = dirVectorLookup[dirLookup[getOrientation(bi)][1]];
				block bc = GETBLOCK_L2(o,x+pc.x,y+pc.y,z+pc.z);

				if (getBlockType(bc) != empty)
				{
					alterConnect(&b,dirLookup[getOrientation(bi)][1],getBlockType(bi)==fuse);
				}
			}
		}

		CHECK_FUSEUNFUSE(-1,-1, 0,east,north)
		CHECK_FUSEUNFUSE( 1,-1, 0,west,north)
		CHECK_FUSEUNFUSE(-1, 1, 0,east,south)
		CHECK_FUSEUNFUSE( 1, 1, 0,west,south)
		CHECK_FUSEUNFUSE(-1, 0,-1,east,front)
		CHECK_FUSEUNFUSE( 1, 0,-1,west,front)
		CHECK_FUSEUNFUSE(-1, 0, 1,east,back)
		CHECK_FUSEUNFUSE( 1, 0, 1,west,back)
		CHECK_FUSEUNFUSE( 0,-1,-1,front,north)
		CHECK_FUSEUNFUSE( 0, 1,-1,front,south)
		CHECK_FUSEUNFUSE( 0,-1, 1,back,north)
		CHECK_FUSEUNFUSE( 0, 1, 1,back,south)
	}
	else
	{
		flags = EMPTY_FLAG|NOTPROP_FLAG|PROPSTATIC_FLAG|NOTACTIVE_FLAG;
	}

	return hashCheckBlock(b,flags);
}

otNode *processBlock(int x, int y, int z, otNode *o)
{
	block b = GETBLOCK_L2(o,x,y,z);
	unsigned char flags;

	unsigned tp = getBlockType(b);
	unsigned orientation = getOrientation(b);

	if (!(o->flags & NOTACTIVE_FLAG))
	{
		return hashCheckBlock(b,GETFLAGS_L2(o,x,y,z));
	}

	{
		switch(tp)
		{
		case slide:
		case fuse:
		case unfuse:
		case rotate:
			{
				vector p = dirVectorLookup[dirLookup[orientation][0]];
				block bp = GETBLOCK_L2(o,x+p.x,y+p.y,z+p.z);

				if (getBlockType(bp) != empty)
				{
					block bi = GETBLOCK_L2(o,x-p.x,y-p.y,z-p.z);
		
					if (getOutput(bi) && 
						 ((getBlockType(bi) == wire
						&& dirLookup[getOrientation(bi)][0] != opposite[dirLookup[orientation][0]]) ||
						(getBlockType(bi) == nor
						&& dirLookup[getOrientation(bi)][0] == dirLookup[orientation][0])))
					{
						if (tp != fuse && tp != unfuse)
						{
							setOutput(&b,1);
							flags = NOTPROP_FLAG|PROPSTATIC_FLAG;
						}
						else 
						{
							vector po = dirVectorLookup[dirLookup[orientation][1]];
							block bo = GETBLOCK_L2(o,x+p.x+po.x,y+p.y+po.y,z+p.z+po.z);

							if 	(getBlockType(bo) != empty &&
								((tp==fuse && !(getConnect(bp) & (1<<dirLookup[orientation][1]))) ||
								(tp==unfuse && (getConnect(bp) & (1<<dirLookup[orientation][1]))))
								)
							{
								setOutput(&b,1);
								flags = NOTPROP_FLAG|PROPSTATIC_FLAG;
							}
							else
							{
								setOutput(&b,0);
								flags = NOTPROP_FLAG|NOTACTIVE_FLAG|PROPSTATIC_FLAG;
							}
						}
					}
					else
					{
						setOutput(&b,0);
						flags = NOTPROP_FLAG|NOTACTIVE_FLAG|PROPSTATIC_FLAG;
					}
				}
				else
				{
					setOutput(&b,0);
					flags = NOTPROP_FLAG|NOTACTIVE_FLAG|PROPSTATIC_FLAG;
				}
			}

			break;
		case wire:
			{
				vector p = dirVectorLookup[dirLookup[orientation][0]];
				block bi = GETBLOCK_L2(o,x-p.x,y-p.y,z-p.z);
		
				if (getOutput(bi) && 
					 ((getBlockType(bi) == wire
					&& dirLookup[getOrientation(bi)][0] != opposite[dirLookup[orientation][0]]) ||
					(getBlockType(bi) == nor
					&& dirLookup[getOrientation(bi)][0] == dirLookup[orientation][0])))
				{
					setOutput(&b,1);
				}
				else
				{
					setOutput(&b,0);
				}
			}

			flags = NOTPROP_FLAG|NOTACTIVE_FLAG|PROPSTATIC_FLAG;
			break;
		case nor:
			{
				unsigned op = 1;

				for(unsigned i = 0; i<6 && op; i++)
				{
					if (i != dirLookup[orientation][0])
					{
						vector p = dirVectorLookup[i];
						block bi = GETBLOCK_L2(o,x+p.x,y+p.y,z+p.z);

						if (getOutput(bi) && 
							((getBlockType(bi) == wire
							&& dirLookup[getOrientation(bi)][0] != i) ||
							(getBlockType(bi) == nor
							&& dirLookup[getOrientation(bi)][0] == opposite[i])))
						{
							op = 0;
						}
					}
				}

				setOutput(&b,op);
			}
			flags = NOTPROP_FLAG|NOTACTIVE_FLAG|PROPSTATIC_FLAG;
			break;
		default:
			flags = EMPTY_FLAG|NOTPROP_FLAG|NOTACTIVE_FLAG|PROPSTATIC_FLAG;
			break;
		}
	}

	return hashCheckBlock(b,flags);
}

otNode *otLevel2NodeCalc(otNode *o, calcFuncPtr calcFunc)
{
	otNode *rval = createNode(1,false,false);

	for(int i = 0; i<2; i++)
		for(int j = 0; j<2; j++)
			for(int k = 0; k<2; k++)
			{
				rval->u.children[i][j][k] = calcFunc(i+1,j+1,k+1,o);

				rval->flags = rval->flags & rval->u.children[i][j][k]->flags;				
			}

	return hashCheck(rval);
}

otNode *populateGapNode(otNode *o, int i2, int j2, int k2, int i1, int j1, int k1)
{
	otNode *rval = createNode(o->level-2,false,false);

	for(int i = 0; i<2; i++)
		for(int j = 0; j<2; j++)
			for(int k = 0; k<2; k++)
			{
				int x = 1+i2*2+i1*2+i;
				int y = 1+j2*2+j1*2+j;
				int z = 1+k2*2+k1*2+k;

				rval->u.children[i][j][k] = 
					o->u.children[(x&4)>>2][(y&4)>>2][(z&4)>>2]
					 ->u.children[(x&2)>>1][(y&2)>>1][(z&2)>>1]
					 ->u.children[x&1][y&1][z&1];

				rval->flags = rval->flags & rval->u.children[i][j][k]->flags;							
			}
	
	return hashCheck(rval);
}

otNode *makeTempChild(otNode *o, int i1, int j1, int k1)
{
	otNode *rval = createNode(o->level-1,false,false);

	for(int i = 0; i<2; i++)
		for(int j = 0; j<2; j++)
			for(int k = 0; k<2; k++)
			{
				rval->u.children[i][j][k] = populateGapNode(o,i1,j1,k1,i,j,k);

				rval->flags = rval->flags & rval->u.children[i][j][k]->flags;			
			}

	return hashCheck(rval);
}

otNode *makeIntermed(otNode *o, int i1, int j1, int k1)
{
	otNode *rval = createNode(o->level-1,false,false);
	
	for(int i=0;i<2;i++)
		for(int j=0;j<2;j++)
			for(int k=0;k<2;k++)
			{
				int x = i1+i;
				int y = j1+j;
				int z = k1+k;

				rval->u.children[i][j][k] = 
					o->u.children[x>>1][y>>1][z>>1]
 					->u.children[x&1][y&1][z&1];

				rval->flags = rval->flags & rval->u.children[i][j][k]->flags;
			}

	return hashCheck(rval);
}

otNode *makeInterChild(otNode **intermed,int i1, int j1, int k1)
{
	otNode *rval = createNode((*intermed)->level+1,false,false);
	
	for(int i=0;i<2;i++)
		for(int j=0;j<2;j++)
			for(int k=0;k<2;k++)
			{
				int x = i + i1;
				int y = j + j1;
				int z = k + k1;

				rval->u.children[i][j][k] = *(intermed + 3*(3*x+y) + z); 

				rval->flags = rval->flags & rval->u.children[i][j][k]->flags;
			}

	return hashCheck(rval);
}

// returns a pointer to a node one level down and centred
otNode *otNodeCalc(otNode *o, calcFuncPtr calcFunc, unsigned calcIndex)
{
	otNode *rval, *tmp;

	if (o->calc[calcIndex] == NULL)
	{
		if (o->level==2)
		{
			o->calc[calcIndex] = otLevel2NodeCalc(o,calcFunc);
		}
		else if (calcIndex == CALCINDEX_CALC2 && (o->level==3 || 
o->level==4 || o->level==5))
		{
			unsigned nextCalcIndex = (o->level==3)?CALCINDEX_CALC:CALCINDEX_CALC2;
			rval = createNode(o->level-1,false,false); 

			otNode *intermed[3][3][3];

			for (int i=0; i<3; i++)
				for(int j=0; j<3; j++)
					for(int k=0; k<3; k++)
					{
						intermed[i][j][k] = makeIntermed(o,i,j,k);
						intermed[i][j][k] = otNodeCalc(intermed[i][j][k],calcFunc,nextCalcIndex);
					}

			for(int i=0; i<2; i++)
				for(int j=0; j<2; j++)
					for(int k=0; k<2; k++)
					{				
						rval->u.children[i][j][k] = makeInterChild(&intermed[0][0][0],i,j,k);
						rval->u.children[i][j][k] = otNodeCalc(rval->u.children[i][j][k],calcFunc,nextCalcIndex);
						rval->flags = rval->flags & rval->u.children[i][j][k]->flags;
					}

			o->calc[calcIndex] = hashCheck(rval);
		}
		else
		{
			rval = createNode(o->level-1,false,false);

			for (int i=0; i<2; i++)
				for(int j=0; j<2; j++)
					for(int k=0; k<2; k++)
					{
						rval->u.children[i][j][k] = makeTempChild(o,i,j,k);

						rval->u.children[i][j][k] = otNodeCalc(rval->u.children[i][j][k],calcFunc,calcIndex);

						rval->flags = rval->flags & rval->u.children[i][j][k]->flags;			
					}

			o->calc[calcIndex] = hashCheck(rval);
		}

		return o->calc[calcIndex];
	}
	else
	{
		return o->calc[calcIndex];
	}
}


// determine whether the universe needs to be expanded, and if so then do it
otNode *expandUniverse(otNode *o, bool testNeed)
{
	bool doExpand;

	if (testNeed && o->level >= 2)
	{
		doExpand = false;

		for(int i = 0; i<2; i++)
			for(int j = 0; j<2; j++)
				for(int k = 0; k<2; k++)
					for(int i1 = 0; i1 < 2; i1++)
						for(int j1 = 0; j1<2; j1++)
							for(int k1 = 0; k1<2; k1++)
							{
								if (i1 != 1-i || j1 != 1-j || k1 != 1-k)
								{
									if (!o->u.children[i][j][k]->u.children[i1][j1][k1]->flags & EMPTY_FLAG)
									{
										doExpand = true;
									}
								}
							}
	}
	else
	{
		doExpand = true;
	}

	if (doExpand)
	{
		otNode *rval = createNode(o->level+1,false,false);

		rval->flags = o->flags;

		for(int i = 0; i<2; i++)
			for(int j = 0; j<2; j++)
				for(int k = 0; k<2; k++)
				{
					otNode *tmp = createNode(o->level,true,true);
					tmp->u.children[1-i][1-j][1-k] = o->u.children[i][j][k];
					tmp->flags = o->u.children[i][j][k]->flags;
					rval->u.children[i][j][k] = hashCheck(tmp);
				}

		return hashCheck(rval);
	}
	else
	{
		return o;
	}
}

// x,y,z can be negative, so treat block 0,0,0 as the first block in the 1,1,1 child
otNode *loadBlock(block b, int x, int y, int z, otNode *o)
{
	int d = (1<<o->level);

	while(x < -d/4 || y < -d/4 || z < -d/4 || x>=d/4 || y>=d/4 || z>=d/4)
	{
		o = expandUniverse(o,false);
		d = (1<<o->level);
	}

	o = setBlockAtLocation(o,x+d/2,y+d/2,z+d/2,b);

	return o;
}



int tpConvertToGrd(int tp)
{
	int typeMap[] =
	{
		-1,
		wire,
		nor,
		-1,
		slide,
		rotate,
		fuse,
		unfuse,
		-1,
		-1,
		-1,
	};

	for(int i = 0; i<= 10; i++)
	{
		if (typeMap[i] == tp)
			return i;
	}

	return 0;
}

// Convert tp from GRD file format to values suitable for this implementation
int tpConvertFromGrd(int tp)
{
	int typeMap[] =
	{
		-1,
		wire,
		nor,
		-1,
		slide,
		rotate,
		fuse,
		unfuse,
		-1,
		-1,
		-1,
	};

	if (tp >= 0 && tp <=10)
		return typeMap[tp];
	else
		return -1;
}


// N.B. this implementation ignores regions and skips over anything to
// do with regions
otNode *load(std::ifstream &is)
{
	static char token[128];
	int xd,yd,zd;
	int xp,yp,zp;
	int tp;
	int np,ep;
	int tmp;
	bool readregion;

	otNode *o = createNode(2,true,true);

	while(is)
	{
	  is >> token;

	  // If the token read is 'record', this is a new style file format and one of several
	  // types of record will be read. Otherwise it is an old style file format and the token
	  // must be the x coordintate of a region.

	  if (!strcmp(token,"record"))
	  {
		  is >> token;

		  if (!strcmp(token,"region"))
		  {
			  is >> xp;
			  readregion = true;
		  }
		  else if (!strcmp(token,"identifier"))
		  {
			  is >> token;
			  readregion = false;
		  }
		  else if (!strcmp(token,"up"))
		  {
			  readregion = false;
		  }
	  }
	  else
	  {
		  xp = atoi(token);
		  readregion = true;
	  }

	  if (readregion)
	  {
		  is >> yp >> zp;
		  is >> xd >> yd >> zd;

		  if (is)
		  {
			  for(unsigned i = 0; i<yd; i++)
			  {
				for(unsigned j = 0; j<zd; j++)
				{
					for(unsigned k = 0; k<xd; k++)
					{
						is >> tp;
						is >> np >> ep;

						block b = 0;

						tp = tpConvertFromGrd(tp);

						if (tp != -1)
						{
						  setBlockType(&b,tp);
						  setOrientation(&b,lookupOrientationValue((direction)np,(direction)ep));

						  for(unsigned l = 0; l<6; l++)
						  {
							is >> tmp;
						    if (tmp)
								setOutput(&b,1);
						  }
						
						  for(unsigned l = 0; l<6; l++)
						  {
							is >> tmp;
							alterConnect(&b,l,tmp);
						  }

						  o = loadBlock(b,xp+k,yp+i,zp+j,o);
						}
						else
						{
							for(unsigned l = 0; l<12; l++)
							{
								is >> tmp;
							}
						}
					}
				}
			  }
		  }
		}
	}

	return o;
}

void save(otNode *eg, std::ofstream &os, int xo, int yo, int zo)
{
	int d = (1<<eg->level);

	if (eg->flags & EMPTY_FLAG)
	{
		return;
	}
	else if (eg->level < 2)
	{
		return;
	}
	else if (eg->level == 2)
	{
		os << "record region " << xo << " " << yo << " " << zo << " 4 4 4" << std::endl;

		for(int j = 0; j<4 ;j++)
			for(int k = 0; k<4; k++)
				for(int i = 0; i<4; i++)
				{
					block b = getBlockAtLocation(eg,i,j,k);
					unsigned tp = getBlockType(b);
					unsigned orientation = getOrientation(b);
					
					os << tpConvertToGrd(tp) << " ";
					os << dirLookup[orientation][0] << " " << dirLookup[orientation][1] << std::endl;

					unsigned output = getOutput(b);

					for(int o = 0; o<6; o++)
					{
						if (output)
						{
							if ((tp == wire && o != opposite[dirLookup[orientation][0]]) ||
							    (tp == nor && o == dirLookup[orientation][0]))
							{
								os << 1 << " ";
							}
							else
							{
								os << 0 << " ";
							}
						}
						else
						{
							os << 0 << " ";
						}

					}
					os << std::endl;

					for(int c = 0; c<6; c++)
						os << ((getConnect(b)&(1<<c))?1:0) << " ";

					os << std::endl;
				}

		return;
	}
	else
	{
		for(int i = 0; i<2; i++)
			for(int j = 0; j<2; j++)
				for(int k = 0; k<2; k++)
					save(eg->u.children[i][j][k],os,
					xo-d/4+i*(d/2),
					yo-d/4+j*(d/2),
					zo-d/4+k*(d/2)
					);
	}
}

otNode *singleStep(otNode *eg, unsigned &iters)
{
	bool doneNormal = true;
	static int mode = 0;
	static int modeCounter = 0;

	do
	{
		doneNormal = false;

		if (eg->flags & NOTPROP_FLAG)
		{
			if (eg->flags & NOTACTIVE_FLAG)
			{
				otNode *tmp;

				if (mode == 0)
					tmp = expandUniverse(otNodeCalc(eg,processBlock,CALCINDEX_CALC2),false);
				
				if (mode || !(tmp->flags & NOTACTIVE_FLAG))
				{
					eg = expandUniverse(otNodeCalc(eg,processBlock,CALCINDEX_CALC),false);
					iters+=1;
				}
				else
				{
					eg = tmp;
					iters+=8;
				}
 
				if (eg->flags & NOTACTIVE_FLAG)
				{
					modeCounter++;
					doneNormal = true;
				}
				else
				{
					modeCounter = 0;
					mode = 1;
				}

				if (modeCounter > 8) mode = 0;
			}
			else
			{
				eg = expandUniverse(otNodeCalc(eg,processBlockActive,CALCINDEX_ACTIVE),false);

				if (eg->flags & NOTPROP_FLAG)
				{
					doneNormal = true;
				}
			}
		}
		else
		{
			if (eg->flags & PROPSTATIC_FLAG)
			{
				eg = expandUniverse(otNodeCalc(eg,processBlockDoMove,CALCINDEX_DOMOVE),false);
				doneNormal = true;
			}
			else
			{
				eg = expandUniverse(expandUniverse(otNodeCalc(eg,processBlockProp,CALCINDEX_PROP),true),false);
			}
		}

		if (freeCount < GC_THRESHHOLD)
		{
			garbageCollect(eg);
		        printf("GC\n");
		}
	}
	while(!doneNormal);

	return eg;
}
