#include "signalpath.cpp"

/* Decoder that decodes the sequence 0-0-x-1-0 (where the lhb is the first into the decoder) 
 * x is a 'don't care' value (so are -)
 * Assumes that 1 is applied to the input
 */
void MotorDecoder1(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{nor {f n}} {del {b e}}");
  bo.String("{nor {w n}} {del {w n}}");
  bo.String("{del {b e}} {del {f n}}");
  bo.String("{nor {b e}} {nul {n e}}");
  bo.String("{del {n e}} {nul {n e}}");  // Note that the del ne here is for the slide to push.
  bo.Layer();
  bo.String("{nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {del {n e} e}");
  bo.String("{nor {s e}} {del {f e} e}");
  bo.String("{del {w n}} {del {f e} e}");

  bo.SetOutputs(1,0,0,true);
  bo.SetOutputs(1,0,1,true);
  bo.SetOutputs(1,0,2,true);
  bo.SetOutputs(1,1,2,true);
  bo.SetOutputs(1,1,3,true);
  bo.SetOutputs(1,1,4,true);
  bo.SetOutputs(0,1,4,true);
  bo.SetOutputs(0,0,3,true);
  bo.SetOutputs(0,0,2,true);
  
  bo.Terminal("i",1,0,1,true,east);
  bo.Terminal("o",0,0,1,false,west);
}

void MotorDecoder2(CBasicObject &bo, bool b0)
{
  bo.Layer();
  bo.String("{nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}}");
  bo.String("{del {w n}} {nor {w n}}");

  bo.Layer();
  if (b0)
    bo.String("{nor {f n}} {nor {w n}}");
  else
    bo.String("{nor {f n}} {del {b e}}");

  bo.String("{nor {w n}} {del {w n}}");
  bo.String("{del {b n}} {del {f n}}");
  bo.String("{nor {b n}} {nor {f n}}");
  bo.String("{nor {b n}} {del {f n}}");


  if (b0)
  {
    bo.SetOutputs(1,1,0,true);
  }
  else
  {
    bo.SetOutputs(0,1,0,true);
  }
  bo.SetOutputs(1,1,1,true);
  bo.SetOutputs(1,1,2,true);
  bo.SetOutputs(1,0,4,true);
  bo.SetOutputs(0,0,4,true);
  bo.SetOutputs(0,1,3,true);
  bo.SetOutputs(0,1,2,true);

  bo.Terminal("i",1,1,1,true,east);
  bo.Terminal("o",0,1,1,true,west);

  // Special attachment for thing needed for constructee
  bo.Terminal("attach",1,1,4,true,east);
}

// May not be used if MotorDecoder2 is sufficient for both
void MotorDecoder3(CBasicObject &bo, bool b0)
{
  bo.Layer();
  if (b0)
    bo.String("{nor {f n}} {nor {w n}}");
  else
    bo.String("{nor {f n}} {del {b e}}");

  bo.String("{nor {w n}} {del {w n}}");
  bo.String("{del {b n}} {del {f n}}");
  bo.String("{nor {b n}} {nor {f n}}");
  bo.String("{nor {b n}} {del {f n}}");
  bo.String("{del {w n}} {nor {w n}}");


  if (b0)
  {
    bo.SetOutputs(1,0,0,true);
  }
  else
  {
    bo.SetOutputs(0,0,0,true);
  }
  bo.SetOutputs(1,0,1,true);
  bo.SetOutputs(1,0,2,true);
  bo.SetOutputs(1,0,5,true);
  bo.SetOutputs(0,0,5,true);
  bo.SetOutputs(0,0,3,true);
  bo.SetOutputs(0,0,2,true);

  bo.Terminal("i",1,0,1,true,east);
  bo.Terminal("o",0,0,1,true,west);
}

void MotorFrame(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nor {f n}} {del {w n}} {nor {w n}} {nor {w n}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nor {n e}} {nor {f e}} {nul {n e}} {nor {b n}} {del {s e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nor {f e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nor {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.Layer();
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {del {b e}} {nor {e n}} {del {e n}} {del {e n} n}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nor {b e}} {nul {n e}} {nul {n e}} {fus {f n} n}");
  bo.String("{del {w n}} {nor {w n}} {nor {w n}} {del {w n}} {nul {n e}} {nul {n e}} {nor {w n}}");
  bo.String("{del {f n}} {sli {e b}} {nor {f n}} {del {f n}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nor {f n}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {del {f n}} {sli {e f}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.Layer();
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nor {e n}} {nor {e n}} {ufs {e s} e} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");

  bo.SetOutputs(6,1,2,true);
  bo.SetOutputs(2,1,2,true);
  bo.SetOutputs(2,1,3,true);
  bo.SetOutputs(3,1,1,true);
  bo.SetOutputs(3,1,0,true);

  bo.SetOutputs(5,0,2,true);
  bo.SetOutputs(4,0,1,true);
  bo.SetOutputs(1,0,1,true);
  bo.SetOutputs(2,0,2,true);
  bo.SetOutputs(2,0,4,true);

  bo.SetOutputs(3,2,2,true);

  bo.Terminal("i",6,1,2,true,north);
  bo.Terminal("di",3,1,2,true,east); // input from decoder
  bo.Terminal("attach",6,1,2,false,east);
  bo.Terminal("chattach",5,2,2,false,north);
}

// Special attachment that had to be added to the EW motor of the constructee
// to fuse parts when constructing.
void ConstructeeFuseAttach(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{nor {n e} x n}");
  bo.Layer();
  bo.String("{fus {n f} n}");

  bo.Terminal("attach",0,1,0,false,west);
}

void Motor(CCompoundObject &co, unsigned i)
{
  CBasicObject *motorframe = new CBasicObject;
  CBasicObject *decoder = new CBasicObject;

  MotorFrame(*motorframe);
  
  switch(i)
  {
    case 0:
      MotorDecoder1(*decoder);
      break;
    case 1:
      MotorDecoder2(*decoder,true);
      break;
    default:
      MotorDecoder2(*decoder,false);
      break;
  }

  co.AddObject(motorframe,"motorframe",0,0,0);
  co.AddObject(decoder,"decoder","o","motorframe.di");

  if (i==1)
  {
    CBasicObject *cfa = new CBasicObject;
    ConstructeeFuseAttach(*cfa);

    co.AddObject(cfa,"cfa","attach","decoder.attach");
  }
}

// The 'side' parameter below is needed because if the constructor arm is
// mirrored, the signal path connection to the main body of the srm
// needs to be kept as it was before mirroring.
void Path(CBasicObject &bo, unsigned l, bool side=true, unsigned omz = 0)
{
  bo.Layer();
  {for(unsigned i = 0; i<l; i++)
  {
    if (i==l-1)
      if (side)
        bo.String("{del {w n}}");
      else
        bo.String("{del {e n}}");
    else
      bo.String("{del {b n}}");
  }}

  bo.Layer();
  {for(unsigned i = 0; i<l; i++)
  {
    bo.String("{nor {n b} x n}");
  }}

  bo.Layer();
  {for(unsigned i = 0; i<l; i++)
  {
    bo.String("{sli {n b}}");
  }}

  bo.Terminal("om",0,0,omz,false,south);
  bo.Terminal("i",0,0,l-1,true,(side?east:west));
  bo.Terminal("bi",0,3,l-1,true,front);
  bo.Terminal("c",0,1,l-1,true,front);
  bo.Terminal("bpjattach",0,2,l-1,false,east);
}

void PathJoin(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{nul {n e}}");
  bo.String("{del {n e}}");
  bo.String("{nul {n e}}");
  bo.String("{nul {n e}}");
  bo.Layer();
  bo.String("{nul {n e}}");
  bo.String("{del {e n} w}");
  bo.String("{nul {n e}}");
  bo.String("{nul {n e}}");
  bo.Layer();
  bo.String("{nor {w n} w w}");
  bo.String("{nor {n e} w n}");
  bo.String("{nul {n e}}");
  bo.String("{nul {n e}}");
  bo.Layer();
  bo.String("{nul {n e}}");
  bo.String("{del {n e} w newfb}");
  bo.String("{nor {f e} w}");
  bo.String("{del {f e} w}");
  
  bo.Terminal("o",0,3,3,false,north);
  bo.Terminal("attach",0,0,1,true,west); 
}

void BlockPathJoin(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{sli {s e}} {del {s e}}");
  bo.String("{nul {n e}} {del {s e}}");
  bo.Layer();
  bo.String("{nor {s e} x s} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}}");

  bo.Terminal("attach",1,0,1,true,south);
}

// Special BlockPathJoin that contains the extra parts that needed to be
// added to the constructee on the join for the EW to NS paths
void BlockPathJoinConstructeeEWNS(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{sli {s e}} {del {s e}}");
  bo.String("{nul {n e}} {fus {s f}}");
  bo.String("{nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nor {e n} x e}");
  bo.Layer();
  bo.String("{nor {s e} x s} {nul {n e}}");
  bo.String("{nul {n e}} {nor {s e}}");
  bo.String("{nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}}");

  bo.Terminal("attach",1,0,1,true,south);
}


void ConstructorHead(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {fus {w s}} {nor {w n} x w}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");

  bo.Layer();
  bo.String("{nul {n e}} {del {w n}} {del {w n}} {del {w n}} {del {s e}}");
  bo.String("{nul {n e}} {fus {f w}} {fus {s f}} {nul {n e}} {fus {f e}}");
  bo.String("{nul {n e}} {fus {s e}} {nul {n e}} {sli {w s}} {nor {w n} x w}");
  bo.String("{nul {n e}} {ufs {b w}} {fus {s b}} {nul {n e}} {ufs {b e}}");
  bo.String("{nul {n e}} {del {w n}} {del {w n}} {del {w n}} {del {s e}}");

  bo.Layer();
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nor {s b} n s} {del {s e}} {nul {n e}}");
  bo.String("{nul {n e}} {nor {s e} n s} {nul {n e}} {sli {w s}} {nor {w n} x w}");
  bo.String("{nul {n e}} {del {s e} n} {nor {s b} n s} {del {w n}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");

  bo.Layer();
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {sli {w s}} {nor {w n} x w}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");

  bo.Layer();
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nor {s e} s s} {del {w n} s} {del {w n}} {del {w n}} {del {e n}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");

  bo.Terminal("attach",0,4,2,true,west);

  bo.Terminal("fuse-i",4,1,0,true,north);
  bo.Terminal("unfuse-i",4,1,4,true,north);
}


void Constructor(CCompoundObject &co, unsigned omz = 0)
{
  CCompoundObject *motor1 = new CCompoundObject;
  CCompoundObject *motor2 = new CCompoundObject;
  CCompoundObject *motor3 = new CCompoundObject;
  CBasicObject *path1 = new CBasicObject;
  CBasicObject *path2 = new CBasicObject;
  CBasicObject *path3 = new CBasicObject;
  CBasicObject *pathjoin1 = new CBasicObject;
  CBasicObject *pathjoin2 = new CBasicObject;
  CBasicObject *blockpathjoin1 = new CBasicObject;
  CBasicObject *blockpathjoin2 = new CBasicObject;
  CBasicObject *constructorhead = new CBasicObject;
  CCompoundObject *cheadfusepath = new CCompoundObject;
  CCompoundObject *cheadunfusepath = new CCompoundObject;

  Motor(*motor1,0);
  Path(*path1,28,false,omz);
  PathJoin(*pathjoin1);
  BlockPathJoin(*blockpathjoin1);

  Path(*path2,18,false,12);
  path2->Rotate(back);
  path2->Rotate(north);
  Motor(*motor2,1);
  motor2->Rotate(back);
  motor2->Rotate(north);
  PathJoin(*pathjoin2);
  pathjoin2->Rotate(back);
  pathjoin2->Rotate(north);
  BlockPathJoinConstructeeEWNS(*blockpathjoin2);
  blockpathjoin2->Rotate(back);
  blockpathjoin2->Rotate(north);

  Path(*path3,19,false,10);
  path3->Rotate(south);
  path3->Rotate(front);
  Motor(*motor3,2);
  motor3->Rotate(south);
  motor3->Rotate(front);

  ConstructorHead(*constructorhead);
  constructorhead->Rotate(west);

  SignalPathJointed(*cheadfusepath,"e 6 n 2 b 7");
  SignalPathJointed(*cheadunfusepath,"e 2 n 1 e 4 b 5");

  co.AddObject(path1,"path1",0,0,0);
  co.AddObject(motor1,"motor1","motorframe.i","path1.om");
  co.AddObject(pathjoin1,"pathjoin1","attach","motor1.motorframe.attach");
  co.AddObject(path2,"path2","i","pathjoin1.o");
  co.AddObject(blockpathjoin1,"blockpathjoin1","attach","path2.bpjattach");
  co.AddObject(motor2,"motor2","motorframe.i","path2.om");
  co.AddObject(pathjoin2,"pathjoin2","attach","motor2.motorframe.attach");
  co.AddObject(path3,"path3","i","pathjoin2.o");
  co.AddObject(blockpathjoin2,"blockpathjoin2","attach","path3.bpjattach");
  co.AddObject(motor3,"motor3","motorframe.i","path3.om");
  co.AddObject(constructorhead,"constructorhead","attach","motor3.motorframe.chattach");
  co.AddObject(cheadfusepath,"cheadfusepath","o","constructorhead.fuse-i");
  co.AddObject(cheadunfusepath,"cheadunfusepath","o","constructorhead.unfuse-i");
}

#if 0
void Constructor(CCompoundObject &co)
{
  CBasicObject *motor1,*motor2,*motor3;
  CBasicObject *path1,*path2,*path3;
  CBasicObject *pathjoin12,*pathjoin23;
  CBasicObject *chead;
  CCompoundObject *cheadfusepath,*cheadunfusepath;

  motor1 = new CBasicObject;
  motor2 = new CBasicObject;
  motor3 = new CBasicObject;
  path1 = new CBasicObject;
  path2 = new CBasicObject;
  path3 = new CBasicObject;
  pathjoin12 = new CBasicObject;
  pathjoin23 = new CBasicObject;
  chead = new CBasicObject;
  cheadfusepath = new CCompoundObject;
  cheadunfusepath = new CCompoundObject;


  Motor(*motor1,false,false);
  Motor(*motor2,true,false);
  Motor(*motor3,false,true);
  Path(*path1,20,false);
  Path(*path2,20);
  Path(*path3,20);
  PathJoin(*pathjoin12);
  PathJoin(*pathjoin23);
  ConstructorHead(*chead);
  SignalPathJointed(*cheadfusepath,"e 5 n 3 b 9");
  SignalPathJointed(*cheadunfusepath,"e 6 n 2 b 7");
  path2->Rotate(back);
  path2->Rotate(north);

  path3->Rotate(front);
  path3->Rotate(west);

  motor2->Rotate(back);
  motor2->Rotate(north);
  motor3->Rotate(south);
  motor3->Rotate(front);

  pathjoin12->Rotate(back);

  pathjoin23->Rotate(back);
  pathjoin23->Rotate(east);
  pathjoin23->Rotate(back);

  chead->Rotate(back);
  chead->Rotate(front);
  chead->Rotate(west);

  co.AddObject(motor1,"motor1",0,0,0);
  co.AddObject(path1,"path1","om","motor1.i");
  co.AddObject(pathjoin12,"pathjoin12","i","motor1.o");
  co.AddObject(path2,"path2","i","pathjoin12.o");
  co.AddObject(motor2,"motor2","i","path2.om");
  co.AddObject(pathjoin23,"pathjoin23","i","motor2.o");
  co.AddObject(path3,"path3","i","pathjoin23.o");
  co.AddObject(motor3,"motor3","i","path3.om");
  co.AddObject(chead,"chead","i","motor3.o");
  co.AddObject(cheadfusepath,"cheadfusepath","o","chead.fuse-i");
  co.AddObject(cheadunfusepath,"cheadunfusepath","o","chead.unfuse-i");
}
#endif