#include "memorycontrol.cpp"
#include "instructiondecode.cpp"
#include "orientation.cpp"
#include "inhibit.cpp"
#include "call.cpp"


void SerialEncoder(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{del {w n}} {del {w n}} {del {s e}}");
  bo.String("{nor {f n}} {del {w n}} {del {w n}}");
  bo.String("{nor {f n}} {nul {n e}} {nul {n e}}");
  bo.String("{nor {f n}} {del {w n}} {del {s e}}");
  bo.Layer();
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {del {w n}}");
  bo.String("{nul {n e}} {nul {n e}} {del {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {del {s e}}");
  bo.Layer();
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{del {w n}} {del {w n}} {del {n e}}");
  bo.String("{del {f n}} {del {b e}} {del {e n}}");
  bo.String("{del {f n}} {del {e n}} {del {f e}}");

  bo.Terminal("o",0,0,4,false,front);
  bo.Terminal("di",2,0,2,true,east);
  bo.Terminal("pti",2,1,1,true,east);

  bo.SetOutputs(0,0,2,true);
  bo.SetOutputs(0,0,4,true);
}

void SerialEncoderExtension(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{del {w n}} {del {w n}}");
  bo.Terminal("i",1,0,0,true,east);
  bo.Terminal("o",0,0,0,false,west);
}

void InhibitSignal(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{del {e n}} {nor {f n}}");
  bo.String("{nul {n e}} {del {f n}}");
  bo.Layer();
  bo.String("{nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {del {n e}}");
  bo.Layer();
  bo.String("{nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {del {n e}}");

  bo.SetOutputs(0,0,0,true);

  bo.Terminal("i",0,0,0,true,west);
  bo.Terminal("o",1,2,1,false,north);
}

// The return path goes from the return decoder to two places on
// the counterstack. The path contains a 1-to-4 pulse lengthener
void ReturnPath(CCompoundObject &co)
{
  CCompoundObject *path1 = new CCompoundObject();
  CCompoundObject *path2 = new CCompoundObject();
  CBasicObject *pg = new CBasicObject();
  CBasicObject *path3 = new CBasicObject();
  CCompoundObject *countinhibit = new CCompoundObject();


  SignalPathJointed(*path1,"w 1 f 8 e 0");
  PulseGen2Long(*pg);
  pg->Rotate(west);
  SignalPathJointed(*path2,"e 3 s 6 f 3");
  SignalPath(*path3,3,1,1,west,front,north,back,west,true,true);
  SignalPathJointed(*countinhibit,"e 2 s 1 e 1");

  co.AddObject(path1,"path1",0,0,0);
  co.AddObject(pg,"pg","i","path1.o");
  co.AddObject(path2,"path2","i","pg.o2");
  co.AddObject(path3,"path3","i","path2.o");
  co.AddObject(countinhibit,"countinhibit","i","pg.o3");

  co.Terminal("i","path1.i");
}

void ControlUnit(CCompoundObject &co, unsigned *prog, unsigned length)
{
  CCompoundObject *mc,*orientation;
  CBasicObject *mpa,*mpc,*mpd,*orientdp;
  CBasicObject *idprb,*idpmo,*serb,*seerb,*semo,*seemo;
  CBasicObject *inhibitg[6];
  CBasicObject *ilw,*ilf,*ile,*ilb,*ilee,*ileee;
  CCompoundObject *ihbl;
  CBasicObject *orient2inhibit;
  CBasicObject *call2inhibit;
  CBasicObject *inhibitreset;
  CCompoundObject *call;
  CBasicObject *callretdecode;
  CBasicObject *callpath;
  CBasicObject *callresetpath;
  CBasicObject *callresetdelay19;
  CBasicObject *callreg2;
  CBasicObject *callsigadapt;
  CCompoundObject *callsiggate;
  CCompoundObject *callpulsegate;
  CCompoundObject *callstackinput;
  CBasicObject *counterloadenable;
  CCompoundObject *returnpath;
  CCompoundObject *returnpathbranch;

  mc = new CCompoundObject();
  orientation = new CCompoundObject();
  mpa = new CBasicObject();
  mpc = new CBasicObject();
  mpd = new CBasicObject();
  orientdp = new CBasicObject();
  idprb = new CBasicObject();
  idpmo = new CBasicObject();
  serb = new CBasicObject();
  seerb = new CBasicObject();
  semo = new CBasicObject();
  seemo = new CBasicObject();
  orient2inhibit = new CBasicObject();
  call2inhibit = new CBasicObject();
  inhibitreset = new CBasicObject();

  {for(unsigned i = 0; i<6; i++)
  {
    inhibitg[i] = new CBasicObject();
  }}

  ilw = new CBasicObject();
  ilf = new CBasicObject();
  ile = new CBasicObject();
  ilb = new CBasicObject();
  ilee = new CBasicObject();
  ileee = new CBasicObject();
  ihbl = new CCompoundObject();
  call = new CCompoundObject();
  callretdecode = new CBasicObject();
  callpath = new CBasicObject();
  callresetpath = new CBasicObject();
  callresetdelay19 = new CBasicObject();
  callreg2 = new CBasicObject();
  callsigadapt = new CBasicObject();
  callsiggate = new CCompoundObject();
  callpulsegate = new CCompoundObject();
  callstackinput = new CCompoundObject();
  counterloadenable = new CBasicObject();
  returnpath = new CCompoundObject();
  returnpathbranch = new CCompoundObject();

  MemoryControl(*mc,prog,length);
  Orientation(*orientation);
  MemoryPath(*mpa);
  MemoryPath(*mpc);
  MemoryPath(*mpd);
  InhibitGateWest(*(inhibitg[0]));
  InhibitGateWest(*(inhibitg[1]));
  InhibitGateBack(*(inhibitg[2]));
  InhibitGateBack(*(inhibitg[3]));
  InhibitGateWest(*(inhibitg[4]));
  InhibitGateBack(*(inhibitg[5]));

  InhibitLineWest(*ilw);
  InhibitLineFront(*ilf);
  InhibitLineEast(*ile);
  InhibitLineBack(*ilb);

  SignalPath(*ilee,3,5,1,east,north,back,west,north);
  SignalPath(*ileee,11,2,11,east,back,south,south,west);

//  InhibitSignal(*ihbs);
  InhibitLogic(*ihbl);
  SignalPath(*orientdp,1,3,9,south,front,east,front,front);
  SignalPath(*orient2inhibit,10,2,4,front,east,north,back,north);
  SignalPath(*inhibitreset,1,9,1,south,front,west,back,south);
  SignalPath(*call2inhibit,4,4,3,front,south,west,back,west);

  InstructionDecodePairSerialOut(*idprb,0x00000011,0x00000021,back); // decoder for dispense and for ori
  InstructionDecodePairSerialOut(*idpmo,0x00000001,0x0000000f,back); // decoder for move and for gather
  SerialEncoder(*serb);
  SerialEncoderExtension(*seerb);
  SerialEncoder(*semo);
  SerialEncoderExtension(*seemo);

  InstructionDecodePairCall(*callretdecode,0x0000000b,0x00000031,west,back); // decoder for return and for call
  CallRegister(*call);
  SignalPath(*callpath,1,3,10,south,front,west,front,west);
  SignalPath(*callresetpath,2,1,1,west,back,south,east,south,true,false);
  Delay19(*callresetdelay19);
  callresetdelay19->Rotate(west);
  CallRegister2(*callreg2);
  callreg2->Rotate(south);

  CallSignalsAdapter(*callsigadapt);
  CallSignalsGate(*callsiggate);
  CallPulseGate(*callpulsegate);
  CallStackInputPaths(*callstackinput);

  SignalPath(*counterloadenable,2,5,2,front,south,west,south,west,true,true);

  ReturnPath(*returnpath);
  SignalPathJointed(*returnpathbranch,"f 2 n 3 w 1");

  co.AddObject(mc,"memcontrol",0,0,0);
  co.AddObject(mpa,"mpa","i4","memcontrol.msbp.sbp5.os");
  co.AddObject(idpmo,"idpmo","i4","mpa.o4");
  co.AddObject(mpc,"mpc","i4","idpmo.o4");
  co.AddObject(mpd,"mpd","i4","mpc.o4");
  co.AddObject(idprb,"idprb","i4","mpd.o4");

  co.AddObject(inhibitg[0],"inhw0","i","idpmo.da");
  co.AddObject(inhibitg[1],"inhw1","i","idprb.da");

  co.AddObject(inhibitg[2],"inhb0","i","idpmo.db");
  co.AddObject(inhibitg[3],"inhb1","i","idprb.db");

  co.AddObject(ilw,"ilw","o0","inhw0.g");
  co.AddObject(ilf,"ilf","o","ilw.i");
  co.AddObject(ile,"ile","j","ilf.j");
  co.AddObject(ilb,"ilb","i","ile.o");

  co.AddObject(serb,"serb","di","inhw1.o");
  co.AddObject(seerb,"seerb","o","serb.pti");

  co.AddObject(semo,"semo","di","inhw0.o");
  co.AddObject(seemo,"seemo","o","semo.pti");

  co.AddObject(orientdp,"orientdp","i","inhb1.o");
  co.AddObject(orientation,"orientation","offi.b4-in","idprb.o4");

  co.AddObject(ihbl,"ihbl","inhibitff.o","ilf.i");

  co.AddObject(orient2inhibit,"orient2inhibit","o","ihbl.setor.is");
  co.AddObject(call2inhibit,"call2inhibit","o","ihbl.setor.ie");
  co.AddObject(inhibitreset,"inhibitreset","o","ihbl.inhibitff.r");

  co.AddObject(callretdecode,"callretdecode","i4","memcontrol.msbp.sbp5.o");
  co.AddObject(call,"callreg","off2.set-in","callretdecode.o0");

  co.AddObject(inhibitg[4],"inhw2","i","callretdecode.da");
  co.AddObject(inhibitg[5],"inhb2","i","callretdecode.db");

  co.AddObject(returnpath,"returnpath","i","inhw2.o");
  co.AddObject(returnpathbranch,"returnpathbranch","o","memcontrol.cntstack.loadpathor.iret");

  co.AddObject(ilee,"ilee","i","ilf.oe");
  co.AddObject(ileee,"ileee","i","ilee.o");

  co.AddObject(callpath,"callpath","i","inhb2.o");
  co.AddObject(callresetpath,"callresetpath","o","callreg.offg.reset-in");
  co.AddObject(callresetdelay19,"callresetdelay19","o","callresetpath.i");

  co.AddObject(callreg2,"callreg2","i0","memcontrol.msbpb.o00");

  co.AddObject(callsigadapt,"callsigadapt","i5","callreg2.o5");
  co.AddObject(callsiggate,"callsiggate","sp0.i","callsigadapt.o0");

  co.AddObject(callpulsegate,"callpulsegate","callsigpath.i","callsigadapt.oc");

  co.AddObject(callstackinput,"callstackinput","csip0.i","callsiggate.gate.o0");

  co.AddObject(counterloadenable,"counterloadenable","i","callpulsegate.callpulsepath.o");
}
