#include "counter2.cpp"
#include "signalpath.cpp"
#include "dtype.cpp"

void CounterMux(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {del {e n}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {del {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.Layer();
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nor {b e}}");
  bo.String("{nul {n e}} {del {n e}} {del {e n}} {nor {b e}}");
  bo.String("{nul {n e}} {nul {n e}} {del {n e}} {nor {b e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nor {b e}}");

  bo.SetOutputs(3,1,3,true);
  bo.SetOutputs(3,1,1,true);
  bo.SetOutputs(2,1,2,true);
  bo.SetOutputs(2,0,2,true);

  bo.Terminal("o",3,1,0,false,back);
  bo.Terminal("a",3,1,3,false,front);
  bo.Terminal("b",1,0,1,true,west);
  bo.Terminal("ae",2,1,2,false,north);
}

// Generate flip flops in an x bit by y word array to use
// for the stack
void FlipFlopArray(CBasicObject &bo, unsigned x, unsigned y)
{
  std::string s0last;
  std::string s0;
  std::string s1;
  std::string snull0,snull1;

  {for(unsigned i = 0; i<y; i++)
  {
    if (i!=0)
    {
      s1 += "{nul {n e}}";
      s0 += "{del {b e}}";
    }

    s0 += "{nor {f e}} {nor {f e} x f} {del {b e}}";
    s1 += "{nor {e n} x e} {nor {e n}} {del {e n}}";

    if (i!=0)
      snull0 += "{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}";
    else
      snull0 += "{del {f e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}";

    snull1 += "{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}";
  }}

  s0last = s0;
  s0last = std::string("{del {s e} b} {del {s e}} {del {s e} b} {del {s e} b} {del {s e} b}") + s0last;
  s0 = std::string("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}") + s0;
  s1 = std::string("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}") + s1;
 
  snull0 = std::string("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}") + snull0;
  snull1 = std::string("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}") + snull1;

  {for(unsigned i = 0; i<x; i++)
  {
    if (i!=0)
    {
      bo.Layer();
      bo.String(snull0.c_str());
      bo.String(snull1.c_str());
    }

    bo.Layer();
    if (i==0)
      bo.String(s0last.c_str());
    else
      bo.String(s0.c_str());
      
    bo.String(s1.c_str());
  }}

  bo.Terminal("out-first",3,2*x-2,0,false,back);

  bo.Terminal("fuse",1,0,0,false,south);
}

void StackMover(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{nul {n e}} {del {n e}} {del {e n}} {nul {n e}}");
  bo.String("{nul {n e}} {del {f e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {del {f e}} {del {e n}} {del {e n}}");
  bo.String("{sli {w s} w} {del {w n}} {del {w n}} {del {f e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.Layer();
  bo.String("{nul {n e}} {nul {n e} f} {nor {f e}} {nul {n e}}");
  bo.String("{nul {n e}} {del {w n}} {nor {w n}} {nul {n e}}");
  bo.String("{ufs {w f}} {del {f e}} {del {e n}} {nul {n e}}");
  bo.String("{fus {w b} w} {del {b e}} {del {f e}} {del {e n}}");
  bo.String("{nul {n e}} {del {w n}} {del {w n}} {del {f e}}");
  bo.Layer();
  bo.String("{nul {n e}} {del {f e}} {del {e n}} {nul {n e}}");
  bo.String("{nul {n e}} {del {f e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {del {f e}} {del {e n}} {del {e n}}");
  bo.String("{sli {w n} w} {del {w n}} {del {w n}} {del {f e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.Layer();
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n w}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");

  bo.Terminal("fuse",0,1,3,false,west);
  bo.Terminal("call-in",1,0,0,true,south);
  bo.Terminal("return-in",1,2,0,true,back);
}

void CounterStack(CCompoundObject &co, unsigned n)
{
  char tmps1[16],tmps2[16];

  CCompoundObject *cnt;
  CCompoundObject *stack;
  CBasicObject *stdt;
  CBasicObject *loadpath;
  CBasicObject *ctodpath;
  CBasicObject *cmux;
  CBasicObject *loadpathor;
  CBasicObject *flipfloparray;
  CBasicObject *stackmover;
  CBasicObject *call2moverpath;
  CBasicObject *return2moverpath;

  cnt = new CCompoundObject();
  stack = new CCompoundObject();
  loadpath = new CBasicObject();

  stdt = new CBasicObject[n];
  ctodpath = new CBasicObject[n];
  cmux = new CBasicObject[n];
  loadpathor = new CBasicObject;
  flipfloparray = new CBasicObject;
  stackmover = new CBasicObject;
  call2moverpath = new CBasicObject;
  return2moverpath = new CBasicObject;


  {for(unsigned i = 0; i<n; i++)
  {
    SmallToggleDType(stdt[i]);
    stdt[i].Rotate(back);
    stdt[i].Rotate(west);
    SignalPath(ctodpath[i],2,7,3,west,back,south,north,south);

    CounterMux(cmux[i]);
    cmux[i].Rotate(west);
    cmux[i].Rotate(north);
  }}

  CounterN(*cnt,n);
  SignalPath(*loadpath,2*n,1,1,west,south,front,east,front,false,true);
  DTypeAdaptorSeries(*stack,n,true);

  stack->Rotate(west);
  stack->Rotate(south);

  FlipFlopArray(*flipfloparray,n,10);
  flipfloparray->Rotate(west);
  flipfloparray->Rotate(north);

  StackMover(*stackmover);
  stackmover->Rotate(west);

  SignalPath(*call2moverpath,1,5,1,south,front,west,north,front,false,false);
  SignalPath(*return2moverpath,2,9,7,west,south,back,east,south,false,false);

  co.AddObject(cnt,"cnt",0,0,0);

  {for(unsigned i = 0; i<n; i++)
  {
    sprintf(tmps1,"stdt%d",i);
    sprintf(tmps2,"cnt.ff%d.r1",i);

    co.AddObject(&(stdt[i]),tmps1,"r",tmps2);
  }}

#if 0
  loadpathor->Layer();
  loadpathor->String("{del {b e}} {nul {n e}}");
  loadpathor->String("{del {w n}} {nor {w n}}");
  loadpathor->String("{nul {n e}} {nul {n e}}");
  loadpathor->String("{nul {n e}} {nul {n e}}");
  loadpathor->Layer();
  loadpathor->String("{nor {f n} x f} {nor {w n}}");
  loadpathor->String("{del {f n} x neswf} {nul {n e}}");
  loadpathor->String("{del {f n} x neswf} {nul {n e}}");
  loadpathor->String("{del {f n} x neswf} {nul {n e}}");

  loadpathor->Terminal("o",0,1,1,false,west);
  loadpathor->Terminal("i",1,1,0,true,south);
  loadpathor->Terminal("iret",1,0,-1,false,east); // input for return signal
#endif
#if 0
  // altered 22/12/2007 to make construction easier
  loadpathor->Layer();
  loadpathor->String("{nul {n e}} {nul {n e}}");
  loadpathor->String("{nul {n e}} {nor {f n}}");
  loadpathor->String("{nul {n e}} {del {f n}}");
  loadpathor->String("{nul {n e}} {nul {n e}}");
  loadpathor->Layer();
  loadpathor->String("{nul {n e}} {nor {f n} x f}");
  loadpathor->String("{del {w n} x nswfb} {nor {w n} x w}");
  loadpathor->String("{del {f n} x neswf} {del {n e}}");
  loadpathor->String("{del {f n} x neswf} {nul {n e}}");

  loadpathor->Terminal("o",0,1,1,false,west);
  loadpathor->Terminal("i",1,1,0,true,south);
  loadpathor->Terminal("iret",1,0,-1,false,east); // input for return signal
#endif
#if 1
  // altered 11/10/2008 altered after other changes
  loadpathor->Layer();
  loadpathor->String("{nul {n e}} {nor {n e} x n}");
  loadpathor->String("{nul {n e}} {nor {s e} x s}");
  loadpathor->String("{nul {n e}} {del {s e}}");
  loadpathor->String("{nul {n e}} {nul {n e}}");
  loadpathor->String("{nul {n e}} {nul {n e}}");
  loadpathor->Layer();
  loadpathor->String("{nul {n e}} {nor {f e}}");
  loadpathor->String("{nul {n e}} {del {f n}}");
  loadpathor->String("{nul {n e}} {del {f n}}");
  loadpathor->String("{nul {n e}} {del {f n}}");
  loadpathor->String("{nul {n e}} {nul {n e}}");
  loadpathor->Layer();
  loadpathor->String("{nul {n e}} {nul {n e}}");
  loadpathor->String("{nul {n e}} {del {n e}}");
  loadpathor->String("{del {w n} x nswfb} {nor {w n} x w}");
  loadpathor->String("{del {f n} x neswf} {del {n e}}");
  loadpathor->String("{del {f n} x neswf} {nul {n e}}");

  loadpathor->Terminal("o",0,2,2,false,west);
  loadpathor->Terminal("iret",1,0,0,true,east);
  loadpathor->Terminal("old",1,0,2,false,south);
#endif

  co.AddObject(loadpath,"loadpath","o","stdt0.ld");
  co.AddObject(loadpathor,"loadpathor","o","loadpath.i");

  {for(unsigned i = 0; i<n; i++)
  {
    sprintf(tmps1,"ctodpath%d",i);
    sprintf(tmps2,"cnt.ff%d.qf",i);

    co.AddObject(&(ctodpath[i]),tmps1,"i",tmps2);

    sprintf(tmps1,"cmux%d",i);
    sprintf(tmps2,"stdt%d.d",i);
    co.AddObject(&(cmux[i]),tmps1,"o",tmps2);

  }}

  co.AddObject(stack,"stack","dtb1.q","cmux0.a");
  co.AddObject(flipfloparray,"flipfloparray","out-first","stack.dtb1.from-setreset");
  co.AddObject(stackmover,"stackmover","fuse","flipfloparray.fuse");
  co.AddObject(call2moverpath,"call2moverpath","o","stackmover.call-in");
  co.AddObject(return2moverpath,"return2moverpath","o","stackmover.return-in");
}
