#ifndef REGISTERS_CPP
#define REGISTERS_CPP

#include "pulsegen.cpp"
#include "memory.cpp"
#include "signalpath.cpp"

// This file contains registers used by the control unit in the 6-bit instruction word architecture.

void SlideRegisterBit(CBasicObject &bo, bool set, bool forward)
{
  bo.Layer();

  if (set && !forward)
  {
    bo.String("{nul {n e}}");
    bo.String("{del {f e} eswb}");
    bo.String("{nor {s e} eswf s}");
    bo.String("{nul {n e}}");
    bo.Layer();
    bo.String("{nul {n e}}");
    bo.String("{nul {n e}}");
    bo.String("{nor {f e} newfb f}");
    bo.String("{nul {n e}}");
  }
  else if (!set && forward)
  {
    bo.String("{nul {n e}}");
    bo.String("{nor {s e} eswb s}");
    bo.String("{del {f e} eswf}");
    bo.String("{nul {n e}}");
    bo.Layer();
    bo.String("{nul {n e}}");
    bo.String("{nor {f e} newfb f}");
    bo.String("{nul {n e}}");
    bo.String("{nul {n e}}");
  }
  else if (!set && !forward)
  {
    bo.String("{del {f e} eswb}");
    bo.String("{nor {s e} eswf s}");
    bo.String("{nul {n e}}");
    bo.String("{nul {n e}}");
    bo.Layer();
    bo.String("{nul {n e}}");
    bo.String("{nor {f e} newfb f}");
    bo.String("{nul {n e}}");
    bo.String("{nul {n e}}");
  }
  else if (set && forward)
  {
    bo.String("{nul {n e}}");
    bo.String("{nul {n e}}");
    bo.String("{nor {s e} eswb s}");
    bo.String("{del {f e} neswf}");
    bo.Layer();
    bo.String("{nul {n e}}");
    bo.String("{nul {n e}}");
    bo.String("{nor {f e} newfb f}");
    bo.String("{nul {n e}}");
  }

  bo.Layer();
  bo.String("{nul {n e}}");
  bo.String("{sli {s f}}");
  bo.String("{sli {s b}}");
  bo.String("{nul {n e}}");
  bo.Layer();
  bo.String("{nul {n e}}");
  bo.String("{del {w n}}");
  bo.String("{nul {n e}}");
  bo.String("{nul {n e}}");

  bo.Terminal("reset-in",0,3,1,true,east);
  bo.Terminal("reset-out",0,3,1,false,back);
  bo.Terminal("set-in",0,3,2,true,north);
  
  if (!forward)
    bo.Terminal("out",0,1,2,false,front);
  else
    bo.Terminal("out",0,1,3,false,front);
}

static void JoinDelta(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{del {b e}}");

  bo.Terminal("i",0,0,0,true,front);
  bo.Terminal("o",0,0,0,false,west);
}

void SlideRegisterInput(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{nor {s e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{del {b e}} {nor {s e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{del {w n}} {del {b e}} {nor {s e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {del {w n}} {del {b e}} {nor {s e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {del {w n}} {del {b e}} {nor {s e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {del {w n}} {del {b e}} {nor {s e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {del {w n}} {del {b e}} {del {s e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {del {w n}} {del {n e}}");

  bo.Layer();
  bo.String("{nor {s e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nor {s e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nor {s e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nor {s e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nor {s e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nor {s e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {del {s e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");

  bo.SetOutputs(0,0,1,true);
  bo.SetOutputs(0,0,2,true);
  bo.SetOutputs(1,0,2,true);
  bo.SetOutputs(1,0,3,true);
  bo.SetOutputs(2,0,3,true);
  bo.SetOutputs(2,0,4,true);
  bo.SetOutputs(3,0,4,true);
  bo.SetOutputs(3,0,5,true);
  bo.SetOutputs(4,0,5,true);
  bo.SetOutputs(4,0,6,true);
  bo.SetOutputs(5,0,6,true);
  bo.SetOutputs(5,0,7,true);
  bo.SetOutputs(6,0,7,true);

  bo.SetOutputs(0,1,0,true);
  bo.SetOutputs(1,1,1,true);
  bo.SetOutputs(2,1,2,true);
  bo.SetOutputs(3,1,3,true);
  bo.SetOutputs(4,1,4,true);
  bo.SetOutputs(5,1,5,true);

  bo.Terminal("enable",6,0,7,true,south);

  bo.Terminal("b0-in",0,1,0,true,north);
  bo.Terminal("b1-in",1,1,1,true,north);
  bo.Terminal("b2-in",2,1,2,true,north);
  bo.Terminal("b3-in",3,1,3,true,north);
  bo.Terminal("b4-in",4,1,4,true,north);
  bo.Terminal("b5-in",5,1,5,true,north);
  bo.Terminal("bg-in",6,1,6,true,north);
}

// extrahold is true if extra parts are needed to hold the guard together.
// e.g. if the west outputs are used, extrahold can be false, otherwise it may need to be set to true
void SlideRegisterGuard(CBasicObject &bo, bool extrahold)
{
  bo.Layer();
  bo.String("{nul {n e}} {del {s e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {del {s e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {s e}} {del {s e}} {del {s e}} {del {s e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {del {s e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {s e}} {del {s e}} {del {s e}} {del {s e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {del {s e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {s e}} {del {s e}} {del {s e}} {del {s e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {del {s e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {del {s e}} {del {s e}} {del {s e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {del {s e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {del {s e}} {nul {n e}}");
  bo.Layer();
  bo.String("{nul {n e}} {del {b e}} {del {b e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{del {b e}} {nul {n e}} {nul {n e}} {del {b e}} {del {b e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {del {f e}} {del {f e}} {nul {n e}} {nul {n e}} {del {b e}} {del {b e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {del {f e}} {del {f e}} {nul {n e}} {nul {n e}} {del {b e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {del {f e}} {del {f e}} {nul {n e}} {del {b e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {del {f e}} {nul {n e}}");
  bo.Layer();
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {del {f e}} {nul {n e}}");

  if (extrahold)
  {
    bo.cells[0][2][0].part = delta;
    bo.cells[0][4][2].part = delta;
    bo.cells[0][6][4].part = delta;
  }

  // 'input' is a misnomer here, since these terminals are simply used for alignment
  bo.Terminal("i4",5,2,8,true,back);
  bo.Terminal("i3",4,2,6,true,back);
  bo.Terminal("i2",3,2,6,true,back);
  bo.Terminal("i1",2,2,4,true,back);
  bo.Terminal("i0",1,2,4,true,back);

  bo.Terminal("o4s",5,0,6,false,south);
  bo.Terminal("o3s",4,0,5,false,south);
  bo.Terminal("o2s",3,0,4,false,south);
  bo.Terminal("o1s",2,0,3,false,south);
  bo.Terminal("o0s",1,0,2,false,south);

  bo.Terminal("o4w",5,0,6,false,west);
  bo.Terminal("o3w",4,0,5,false,west);
  bo.Terminal("o2w",3,0,4,false,west);
  bo.Terminal("o1w",2,0,3,false,west);
  bo.Terminal("o0w",1,0,2,false,west);
}

void GuardToOrientPath(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{del {w n}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{del {w n}} {del {w n}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{del {w n}} {del {w n}} {del {w n}} {nul {n e}} {nul {n e}}");
  bo.String("{del {w n}} {del {w n}} {del {w n}} {del {w n}} {nul {n e}}");
  bo.String("{del {w n}} {del {w n}} {del {w n}} {del {w n}} {del {w n}}");

  bo.Terminal("i0",0,0,0,true,east);
  bo.Terminal("o0",0,0,0,false,west);
  bo.Terminal("i1",1,0,1,true,east);
  bo.Terminal("o1",0,0,1,false,west);
  bo.Terminal("i2",2,0,2,true,east);
  bo.Terminal("o2",0,0,2,false,west);
  bo.Terminal("i3",3,0,3,true,east);
  bo.Terminal("o3",0,0,3,false,west);
  bo.Terminal("i4",4,0,4,true,east);
  bo.Terminal("o4",0,0,4,false,west);
}

static void EnableConnector(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{nor {n e} b n}");
  bo.Layer();
  bo.String("{del {n e} x newfb}");

  bo.Terminal("i",0,0,1,true,back);
  bo.Terminal("o",0,1,0,false,north);
}

// extrahold is true if extra parts are needed to hold the guard together.
// e.g. if the west outputs are used, extrahold can be false, otherwise it may need to be set to true
void SlideRegister(CCompoundObject &co, bool extrahold, bool rightmostbit)
{
  CBasicObject *srb[7];
  CBasicObject *jd[6];
  CBasicObject *ec,*srbi;
  CBasicObject *grd;

  for(unsigned i = 0; i<7; i++)
  {
    srb[i] = new CBasicObject;
    SlideRegisterBit(*(srb[i]),i!=6 || rightmostbit,i%2==0);

    if (i<6)
    {
      jd[i] = new CBasicObject;
      JoinDelta(*(jd[i]));
    }
  }

  ec = new CBasicObject;
  srbi = new CBasicObject;
  grd = new CBasicObject;

  EnableConnector(*ec);
  SlideRegisterInput(*srbi);
  SlideRegisterGuard(*grd,extrahold);

  co.AddObject(srb[6],"srbg",0,0,0);
  co.AddObject(jd[5],"jd5","i","srbg.reset-out");
  co.AddObject(srb[5],"srb5","reset-in","jd5.o");
  co.AddObject(jd[4],"jd4","i","srb5.reset-out");
  co.AddObject(srb[4],"srb4","reset-in","jd4.o");
  co.AddObject(jd[3],"jd3","i","srb4.reset-out");
  co.AddObject(srb[3],"srb3","reset-in","jd3.o");
  co.AddObject(jd[2],"jd2","i","srb3.reset-out");
  co.AddObject(srb[2],"srb2","reset-in","jd2.o");
  co.AddObject(jd[1],"jd1","i","srb2.reset-out");
  co.AddObject(srb[1],"srb1","reset-in","jd1.o");
  co.AddObject(jd[0],"jd0","i","srb1.reset-out");
  co.AddObject(srb[0],"srb0","reset-in","jd0.o");

  co.AddObject(grd,"grd","i0","srb0.out");

  co.AddObject(ec,"ec","i","srbg.out");
  co.AddObject(srbi,"srbi","enable","ec.o");
}

void FlipFlopRegisterBit(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{nul {n e}} {nul {n e}} {del {s e}} {nul {n e}}");
  bo.String("{del {b e}} {nor {f e}} {nor {w n}} {del {w n}}");
  bo.String("{del {w n}} {nor {w n}} {nor {w n}} {nor {w n}}");
  bo.String("{nul {n e}} {nul {n e}} {del {s e}} {nul {n e}}");

  bo.SetOutputs(1,0,1,true);
  bo.SetOutputs(3,0,2,true);
  bo.SetOutputs(2,0,0,true);
  bo.SetOutputs(2,0,3,true);
  
  bo.Terminal("i",3,0,1,true,east);
  bo.Terminal("o",0,0,1,false,west);
}

void FlipFlopRegister(CCompoundObject &co)
{
  char tmps[5];
  
  strcpy(tmps,"bit?");

  CBasicObject *bit[5];
  CBasicObject *load = new CBasicObject;

  for(unsigned i = 0; i<5; i++)
  {
    bit[i] = new CBasicObject;
    FlipFlopRegisterBit(*bit[i]);
 
    tmps[3] = '0' + i;
 
    co.AddObject(bit[i],tmps,0,i,0);
  }

  load->Layer();
  load->String("{del {s e}}");
  load->String("{nor {f n}}");
  load->String("{nor {f n}}");
  load->String("{del {f n}}");
  load->Layer();
  load->String("{nor {s e}}");
  load->String("{nul {n e}}");
  load->String("{nul {n e}}");
  load->String("{nul {n e}}");

  load->SetOutputs(0,1,0,true);
  load->SetOutputs(0,0,0,true);
  load->SetOutputs(0,0,2,true);
  load->SetOutputs(0,0,3,true);

  load->Terminal("i",0,1,0,true,north);

  co.AddObject(load,"load",2,5,0);
}

void RegisterLogic(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nor {s e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nor {b e}} {del {w n}} {del {s e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.Layer();
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nor {s e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {del {s e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.Layer();
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {del {w n}} {del {b n}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {del {b e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {del {w n}} {nor {w n}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {del {b e}} {del {e n}} {del {e n}} {nor {b e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nor {b e}} {nul {n e}} {nul {n e}} {del {b e}}");
  bo.String("{nul {n e}} {nul {n e}} {del {s e}} {nul {n e}} {nor {b e}} {nul {n e}} {nul {n e}} {del {b e}}");
  bo.String("{nul {n e}} {nul {n e}} {nor {f e}} {nor {e n}} {nor {b e}} {nor {f e}} {nul {n e}} {del {b e}}");
  bo.String("{nul {n e}} {nor {e n}} {nor {e n}} {del {e n}} {nul {n e}} {del {f e}} {nul {n e}} {del {b e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nor {s e}} {nul {n e}} {del {b e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {del {s e}}");
  bo.Layer();
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {del {s e}} {nor {e n}} {nor {e n}} {nor {f n}} {nor {w n}} {del {b e}}");
  bo.String("{del {f e}} {nul {n e}} {nor {f n}} {nul {n e}} {nor {b n}} {del {f n}} {del {e n}} {del {b e}}");
  bo.String("{nor {e n}} {del {e n}} {nor {e n}} {del {e n}} {nor {e n}} {del {e n}} {del {e n}} {nor {b e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nor {f e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {del {f e}} {del {e n}} {del {e n}}");
  bo.Layer();
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {del {n e}} {del {e n}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {del {w n}} {del {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {del {f e}} {del {e n}} {del {e n}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nor {s e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}} {nul {n e}}");

  bo.SetOutputs(0,3,7,true);
  bo.SetOutputs(1,3,7,true);
  bo.SetOutputs(2,3,6,true);
  bo.SetOutputs(3,3,5,true);
  bo.SetOutputs(4,3,6,true);
  bo.SetOutputs(4,3,7,true);
  bo.SetOutputs(5,3,7,true);
  bo.SetOutputs(6,3,5,true);
  bo.SetOutputs(6,3,7,true);

  bo.SetOutputs(2,2,6,true);
  bo.SetOutputs(3,2,6,true);
  bo.SetOutputs(4,2,5,true);
  bo.SetOutputs(4,1,7,true);

  bo.SetOutputs(3,0,8,true);
  bo.SetOutputs(3,1,7,true);

  bo.SetOutputs(5,2,6,true);
  bo.SetOutputs(5,2,7,true);

  bo.SetOutputs(5,4,8,true);
  bo.SetOutputs(7,2,3,true);

  bo.Terminal("b5",0,3,6,true,back);
  bo.Terminal("g",2,3,5,true,west);
  bo.Terminal("oa",5,2,0,false,west);
  bo.Terminal("ob",7,3,5,false,east);
  bo.Terminal("oo",3,0,7,false,south);
  bo.Terminal("callo",7,3,9,false,east);
  bo.Terminal("callo2",5,3,9,false,west);
//  bo.Terminal("clraout",6,2,3,false,east);
}

void RegisterStructure(CCompoundObject &co)
{
  CCompoundObject *sidebranch = new CCompoundObject;
  CBasicObject *joindel = new CBasicObject;
  CBasicObject *registerLogic = new CBasicObject;
  CBasicObject *pgB5 = new CBasicObject;
  CBasicObject *pgG = new CBasicObject;
  CCompoundObject *clraoutsig = new CCompoundObject;
  CCompoundObject *clrbsig = new CCompoundObject;
  CCompoundObject *orientloadsig = new CCompoundObject;

  CCompoundObject *areg = new CCompoundObject;
  CCompoundObject *breg = new CCompoundObject;
  CCompoundObject *oreg = new CCompoundObject;

  CBasicObject *gto = new CBasicObject;

  MemorySideBranch(*sidebranch,12);
  ((CBasicObject *)sidebranch->GetObject("sbp5"))->Terminal("ol",2,0,0,false,south);
  ((CBasicObject *)sidebranch->GetObject("sbp6"))->Terminal("ol",2,0,0,false,south);

  ((CBasicObject *)sidebranch->GetObject("sbp0"))->Terminal("o2",5,0,0,false,south);
  ((CBasicObject *)sidebranch->GetObject("sbp1"))->Terminal("o2",5,0,0,false,south);
  ((CBasicObject *)sidebranch->GetObject("sbp2"))->Terminal("o2",5,0,0,false,south);
  ((CBasicObject *)sidebranch->GetObject("sbp3"))->Terminal("o2",5,0,0,false,south);
  ((CBasicObject *)sidebranch->GetObject("sbp4"))->Terminal("o2",5,0,0,false,south);
  ((CBasicObject *)sidebranch->GetObject("sbp5"))->Terminal("o2",5,0,0,false,south);
  ((CBasicObject *)sidebranch->GetObject("sbp6"))->Terminal("o2",5,0,0,false,south);

  joindel->Layer();
  joindel->String("{del {s e}}");
  joindel->Terminal("i",0,0,0,true,north);
  joindel->Terminal("o",0,0,0,false,front);

  RegisterLogic(*registerLogic);
  PulseGen2(*pgB5);
  PulseGen2InBack(*pgG);
  SignalPathJointed(*clraoutsig,"e 1 n 1 e 3");
  SignalPathJointed(*clrbsig,"e 4 b 1 i b 1 i b 2 i b 1 i s 1 w 0");
  SignalPathJointed(*orientloadsig,"s 1 w 8 b 5");

  pgB5->Rotate(west);
  pgB5->Rotate(north);

  pgG->Rotate(west);
  pgG->Rotate(north);

  SlideRegister(*areg,false,true);
  SlideRegister(*breg,true,false);
  FlipFlopRegister(*oreg);
  oreg->Rotate(west);

  GuardToOrientPath(*gto);

  co.AddObject(sidebranch,"sidebranch",0,0,0);
  co.AddObject(joindel,"jd","i","sidebranch.sbp5.ol");
  co.AddObject(pgB5,"pgB5","i","jd.o");
  co.AddObject(pgG,"pgG","i","sidebranch.sbp6.ol");
  co.AddObject(registerLogic,"registerlogic","b5","pgB5.oe");
  co.AddObject(clrbsig,"clrbsig","i","registerlogic.ob");
// TODO below is not needed, so remove it and all associated with it
//  co.AddObject(clraoutsig,"clraoutsig","i","registerlogic.clraout");
  co.AddObject(orientloadsig,"orientloadsig","i","registerlogic.oo");
  co.AddObject(areg,"areg","srbi.b0-in","sidebranch.sbp0.o2");
  co.AddObject(breg,"breg","srbi.b0-in","sidebranch.sbp0.o");
  co.AddObject(gto,"gto","i0","areg.grd.o0w");
  co.AddObject(oreg,"oreg","bit0.i","gto.o0");

  // Now setup the terminal aliases
  // inputs needed: data bits and gate
  // outputs: orientation bits, call bits and 'call' outputs

  co.Terminal("i0","sidebranch.sbp0.i");
  co.Terminal("i1","sidebranch.sbp1.i");
  co.Terminal("i2","sidebranch.sbp2.i");
  co.Terminal("i3","sidebranch.sbp3.i");
  co.Terminal("i4","sidebranch.sbp4.i");
  co.Terminal("i5","sidebranch.sbp5.i");
  co.Terminal("ig","sidebranch.sbp6.i");

  co.Terminal("orient-out0","oreg.bit0.o");
  co.Terminal("orient-out1","oreg.bit1.o");
  co.Terminal("orient-out2","oreg.bit2.o");
  co.Terminal("orient-out3","oreg.bit3.o");
  co.Terminal("orient-out4","oreg.bit4.o");

  co.Terminal("call","registerlogic.callo");
  co.Terminal("call2","registerlogic.callo2");

  co.Terminal("call-out0","areg.grd.o0s");
  co.Terminal("call-out1","areg.grd.o1s");
  co.Terminal("call-out2","areg.grd.o2s");
  co.Terminal("call-out3","areg.grd.o3s");
  co.Terminal("call-out4","areg.grd.o4s");
  co.Terminal("call-out5","breg.grd.o0s");
  co.Terminal("call-out6","breg.grd.o1s");
  co.Terminal("call-out7","breg.grd.o2s");
  co.Terminal("call-out8","breg.grd.o3s");
  co.Terminal("call-out9","breg.grd.o4s");
}

#endif