#ifndef SFLIPFLOP_CPP
#define SFLIPFLOP_CPP

// Signal-based flip-flop with initial state given by i
void SignalFlipFlop(CBasicObject &bo, bool i)
{
  bo.Layer();
  bo.String("{nul {n e}} {del {s e}}");
  bo.String("{del {w n}} {del {f n}}");
  bo.Layer();
  bo.String("{nor {e n}} {del {e n}}");
  bo.String("{nor {e n}} {del {e n}}");
  bo.Layer();
  bo.String("{del {w n}} {del {b n}}");
  bo.String("{nul {n e}} {del {n e}}");

  // Set initial values for all outputs based on value of i

  bo.SetOutputs(0,1,0,i);
  bo.SetOutputs(1,1,0,i);
  bo.SetOutputs(1,0,0,i);
  bo.SetOutputs(1,0,1,i);
  bo.SetOutputs(0,0,1,i);

  bo.SetOutputs(0,1,1,!i);
  bo.SetOutputs(1,1,1,!i);
  bo.SetOutputs(1,2,1,!i);
  bo.SetOutputs(1,2,0,!i);
  bo.SetOutputs(0,2,0,!i);

  bo.Terminal("q",1,1,0,false,east);
  bo.Terminal("qb",1,1,1,false,east);
  bo.Terminal("r",0,1,0,true,west);
  bo.Terminal("s",0,1,1,true,west);
}

// Small Signal-based flip-flop with initial state given by i
// fed by pulses 4 cycles long
void SmallSignalFlipFlop(CBasicObject &bo, bool i)
{
  bo.Layer();
  bo.String("{del {w n}} {nor {w n}}");
  bo.String("{del {f n}} {nor {b n}}");

  // Set initial values for all outputs based on value of i

  bo.SetOutputs(0,0,0,i);
  bo.SetOutputs(0,0,1,i);
  bo.SetOutputs(1,0,1,!i);
  bo.SetOutputs(1,0,0,i);

  bo.Terminal("q",0,0,0,false,west);
  bo.Terminal("q2",0,0,1,false,west);
  bo.Terminal("r",1,0,0,true,east);
  bo.Terminal("r2",1,0,0,true,back);
  bo.Terminal("s",1,0,1,true,east);
  bo.Terminal("s2",1,0,1,true,front);
}

void NegativeEdge10(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{del {b n}} {del {e n}} {del {e n}}");
  bo.String("{del {b n}} {del {w n}} {del {f n}}");
  bo.String("{del {b n}} {del {f n}} {del {e n}}");
  bo.String("{del {b n}} {nul {e n}} {nor {f n} x f}");
  bo.String("{del {e n}} {del {e n}} {nor {e n}}");

  bo.Terminal("i",0,0,4,true,west);
  bo.Terminal("o",2,0,4,false,east);
}

void ToggleFeedback(CBasicObject &bo, bool i)
{
  bo.Layer();
  bo.String("{nul {n e}} {nul {n e}} {del {w n}} {del {w n}} {del {w n}} {del {b n}}");
  bo.String("{nul {n e}} {nul {n e}} {del {f n}} {nul {n e}} {nul {n e}} {del {b n}}");
  bo.String("{nul {n e}} {nul {n e}} {nor {f n}} {nul {n e}} {nul {n e}} {del {b n}}");
  bo.String("{nor {e n}} {del {e n}} {nor {e n}} {nul {n e}} {nul {n e}} {del {e n}}");
  bo.String("{nul {n e}} {del {f n}} {nor {e n}} {nul {n e}} {nul {n e}} {del {e n}}");
  bo.String("{nul {n e}} {nul {n e}} {nor {b n}} {nul {n e}} {nul {n e}} {del {f n}}");
  bo.String("{nul {n e}} {nul {n e}} {del {b n}} {nul {n e}} {nul {n e}} {del {f n}}");
  bo.String("{nul {n e}} {nul {n e}} {del {w n}} {del {w n}} {del {w n}} {del {f n}}");

  bo.SetOutputs(0,0,3,true);
  bo.SetOutputs(1,0,3,true);
  bo.SetOutputs(1,0,4,true);

  bo.SetOutputs(5,0,3,i);
  bo.SetOutputs(5,0,2,i);
  bo.SetOutputs(5,0,1,i);
  bo.SetOutputs(5,0,0,i);
  bo.SetOutputs(4,0,0,i);
  bo.SetOutputs(3,0,0,i);
  bo.SetOutputs(2,0,0,i);
  bo.SetOutputs(2,0,1,i);
  bo.SetOutputs(2,0,2,i);
  bo.SetOutputs(2,0,3,!i);

  bo.SetOutputs(5,0,4,!i);
  bo.SetOutputs(5,0,5,!i);
  bo.SetOutputs(5,0,6,!i);
  bo.SetOutputs(5,0,7,!i);
  bo.SetOutputs(4,0,7,!i);
  bo.SetOutputs(3,0,7,!i);
  bo.SetOutputs(2,0,7,!i);
  bo.SetOutputs(2,0,6,!i);
  bo.SetOutputs(2,0,5,!i);
  bo.SetOutputs(2,0,4,i);

  bo.Terminal("q",4,0,3,false,east);
  bo.Terminal("qb",4,0,4,false,east);
  bo.Terminal("t",0,0,3,true,west);

  // This terminal is for the purpose of aligning with the reset input
  // of the flipflop
  bo.Terminal("ra",2,0,3,false,east);

}

void SmallToggle(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{del {b e}} {nor {e n}} {del {e n}} {nul {n e}}");
  bo.String("{del {w n}} {nor {w n}} {nor {w n}} {del {b e}}");
  bo.String("{del {f n}} {nor {b n}} {nor {w n}} {del {b e}}");
  bo.String("{del {f e}} {nor {e n}} {nor {b e}} {nor {b e}}");


  bo.SetOutputs(3,0,3,true);
  bo.SetOutputs(3,0,2,true);
  bo.SetOutputs(3,0,1,true);
  bo.SetOutputs(1,0,0,true);
  bo.SetOutputs(2,0,0,true);
  bo.SetOutputs(1,0,2,true);
  bo.SetOutputs(1,0,3,true);

  bo.Terminal("t",3,0,3,true,front);
  bo.Terminal("ts",3,0,3,true,south);
  bo.Terminal("q",0,0,3,false,front);

  bo.Terminal("r1",1,0,1,true,north);
  bo.Terminal("s1",1,0,2,true,north);
  bo.Terminal("r2",1,0,1,true,south);
  bo.Terminal("s2",1,0,2,true,south);
}

//An adapter that fits onto SmallToggle above and turns it into a d-type
void SmallToggleDType(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{nul {n e}} {nor {s e}} {nor {w n}}");
  bo.String("{nul {n e}} {nor {s e}} {nor {b e}}");
  bo.String("{nul {n e}} {nor {b e}} {nor {b e}}");
  bo.String("{nul {n e}} {del {b e}} {del {b e}}");
  bo.String("{nul {n e}} {nor {b e}} {nor {b e}}");
  bo.String("{nul {n e}} {nor {b e}} {nor {b e}}");
  bo.String("{nul {n e}} {del {b e}} {nor {b e}}");

  bo.Layer();
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {e n}} {del {e n}} {del {e n}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}}");

  bo.SetOutputs(2,1,4,true);
  bo.SetOutputs(1,1,4,true);
  bo.SetOutputs(0,1,4,true);

  bo.SetOutputs(2,0,6,true);
  bo.SetOutputs(1,0,5,true);

  bo.SetOutputs(1,0,2,true);
  bo.SetOutputs(2,0,2,true);

  bo.SetOutputs(2,0,0,true);

  bo.Terminal("r",1,0,0,false,south);
  bo.Terminal("s",1,0,1,false,south);
  bo.Terminal("d",1,0,6,true,front);
  bo.Terminal("ld",1,1,4,true,west);
}

void NegativeEdge4(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{del {s e}} {del {e n}} {nul {n e}} {nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {del {f n}} {nul {n e}} {del {b e} x neswb} {nor {f e}}");
  bo.String("{nul {n e}} {del {f n}} {nul {n e}} {nor {b e} x b} {nor {f e} x f}");
  bo.String("{nul {n e}} {del {f n}} {del {e n}} {del {e n}} {nor {n e}}");

  bo.Terminal("i",0,0,0,true,north);
  bo.Terminal("o",4,0,3,false,north);
}

void NegativeEdge4Side(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{nul {n e}} {del {b n}} {nor {f e} x f}");
  bo.String("{nul {n e}} {nor {b n}} {nor {f e}}");
  bo.String("{del {e n}} {del {e n}} {nor {e f}}");

  bo.SetOutputs(0,0,2,true);
  bo.SetOutputs(1,0,2,true);
  bo.Terminal("i",0,0,2,true,west);
  bo.Terminal("o",2,0,2,false,east);
}

void PositiveEdge4(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{nul {n e}} {del {b n} x neswb} {nor {f e}}");
  bo.String("{nul {n e}} {nor {b n} x b} {nor {f e} x f}");
  bo.String("{nor {e n} x e} {del {e n}} {nor {e n}}");

  bo.Terminal("i",0,0,2,true,west);
  bo.Terminal("o",2,0,2,false,east);
}

void NegativeEdge2Side(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{del {b n}} {nor {f e} x f}");
  bo.String("{del {e n}} {nor {e f}}");

  bo.Terminal("i",0,0,1,true,west);
  bo.Terminal("o",1,0,1,false,east);
}

void PositiveEdge4Side(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{nul {n e}} {del {b n} x neswb} {nor {f e}}");
  bo.String("{nul {n e}} {nor {b n} x b} {nor {f e} x f}");
  bo.String("{nor {e n}} {del {e n}} {nor {f n}}");

  bo.Terminal("i",0,0,2,true,front);
  bo.Terminal("i2",0,0,2,true,west);
  bo.Terminal("o",2,0,2,false,front);
}

void PositiveEdge2Side(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{nul {n e}} {del {b n} x} {nor {f e} x f}");
  bo.String("{nor {e n}} {del {e n}} {nor {e f}}");

  bo.Terminal("i",0,0,1,true,front);
  bo.Terminal("i2",0,0,1,true,west);
  bo.Terminal("o",2,0,1,false,east);
}

void SmallToggleV2(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{del {b e}} {nor {e n}} {del {e n}} {nul {n e}} {nul {n e}}");
  bo.String("{del {w n}} {nor {w n}} {nor {w n}} {del {b e}} {nul {n e}}");
  bo.String("{del {f n}} {nor {b n}} {nor {w n}} {del {w n}} {nor {w n}}");
  bo.String("{del {f e}} {nor {e n}} {nor {b e}} {nul {n e}} {nul {n e}}");


  bo.SetOutputs(4,0,2,true);
  bo.SetOutputs(3,0,2,true);
  bo.SetOutputs(3,0,1,true);
  bo.SetOutputs(1,0,0,true);
  bo.SetOutputs(2,0,0,true);
  bo.SetOutputs(1,0,2,true);
  bo.SetOutputs(1,0,3,true);

  bo.Terminal("t",4,0,2,true,east);
  bo.Terminal("tf",4,0,2,true,front);
  bo.Terminal("tb",4,0,2,true,back);
  bo.Terminal("tn",4,0,2,true,north);
  bo.Terminal("q",0,0,1,false,west);
  bo.Terminal("q2",0,0,0,false,west);
  bo.Terminal("qb",2,0,0,false,east);
  bo.Terminal("qb2",2,0,0,false,back);
  bo.Terminal("qf",0,0,3,false,front);

  bo.Terminal("r1",1,0,1,true,north);
  bo.Terminal("s1",1,0,2,true,north);
  bo.Terminal("r2",1,0,1,true,south);
  bo.Terminal("s2",1,0,2,true,south);

}

void SmallToggleV3(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{del {b e}} {nor {e n}} {del {e n}} {nul {n e}} {nul {n e}}");
  bo.String("{del {w n}} {nor {w n}} {nor {w n}} {nor {w n}} {del {b e}}");
  bo.String("{del {f n}} {nor {b n}} {nor {w n}} {nor {w n}} {del {s e}}");
  bo.String("{del {f e}} {nor {e n}} {nor {b e}} {nul {n e}} {nul {n e}}");


  bo.SetOutputs(3,0,2,true);
  bo.SetOutputs(3,0,1,true);
  bo.SetOutputs(1,0,0,true);
  bo.SetOutputs(2,0,0,true);
  bo.SetOutputs(1,0,2,true);
  bo.SetOutputs(1,0,3,true);

  bo.Terminal("t",4,0,2,true,north);
  bo.Terminal("q",0,0,1,false,west);
  bo.Terminal("q2",0,0,0,false,west);
  bo.Terminal("qb",2,0,0,false,east);

  bo.Terminal("r1",1,0,1,true,north);
  bo.Terminal("s1",1,0,2,true,north);
  bo.Terminal("r2",1,0,1,true,south);
  bo.Terminal("s2",1,0,2,true,south);
}

void SyncStageN(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}}");
  bo.String("{nor {b e}} {nul {n e}}");
  bo.String("{del {w n}} {del {w n}}");
  bo.Layer();
  bo.String("{nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}}");
  bo.String("{del {s e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}}");
  bo.Layer();
  bo.String("{del {f e}} {nul {n e}}");
  bo.String("{nor {e f}} {nor {f e}}");
  bo.String("{del {e f}} {del {f n}}");
  bo.String("{nul {n e}} {nul {n e}}");

  bo.SetOutputs(0,2,0,true);
  bo.SetOutputs(0,2,2,true);
  bo.SetOutputs(1,2,1,true);
  bo.SetOutputs(1,2,2,true);
  bo.SetOutputs(0,1,2,true);

  bo.SetOutputs(0,0,3,true);
  bo.SetOutputs(1,0,3,true);

  bo.Terminal("out",0,0,2,false,back);
  bo.Terminal("count-in",1,0,3,true,east);
  bo.Terminal("count-out",0,0,3,false,west);
}

void SyncStage0(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}}");
  bo.String("{nor {b e}} {nul {n e}}");
  bo.String("{del {w n}} {del {w n}}");
  bo.Layer();
  bo.String("{nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}}");
  bo.String("{nul {n e}} {nul {n e}}");
  bo.Layer();
  bo.String("{del {f e}} {nul {n e}}");
  bo.String("{del {f e}} {nul {n e}}");
  bo.String("{del {f e}} {del {e n}}");
  bo.String("{nul {n e}} {nul {n e}}");

  bo.SetOutputs(0,2,0,true);
  bo.SetOutputs(0,2,1,true);
  bo.SetOutputs(0,2,2,true);
  bo.SetOutputs(1,2,2,true);

  bo.SetOutputs(0,0,3,true);
  bo.SetOutputs(1,0,3,true);

  bo.Terminal("out",0,0,2,false,back);
  bo.Terminal("count-in",1,0,3,true,east);
  bo.Terminal("count-out",0,0,3,false,west);
}

#endif
