#ifndef SIGNALPATH
#define SIGNALPATH

void SignalPath(CBasicObject &bo, unsigned x, unsigned y, unsigned z,
                Direction d1, Direction d2, Direction d3, Direction id, Direction od, bool inverter = false, bool inverted = false, unsigned ilength = 0)
{
  int sx=0,sy=0,sz=0;
  Direction d[3] = {d1,d2,d3};
  unsigned di = 0;

  if (ilength == 0)
    ilength = x + y + z;

  if (d1 == west || d2 == west || d3 == west)
    sx = x-1;
  if (d1 == south || d2 == south || d3 == south)
    sy = y-1;
  if (d1 == back || d2 == back || d3 == back)
    sz = z-1;

  int px = sx, py = sy, pz = sz;

  std::string tmps;

  for(unsigned k = 0; k<x; k++)
    tmps += "{nul {n e}}";

  for(unsigned i = 0; i<y; i++)
  {
    bo.Layer();

    for(unsigned j = 0; j<z; j++)
    {
      bo.String(tmps.c_str());
    }
  }

  unsigned numdone = 0;

  for(unsigned n=0; n<x+y+z; n++)
  {
    if (bo.cells[py][pz][px].part == nul)
    {
      bo.cells[py][pz][px].part = ((n==0&&inverter)?nor:delta);
      bo.cells[py][pz][px].or1 = ((n==0 && !inverter)?OppositeDir(id):d[di]);
      bo.cells[py][pz][px].or2 = ( (DirectionToY(bo.cells[py][pz][px].or1) == 0) ? north : east);

      numdone++;

      if (inverted)
      {
        bo.SetOutputs(px,py,pz,true);
      }
    }


    px += DirectionToX(d[di]);
    py += DirectionToY(d[di]);
    pz += DirectionToZ(d[di]);

    if (px < 0 || py < 0 || pz < 0 || px >=x || py>=y || pz>=z)
    {
      px -= DirectionToX(d[di]);
      py -= DirectionToY(d[di]);
      pz -= DirectionToZ(d[di]);

      di++;

      if (n==0&&inverter)
      {
        bo.cells[py][pz][px].or1 = d[di];
        bo.cells[py][pz][px].or2 = ( (DirectionToY(bo.cells[py][pz][px].or1) == 0) ? north : east);
        
        if (inverted)
        {
          bo.SetOutputs(px,py,pz,true);
        }
      }

      if (n==x+y+z-1 && bo.cells[py][pz][px].part == nor)
      {
        bo.cells[py][pz][px].or1 = od;
        bo.cells[py][pz][px].or2 = ( (DirectionToY(od) == 0) ? north : east);

        
        if (inverted)
        {
          bo.SetOutputs(px,py,pz,true);
        }
      }
    }

    if (numdone == ilength)
    {
        inverted = !inverted;
        ilength = 0;
    }
  }

  bo.Terminal("i",sx,sy,sz,true,id);
  bo.Terminal("o",px,py,pz,false,od);
}

void Delay19(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{nul {n e}} {nul {n e}} {del {w n}} {del {b n}}");
  bo.String("{nor {w n} x w} {del {b n}} {del {f n}} {del {b n}}");
  bo.String("{nul {n e}} {del {w n}} {del {f n}} {del {s e}}");

  bo.Layer();
  bo.String("{del {e n}} {del {e n}} {del {b n}} {del {e n}}");
  bo.String("{nul {n e}} {del {f n}} {del {b n}} {del {f n}}");
  bo.String("{nul {n e}} {del {f n}} {del {e n}} {del {f n}}");

  bo.Terminal("i",0,1,0,true,west);
  bo.Terminal("o",0,0,1,false,west);

}

void SignalPathBunch(CBasicObject &bo, unsigned x, unsigned z)
{
  std::string tmps;

  {for(unsigned i = 0; i<x; i++)
  {
    tmps += "{del {e n}}";
  }}

  bo.Layer();

  {for(unsigned i = 0; i<z; i++)
  {
    bo.String(tmps.c_str());
  }}

  tmps = "";

  {for(unsigned i = 0; i<z; i++)
  {
    tmps = "ixx";
    tmps[1] = ((z-i-1)/10)+'0';
    tmps[2] = ((z-i-1)%10)+'0';
    bo.Terminal(tmps.c_str(),0,0,i,true,west);
    tmps[0] = 'o';
    bo.Terminal(tmps.c_str(),x-1,0,i,false,east);
  }}
}

// s contains a sequence of the form "e 4 n 3 f 2" etc...
void SignalPathJointed(CCompoundObject &co, char *s, bool inverted = false)
{
  static char tmp[10];
  char *ptr;
  Direction d1,d2,d3,odir;
  int l;
  char curname[2] = "a";
  char lastname[4] = "a.o";
  CBasicObject *bo;
  bool first = true;
  bool invertersection = false;

  while(*s)
  {
    if (*s == 'i')
    {
      invertersection = true;
      s+=2;
    }
    else
    {
      d1 = CharToDirection(*s);
      if (d1 == east || d1 == west)
      {
        d2 = front;
        d3 = north;
      }
      else if (d1 == front || d1 == back)
      {
        d2 = east;
        d3 = north;
      }
      else
      {
        d2 = front;
        d3 = east;
      }

      s++; s++;
      ptr=tmp;
      while(*s && (*s != ' '))
      {
        *ptr = *s;
        ptr++;
        s++;
      }
      *ptr = (char)0;
      if (*s)
      {
        s++;
        if (*s == 'i')
        {
          odir = CharToDirection(*(s+2));
        }
        else
        {
          odir = CharToDirection(*s);
        }
      }
      else
      {
        odir = d1;
      }
  
      l = atoi(tmp);

      bo = new CBasicObject();

      if (l)
      {
        if (invertersection) inverted = !inverted;

        SignalPath(*bo,1+abs(DirectionToX(d1)*(l-1)),
                  1+abs(DirectionToY(d1)*(l-1)),
                  1+abs(DirectionToZ(d1)*(l-1)),
                  d1,d2,d3,OppositeDir(d1),odir,invertersection,inverted);

        invertersection = false;

        if (first)
        {
          co.AddObject(bo,curname,0,0,0);
          first = false;
        }
        else
        {
          co.AddObject(bo,curname,"i",lastname);
          lastname[0]++;
        }
      }

      curname[0]++;
    }
  }

  co.Terminal("i","a.i");
  co.Terminal("o",lastname);
}

#endif
