#include <fstream>
#include <math.h>
#include "objects.h"

#define PI 3.1415926535

static int blockcount[nopart] = {0,0,0,0,0,0,0,0,0,0,0};

static bool defaultconnect = true;

void SetConnect(bool c)
{
  defaultconnect = c;
}

void ShowCount(void)
{
  std::cout << blockcount[1] << " del" << std::endl;
  std::cout << blockcount[2] << " nor" << std::endl;
  std::cout << blockcount[3] << " psh" << std::endl;
  std::cout << blockcount[4] << " sli" << std::endl;
  std::cout << blockcount[5] << " rot" << std::endl;
  std::cout << blockcount[6] << " fus" << std::endl;
  std::cout << blockcount[7] << " ufs" << std::endl;
  std::cout << blockcount[8] << " key" << std::endl;
  std::cout << blockcount[9] << " sig" << std::endl;
  std::cout << blockcount[9] << " sln" << std::endl;
 
  unsigned tot = 0;
  for(unsigned i = 0; i<nopart-1; i++)
  {
    tot += blockcount[i];
  }

  std::cout << tot << " total" << std::endl;
}

Direction RotateFaceAboutAxis(Direction a, Direction f)
{
  static Direction array[6][6] =
  {
    {north,front,south,back,west,east},  // North
    {back,east,front,west,north,south},  // East
    {north,back,south,front,east,west},  // South
    {front,east,back,west,south,north},  // East
    {east,south,west,north,front,back},  // Front
    {west,north,east,south,front,back},  // Back
  };

  return array[a][f];
}

Direction OppositeDir(Direction d)
{
  static Direction array[6] =
  {
    south,west,north,east,back,front
  };

  return array[d];
}

int DirectionToX(Direction p)
{
  switch(p)
  {
    case east:
      return 1;
    case west:
      return -1;
    default:
      return 0;
  }
}

int DirectionToY(Direction p)
{
  switch(p)
  {
    case north:
      return 1;
    case south:
      return -1;
    default:
      return 0;
  }
}

int DirectionToZ(Direction p)
{
  switch(p)
  {
    case front:
      return 1;
    case back:
      return -1;
    default:
      return 0;
  }
}

void CPartCell::Rotate(Direction a)
{
  bool tmpconnect[6],tmpoutput[6];

  or1 = RotateFaceAboutAxis(a,or1);
  or2 = RotateFaceAboutAxis(a,or2);

  {for(unsigned i = 0; i<6; i++)
  {
    tmpconnect[i] = connect[i];
    tmpoutput[i] = output[i];
  }}

  {for(unsigned i = 0; i<6; i++)
  {
    unsigned j = (Direction)RotateFaceAboutAxis(a,(Direction)i);
    connect[j] = tmpconnect[i];
    output[j] = tmpoutput[i];
  }}  
}

void CBasicObject::Layer(void)
{
  cells.push_back(std::vector<std::vector<CPartCell> >());
}

const char *PartToString(Part p)
{
  switch(p)
  {
    case nul:
      return "nul";
      break;
    case delta:
      return "del";
      break;
    case nor:
      return "nor";
      break;
    case push:
      return "psh";
      break;
    case slide:
      return "sli";
      break;
    case rotate:
      return "rot";
      break;
    case fuse:
      return "fus";
      break;
    case unfuse:
      return "ufs";
      break;
    case keypress:
      return "key";
      break;
    case signal:
      return "sig";
      break;
    case slideon:
      return "sln";
      break;
  }
}

Part StringToPart(const char *p)
{
  if (!strncmp(p,"nul",3))
  {
    return nul;
  }
  if (!strncmp(p,"del",3))
  {
    return delta;
  }
  if (!strncmp(p,"nor",3))
  {
    return nor;
  }
  if (!strncmp(p,"psh",3))
  {
    return push;
  }
  if (!strncmp(p,"sli",3))
  {
    return slide;
  }
  if (!strncmp(p,"rot",3))
  {
    return rotate;
  }
  if (!strncmp(p,"fus",3))
  {
    return fuse;
  }
  if (!strncmp(p,"ufs",3))
  {
    return unfuse;
  }
  if (!strncmp(p,"key",3))
  {
    return keypress;
  }
  if (!strncmp(p,"sig",3))
  {
    return signal;
  }
  if (!strncmp(p,"sln",3))
  {
    return slideon;
  }

  return nopart;
}

Direction CharToDirection(char o)
{
  if (o == 'n')
    return north;
  if (o == 'e')
    return east;
  if (o == 's')
    return south;
  if (o == 'w')
    return west;
  if (o == 'f')
    return front;
  if (o == 'b')
    return back;

  return nodirection;
}

const char *StrNChr(const char *s, char c, unsigned n)
{
  while(n)
  {
    if (*s == c)
      return s;

    s++;
    n--;
  }

  return NULL;
}

void SetBoolArray(bool *a, const char *p, unsigned c)
{
  a[north] = a[south] = a[east] = a[west] = a[front] = a[back] = false;

  if (p)
  {
    if (StrNChr(p,'n',c))
      a[north] = true;
    if (StrNChr(p,'e',c))
      a[east] = true;
    if (StrNChr(p,'s',c))
      a[south] = true;
    if (StrNChr(p,'w',c))
      a[west] = true;
    if (StrNChr(p,'f',c))
      a[front] = true;
    if (StrNChr(p,'b',c))
      a[back] = true;
  }
}

void CObject::WriteMem(char *part, char *t, unsigned offset, unsigned bits, unsigned *d, unsigned mask)
{
  if (strlen(part) == 0)
  {
    int x=0,y=0,z=0;
    CTerminal *term;
    bool falseVal = false;
    bool trueVal = true;

    term = GetTerminal(t,x,y,z);

    while(offset + bits)
    {
      int xd,yd,zd;
      Direction dir;
      Part pt;
      CPartCell *pc = GetPartCell(x,y,z);

	  if (!pc) break;

      pt = pc->part;
      dir = pc->or1;

      if (offset)
        offset--;
      else
      {
        SetOutputs(x,y,z,(((*d)&mask)?falseVal:trueVal));   // This is inverted

        bits--;
        d++;
      }

      if (pt == nor)
      {
        falseVal = !falseVal;
        trueVal = !trueVal;
      }

      if (pt == delta)
      {
        x -= DirectionToX(dir);
        y -= DirectionToY(dir);
        z -= DirectionToZ(dir);
      }
      else  // Look for a part with an output facing in this direction
      {
        bool found = false;

        for(unsigned i=0; i<6 && !found; i++)
        {
          if ((Direction)i != dir)
          {
            int xo = x + DirectionToX((Direction)i);
            int yo = y + DirectionToY((Direction)i);
            int zo = z + DirectionToZ((Direction)i);

            pc = GetPartCell(xo,yo,zo);

            if (pc && 
                 (pc->part == delta && pc->or1 != i || 
                  pc->part == nor && pc->or1 == OppositeDir((Direction)i)))
            {
              x = xo; y = yo; z = zo;
              found = true;
            }
          }
        } 
      }
    }
  }
  else
  {
    CObject *o = GetObject(part);

    if (o)
    {
      o->WriteMem("",t,offset,bits,d,mask);
    }
    else
    {
      std::cout << "Unable to find object " << part << " in WriteMem" << std::endl;
    }
  }
}

void CObject::SetOutputs(int x, int y, int z, bool i)
{
  CPartCell *pc;

  pc = GetPartCell(x,y,z);

  for(unsigned j = 0; j<6; j++)
  {
    if (pc)
      pc->output[j] = false;
  }

  if (pc)
    pc->output[pc->or1] = i;

  if (pc && pc->part == delta)
  {
    for(unsigned j = 0; j<6; j++)
    {
      if (j != OppositeDir(pc->or1))
        pc->output[j] = i;
    }
  }
}

void CBasicObject::SetAllOutputs(bool b)
{
  unsigned ni = cells.back().back().size();
  unsigned nj = cells.size();
  unsigned nk = cells.back().size();

  {for(unsigned j = 0; j<nj; j++)
    for(unsigned k = 0; k<nk; k++)
      for(unsigned i = 0; i<ni; i++)
      {
        if (cells[j][k][i].part != nul && cells[j][k][i].part != nopart)
        {
          SetOutputs(i,j,k,b);
        }
      }
  }
}

bool CBasicObject::String(const char *s)
{
  /* Parse the string s, which has the following format:
   * {{ <part> <o1> <o2> } <connectivity> <outputs>} ...
   */

  std::vector<CPartCell> str;
  CPartCell tmp;
  int state = 1;
  const char *sptr = s;
  const char *pstart;
  const char *o1,*o2,*constart,*outstart;
  unsigned concount=0,outcount=0;

  while(*sptr && state != 0)
  {

    switch(state)
    {
      case 1:
        if (*sptr == '{')
        {
          state = 3;
          sptr++;
        }
        else if (isspace(*sptr))
        {
          sptr++;
        }
        else
          state = 0;
        break;
      case 3:
        if (isspace(*sptr))
        {
          sptr++;
        }
        else
        {
          pstart = sptr;
          state = 4;
          sptr++;
        }
        break;
      case 4:
        if (isspace(*sptr))
        {
          state = 5;
          sptr++;
        }
        else
          sptr++;
        break;
      case 5:
        if (isspace(*sptr))
        {
          sptr++;
        }
        else if (*sptr=='{')
        {
          state = 6;
          sptr++;
        }
        else
          state = 0;
        break;
      case 6:
        if (isspace(*sptr))
        {
          sptr++;
        }
        else
        {
          o1 = sptr;
          sptr++;
          state = 7;
        }
        break;
      case 7:
        if (isspace(*sptr))
        {
          sptr++;
          state = 8;
        }
        else
          sptr++;
        break;
      case 8:
        if (isspace(*sptr))
        {
          sptr++;
        }
        else
        {
          o2 = sptr;
          sptr++;
          state = 9;
        }
        break;
      case 9:
        if (isspace(*sptr))
        {
          state = 10;
          sptr++;
        }
        else if (*sptr=='}')
        {
          state = 11;
          sptr++;
        }
        else
          sptr++;
        break;
      case 10:
        if (isspace(*sptr))
        {
          sptr++;
        }
        else if (*sptr=='}')
        {
          state = 11;
          sptr++;
        }
        else
          state = 0;
        break;
      case 11:
        if (isspace(*sptr))
        {
          sptr++;
        }
        else if (*sptr == '}')
        {
          constart = NULL;
          outstart = NULL;
          state = 16;
        }
        else
        {
          constart = sptr;
          concount = 1;
          sptr++;
          state = 12;
        }
        break;
      case 12:
        if (isspace(*sptr))
        {
          state = 13;
          sptr++;
        }
        else if (*sptr=='}')
        {
          outstart = NULL;
          state = 16;
        }
        else
        {
          sptr++;
          concount++;
        }
        break;
      case 13:
        if (isspace(*sptr))
        {
          sptr++;
        }
        else if (*sptr=='}')
        {
          outstart = NULL;
          state = 16;
        }
        else
        {
          outstart = sptr;
          outcount = 1;
          sptr++;
          state = 14;
        }
        break;
      case 14:
        if (isspace(*sptr))
        {
          state = 15;
          sptr++;
        }
        else if (*sptr == '}')
        {
          state = 16;
        }
        else
        {
          outcount++;
          sptr++;
        }
        break;
      case 15:
        if (isspace(*sptr))
        {
          sptr++;
        }
        else if (*sptr == '}')
        {
          state = 16;
        }
        else
          state = 0;
        break;
      case 16:
        /* done it, we now have pointers and lengths to everything that
         * has been parsed */
        tmp.part = StringToPart(pstart);
        tmp.or1 = CharToDirection(*o1);
        tmp.or2 = CharToDirection(*o2);
        SetBoolArray(tmp.connect,constart,concount);
        SetBoolArray(tmp.output,outstart,outcount);

        str.push_back(tmp);

        state = 1;
        sptr++;
        break;
    }
  }

  if (state == 1)
  {
    cells.back().push_back(str);
    return 1;
  }
  else
  {
    std::cout << "Parse Error" << std::endl;
    return 0;
  }
}

bool CBasicObject::NulString(unsigned n)
{
  std::vector<CPartCell> str;
  CPartCell tmp;

  for(unsigned i = 0; i<n; i++)
  {
    str.push_back(tmp);
  }

  cells.back().push_back(str);

  return true;
}

CPartCell *CBasicObject::GetPartCell(int x, int y, int z)
{
  if (x>=0 && x<XSize() && y>=0 && y<YSize() && z>=0 && z<ZSize())
  {
    return &cells[y][z][x];
  }
  else
  {
    return NULL;
  }
}

void CBasicObject::Terminal(const char *n, int x, int y, int z, bool i, Direction f)
{
  terminals.push_back(CTerminal(n,x,y,z,i,f));
}

// Rotate anticlockwise about the specified axis and return the x,y and z
// coords (relative to this object) of the block that was previously
// at 0,0,0
void CBasicObject::Rotate(Direction a, int &xnew, int &ynew, int &znew, int x, int y, int z)
{
  // make a copy of the array with the dimensions swapped
  std::vector<std::vector<std::vector<CPartCell> > > newcells;
  unsigned xdnew,ydnew,zdnew;
  unsigned xdold,ydold,zdold;

  xdold = cells.back().back().size();
  zdold = cells.back().size();
  ydold = cells.size();

  switch(a)
  {
    case north:
    case south:
	  xdnew = zdold;
	  zdnew = xdold;
	  ydnew = ydold;
      break;
    case east:
    case west:
	  ydnew = zdold;
	  zdnew = ydold;
	  xdnew = xdold;
      break;
    case front:
    case back:
	  xdnew = ydold;
	  ydnew = xdold;
	  zdnew = zdold;
      break;
  }

  newcells.resize(ydnew);
  {for(unsigned i = 0; i<ydnew; i++)
  {
    newcells[i].resize(zdnew);
    for(unsigned j = 0; j<zdnew; j++)
    {
      newcells[i][j].resize(xdnew);
    }
  }}

  {for(unsigned i = 0; i<xdnew; i++)
  {
    for(unsigned j = 0; j<ydnew; j++)
    {
      for(unsigned k = 0; k<zdnew; k++)
      {
        switch(a)
        {
          case north:
            newcells[j][k][i] = cells[j][xdnew-i-1][k];
            xnew = zdold - 1;
            ynew = 0;
            znew = 0;
            break;
          case south:
            newcells[j][k][i] = cells[j][i][zdnew-k-1];
            xnew = 0;
            ynew = 0;
            znew = xdold - 1;
            break;
          case east:
            newcells[j][k][i] = cells[zdnew-k-1][j][i];
            xnew = 0;
            ynew = 0;
            znew = ydold - 1;
            break;
          case west:
            newcells[j][k][i] = cells[k][ydnew-j-1][i];
            xnew = 0;
            ynew = zdold - 1;
            znew = 0;
            break;
          case front:
            newcells[j][k][i] = cells[i][k][ydnew-j-1];
            xnew = 0;
            ynew = xdold - 1;
            znew = 0;
            break;
          case back:
            newcells[j][k][i] = cells[xdnew-i-1][k][j];
            xnew = ydold - 1;
            ynew = 0;
            znew = 0;
            break;
        }

        newcells[j][k][i].Rotate(a);
      }
    }
  }}

  cells = newcells;

  // now recalculate terminal positions and faces
  {for(std::vector<CTerminal>::iterator i = terminals.begin(); i!=terminals.end(); i++)
  {
    int tmp;
    switch(a)
    {
      case north:
        tmp = (*i).xpos;
        (*i).xpos = zdold - (*i).zpos - 1;
        (*i).zpos = tmp;
        break;
      case south:
        tmp = (*i).xpos;
        (*i).xpos = (*i).zpos;
        (*i).zpos = xdold - tmp - 1;
        break;
      case east:
        tmp = (*i).ypos;
        (*i).ypos = (*i).zpos;
        (*i).zpos = ydold - tmp - 1;
        break;
      case west:
        tmp = (*i).ypos;
        (*i).ypos = zdold - (*i).zpos - 1;
        (*i).zpos = tmp;
		break;
      case front:
        tmp = (*i).xpos;
        (*i).xpos = (*i).ypos;
        (*i).ypos = xdold - tmp - 1;
        break;
      case back:
        tmp = (*i).xpos;
        (*i).xpos = ydold - (*i).ypos - 1;
        (*i).ypos = tmp;
        break;
    }
    (*i).face = RotateFaceAboutAxis(a, (*i).face);
  }}
}

int CBasicObject::XSize(void)
{
  return cells.back().back().size();
}

int CBasicObject::YSize(void)
{
  return cells.size();
}

int CBasicObject::ZSize(void)
{
  return cells.back().size();
}

// Mirror an object as if the mirror lies in a plane perpendicular to the direction
// a
void CBasicObject::Mirror(Direction a)
{
  unsigned xd,yd,zd;
  bool tmp;
  
  xd = cells.back().back().size();
  zd = cells.back().size();
  yd = cells.size();

  {for(unsigned i = 0; i<yd; i++)
  {
    for(unsigned j = 0; j<zd; j++)
    {
      for(unsigned k = 0; k<xd; k++)
      {
        if (cells[i][j][k].or1 == a)
          cells[i][j][k].or1 = OppositeDir(a);
        else if (cells[i][j][k].or1 == OppositeDir(a))
          cells[i][j][k].or1 = a;

        if (cells[i][j][k].or2 == a)
          cells[i][j][k].or2 = OppositeDir(a);
        else if (cells[i][j][k].or2 == OppositeDir(a))
          cells[i][j][k].or2 = a;

        tmp = cells[i][j][k].connect[a];
        cells[i][j][k].connect[a] = cells[i][j][k].connect[OppositeDir(a)];
        cells[i][j][k].connect[OppositeDir(a)] = tmp;

        tmp = cells[i][j][k].output[a];
        cells[i][j][k].output[a] = cells[i][j][k].output[OppositeDir(a)];
        cells[i][j][k].output[OppositeDir(a)] = tmp;
      }
    }
  }}


  if (a==east || a==west)
  {
    CPartCell tmp;

    {for(unsigned i = 0; i<yd; i++)
    {
      for(unsigned j = 0; j<zd; j++)
      {
        for(unsigned k = 0; k<xd/2; k++)
        {
          tmp = cells[i][j][k];
          cells[i][j][k] = cells[i][j][xd-k-1];
          cells[i][j][xd-k-1] = tmp;
        }
      }
    }}
  }

  if (a==north || a==south)
  {
    std::vector<std::vector<CPartCell> > tmp;

    {for(unsigned i = 0; i<yd; i++)
    {
      tmp = cells[i];
      cells[i] = cells[yd-i-1];
      cells[yd-i-1] = tmp;
    }}
  }

  if (a==front || a==back)
  {
    std::vector<CPartCell> tmp;

    {for(unsigned i = 0; i<yd; i++)
    {
      for(unsigned j = 0; j<zd; j++)
      {
        tmp = cells[i][j];
        cells[i][j] = cells[i][zd-j-1];
        cells[i][zd-j-1] = tmp;
      }
    }}
  }

  {for(std::vector<CTerminal>::iterator i = terminals.begin(); i != terminals.end(); i++)
  {
    if (a==east || a==west)
    {
      (*i).xpos = xd-(*i).xpos-1;
    }

    if (a==north || a==south)
    {
      (*i).ypos = yd-(*i).ypos-1;
    }

    if (a==front || a==back)
    {
      (*i).zpos = zd-(*i).zpos-1;
    }

    if (a==(*i).face || OppositeDir(a)==(*i).face)
      (*i).face = OppositeDir((*i).face);
  }}
}

CTerminal *CBasicObject::GetTerminal(const char *n, int &x, int &y, int &z)
{
  for(std::vector<CTerminal>::iterator i = terminals.begin(); i != terminals.end(); i++)
  {
    if ((*i).name == n)
    {
      x += (*i).xpos;
      y += (*i).ypos;
      z += (*i).zpos;

      return &(*i);
    }
  }

  return NULL;;
}


void CBasicObject::Output(std::ostream &osg, std::ostream &ost, int x, int y, int z)
{
  bool connect[6];

  unsigned ni = cells.back().back().size();
  unsigned nj = cells.size();
  unsigned nk = cells.back().size();

  osg << "record region" << std::endl;

  osg << x << " " << y << " " << z << std::endl;

  osg << ni << " " << nj << " " << nk << std::endl;

  {for(unsigned j = 0; j<nj; j++)
    for(unsigned k = 0; k<nk; k++)
      for(unsigned i = 0; i<ni; i++)
      {
        {for(unsigned l = 0; l<6; l++)
          if (cells[j][k][i].connect[l])
          {
            connect[l] = !defaultconnect;
          }
          else
          {
            connect[l] = defaultconnect;
          }
        }

        if (cells[j][k][i].part != nul && cells[j][k][i].part != nopart)
        {
          blockcount[cells[j][k][i].part]++;
        }

        osg << (unsigned)(cells[j][k][i].part) << std::endl;
        osg << (unsigned)(cells[j][k][i].or1) << std::endl;
        osg << (unsigned)(cells[j][k][i].or2) << std::endl;

        {for(unsigned l = 0; l<6; l++)
          osg << cells[j][k][i].output[l] << " ";
        }
        osg << std::endl;

        {for(unsigned l = 0; l<6; l++)
          osg << connect[l] << " ";
        }
        osg << std::endl;
      }
  }

  for(std::vector<CTerminal>::iterator i = terminals.begin(); i != terminals.end(); i++)
  {
    ost << (*i).name << std::endl;
    ost << (*i).xpos + x << " " << (*i).ypos + y << " " << (*i).zpos + z << std::endl;
    ost << (unsigned)(*i).face << std::endl;
    ost << (((*i).input)?1:0) << std::endl;
  }
}

CObject *CBasicObject::GetObject(const char *n)
{
  if (strlen(n) != 0)
  {
    std::cout << "GetObject called for CBasicObject with non empty string " << n << std::endl;
  }

  return NULL;
}

void CBasicObject::ResolveConnections(CObject *topLevel, int x, int y, int z)
{
  std::cout << "Resolving connections for BO at " << x << " " << y << " " << z << std::endl;
  unsigned ni = cells.back().back().size();
  unsigned nj = cells.size();
  unsigned nk = cells.back().size();

  {for(unsigned j = 0; j<nj; j++)
    for(unsigned k = 0; k<nk; k++)
      for(unsigned i = 0; i<ni; i++)
      {
        if (cells[j][k][i].part != nul)
	{
          for(unsigned l = 0; l<6; l++)
          {
           if (cells[j][k][i].connect[l] != true)
           {
            CPartCell *p = NULL;
	    
            int id = i+DirectionToX((Direction)l);
            int jd = j+DirectionToY((Direction)l);
            int kd = k+DirectionToZ((Direction)l);

	    if (id>=0 && id < ni && jd>=0 && jd < nj && kd>=0 && kd < nk)
            {
		p = &cells[jd][kd][id];

		if (p->part == nul)
		  p = NULL;
	    }
	    
            if (p==NULL)
	    {
	 	p = topLevel->GetPartCell(x+id,y+jd,z+kd);
	    }

            if (p && p->part != nul)
            {
              if (!cells[j][k][i].connect[l] && p->connect[OppositeDir((Direction)l)])
              {
                cells[j][k][i].connect[l] = true;
              }
            }
            else
            {
              cells[j][k][i].connect[l] = true;
            }
           }
          }
	}
      }
  }
}

void DrawLine(std::ostream &os, float x1, float y1, float x2, float y2)
{
	os << "<draw:line draw:style-name=\"gr1\" draw:text-style-name=\"P1\" draw:layer=\"layout\" svg:x1=\"";
	os << x1 << "cm\" svg:y1=\"";
	os << y1 << "cm\" svg:x2=\"";
	os << x2 << "cm\" svg:y2=\"";
	os << y2 << "cm\">";
	os << "<text:p/></draw:line>" << std::endl;
}

void DrawLineRotated(std::ostream &os, float x1, float y1, float x2, float y2, float xo, float yo, float angle)
{
	float x1d = x1-xo;
	float y1d = y1-yo;
	float x2d = x2-xo;
	float y2d = y2-yo;
	
	float x1n = xo + x1d*cos(angle)-y1d*sin(angle);
	float y1n = yo + y1d*cos(angle)+x1d*sin(angle);
	float x2n = xo + x2d*cos(angle)-y2d*sin(angle);
	float y2n = yo + y2d*cos(angle)+x2d*sin(angle);
	
	DrawLine(os,x1n,y1n,x2n,y2n);
}

void DrawLineFraction(std::ostream &os, float x1, float y1, float x2, float y2, float xf1, float yf1, float xf2, float yf2, float angle)
{
	DrawLineRotated(os,xf1*(x2-x1)+x1,yf1*(y2-y1)+y1,xf2*(x2-x1)+x1,yf2*(y2-y1)+y1,(x1+x2)/2.0,(y1+y2)/2.0,angle);
}

float RotatedWidth(float w, float h, float angle)
{
	float wrval = w*cos(angle);
	float hrval = -h*sin(angle);
	
	float rval = wrval;
	if (fabs(hrval) > fabs(wrval)) rval = hrval;
	
	return rval;
}

float RotatedHeight(float w, float h, float angle)
{
	float wrval = w*sin(angle);
	float hrval = h*cos(angle);
	
	float rval = wrval;
	if (fabs(hrval) > fabs(wrval)) rval = hrval;
	
	return rval;
}


void DrawEllipseFraction(std::ostream &os, float x1, float y1, float x2, float y2, float xf, float yf, float rxf, float ryf, float angle, bool filled)
{
	float xo = (x1+x2)/2;
	float yo = (y1+y2)/2;
	
	float xd = xf*(x2-x1)+x1-xo;
	float yd = yf*(y2-y1)+y1-yo;
	
	float x1n = xo + xd*cos(angle)-yd*sin(angle) - RotatedWidth(rxf*(x2-x1),ryf*(y2-y1),angle);
	float y1n = yo + yd*cos(angle)+xd*sin(angle) - RotatedHeight(rxf*(x2-x1),ryf*(y2-y1),angle);
		
	os << "<draw:ellipse draw:style-name=\"";
	os << (filled?"gr3":"gr2");
	os << "\" draw:text-style-name=\"P1\" ";
	os << "draw:layer=\"layout\" ";
	os << "svg:width=\"" << 2.0*rxf*(x2-x1) << "cm\" ";
	os << "svg:height=\"" << 2.0*ryf*(y2-y1) << "cm\" ";
	os << "draw:transform=\"rotate (" << -angle << ")  translate (" << x1n << "cm " << y1n << "cm)\"><text:p/></draw:ellipse>";
	os << std::endl;
}

void DrawCircleFraction(std::ostream &os, float x1, float y1, float x2, float y2, float xf, float yf, float rf, float angle, bool filled)
{
	float xo = (x1+x2)/2;
	float yo = (y1+y2)/2;
	
	float xd = xf*(x2-x1)+x1-xo;
	float yd = yf*(y2-y1)+y1-yo;
	
	float x1n = xo + xd*cos(angle)-yd*sin(angle) - rf*(x2-x1);
	float y1n = yo + yd*cos(angle)+xd*sin(angle) - rf*(y2-y1);
	
	os << "<draw:circle draw:style-name=\"";
	os << (filled?"gr3":"gr2");
	os << "\" draw:text-style-name=\"P1\" ";
	os << "draw:layer=\"layout\" ";
	os << "svg:width=\"" << 2.0*rf*(x2-x1) << "cm\" ";
	os << "svg:height=\"" << 2.0*rf*(y2-y1) << "cm\" ";
	os << "svg:x=\"" << x1n << "cm\" ";
	os << "svg:y=\"" << y1n << "cm\"><text:p/></draw:circle>";
	os << std::endl;
}

int DirToN(Direction d)
{
	int n = 0;
	
	switch(d)
	{
	case east:
		n = 1;
		break;
	case west:
		n = 3;
		break;
	case front:
		n = 2;
		break;
	case back:
		n = 0;
		break;
	}
		
	return n;
}

void DrawWire(std::ostream &os, float x1, float y1, float x2, float y2, Direction pri, Direction sec)
{
	if (pri != north && pri != south)
	{		
		int n = DirToN(pri);
		DrawLineFraction(os,x1,y1,x2,y2,0.5,0,0.5,1,(float)n*PI/2.0);
		DrawLineFraction(os,x1,y1,x2,y2,0.4,0.1,0.5,0,(float)n*PI/2.0);
		DrawLineFraction(os,x1,y1,x2,y2,0.6,0.1,0.5,0,(float)n*PI/2.0);

		DrawLineFraction(os,x1,y1,x2,y2,0,0.5,1,0.5,(float)n*PI/2.0);

		DrawLineFraction(os,x1,y1,x2,y2,0.1,0.4,0,0.5,(float)n*PI/2.0);
		DrawLineFraction(os,x1,y1,x2,y2,0.1,0.6,0,0.5,(float)n*PI/2.0);

		DrawLineFraction(os,x1,y1,x2,y2,0.9,0.4,1,0.5,(float)n*PI/2.0);
		DrawLineFraction(os,x1,y1,x2,y2,0.9,0.6,1,0.5,(float)n*PI/2.0);

	}
	else if (pri == north)
	{
		DrawCircleFraction(os,x1,y1,x2,y2,0.5,0.5,0.1,0,true);
	}
	else if (pri == south)
	{
		DrawCircleFraction(os,x1,y1,x2,y2,0.5,0.5,0.1,0,false);
	}
	
}

void DrawNor(std::ostream &os, float x1, float y1, float x2, float y2, Direction pri, Direction sec)
{
	if (pri != north && pri != south)
	{
		int n = DirToN(pri);
		
		DrawLineFraction(os,x1,y1,x2,y2,0.5,0.2,0.5,1,(float)n*PI/2.0);
		DrawCircleFraction(os,x1,y1,x2,y2,0.5,0.1,0.1,(float)n*PI/2.0,false);

		DrawLineFraction(os,x1,y1,x2,y2,0.3,0.5,0.7,0.5,(float)n*PI/2.0);
	}
	else if (pri == north)
	{
		DrawCircleFraction(os,x1,y1,x2,y2,0.5,0.5,0.1,0,false);
		DrawLineFraction(os,x1,y1,x2,y2,0.3,0.5,0.4,0.5,0);
		DrawLineFraction(os,x1,y1,x2,y2,0.5,0.3,0.5,0.4,0);
		DrawLineFraction(os,x1,y1,x2,y2,0.6,0.5,0.7,0.5,0);
		DrawLineFraction(os,x1,y1,x2,y2,0.5,0.6,0.5,0.7,0);
	}
	else if (pri == south)
	{
		DrawCircleFraction(os,x1,y1,x2,y2,0.5,0.5,0.1,0,false);
		DrawLineFraction(os,x1,y1,x2,y2,0.3,0.5,0.7,0.5,0);
		DrawLineFraction(os,x1,y1,x2,y2,0.5,0.3,0.5,0.7,0);
	}
}

void DrawSli(std::ostream &os, float x1, float y1, float x2, float y2, Direction pri, Direction sec)
{
	if (pri != north && pri != south)
	{
		int n = DirToN(pri);
		
		DrawLineFraction(os,x1,y1,x2,y2,0.5,0.2,0.5,1,(float)n*PI/2.0);
		
		if (sec == south)
		{
			DrawLineFraction(os,x1,y1,x2,y2,0.0,0.2,1.0,0.2,(float)n*PI/2.0);
			DrawLineFraction(os,x1,y1,x2,y2,0.0,0.2,0.2,0.0,(float)n*PI/2.0);
			DrawLineFraction(os,x1,y1,x2,y2,1.0,0.2,0.8,0.0,(float)n*PI/2.0);
		}
		else if (sec == north)
		{
			DrawLineFraction(os,x1,y1,x2,y2,0.2,0.2,0.8,0.2,(float)n*PI/2.0);
			DrawLineFraction(os,x1,y1,x2,y2,0.2,0.2,0.0,0.0,(float)n*PI/2.0);
			DrawLineFraction(os,x1,y1,x2,y2,0.8,0.2,1.0,0.0,(float)n*PI/2.0);
		}
		else if ((pri == back && sec == east) || (pri == east && sec == front) ||
                         (pri == front && sec == west) || (pri == west && sec == back))
		{
			DrawLineFraction(os,x1,y1,x2,y2,0.0,0.4,1.0,0.0,(float)n*PI/2.0);
		}
		else
		{
			DrawLineFraction(os,x1,y1,x2,y2,0.0,0.0,1.0,0.4,(float)n*PI/2.0);
		}
	}
	else if (pri == north)
	{
	    int n = DirToN(sec);
		
		DrawLineFraction(os,x1,y1,x2,y2,0.2,0.1,0.8,0.1,(float)n*PI/2.0);
		DrawLineFraction(os,x1,y1,x2,y2,0.1,0.9,0.9,0.9,(float)n*PI/2.0);
		DrawLineFraction(os,x1,y1,x2,y2,0.2,0.1,0.1,0.9,(float)n*PI/2.0);
		DrawLineFraction(os,x1,y1,x2,y2,0.8,0.1,0.9,0.9,(float)n*PI/2.0);
	}
	else if (pri == south)
	{
	    int n = DirToN(sec);
	
		DrawCircleFraction(os,x1,y1,x2,y2,0.5,0.5,0.1,0,false);		
		DrawLineFraction(os,x1,y1,x2,y2,0.2,0.1,0.8,0.1,(float)n*PI/2.0);
		DrawLineFraction(os,x1,y1,x2,y2,0.1,0.9,0.9,0.9,(float)n*PI/2.0);
		DrawLineFraction(os,x1,y1,x2,y2,0.2,0.1,0.1,0.9,(float)n*PI/2.0);
		DrawLineFraction(os,x1,y1,x2,y2,0.8,0.1,0.9,0.9,(float)n*PI/2.0);
	}
}

void DrawFusUfs(std::ostream &os, float x1, float y1, float x2, float y2, Direction pri, Direction sec, bool filled)
{
	if (pri != north && pri != south)
	{
		int n = DirToN(pri);
				
		if (sec == south)
		{
			DrawLineFraction(os,x1,y1,x2,y2,0.5,0.0,0.5,1,(float)n*PI/2.0);
			DrawLineFraction(os,x1,y1,x2,y2,0.25,0.5,0.5,0.0,(float)n*PI/2.0);
			DrawLineFraction(os,x1,y1,x2,y2,0.75,0.5,0.5,0.0,(float)n*PI/2.0);
		}
		else if (sec == north)
		{
			DrawLineFraction(os,x1,y1,x2,y2,0.5,0.5,0.5,1,(float)n*PI/2.0);
			DrawLineFraction(os,x1,y1,x2,y2,0.5,0.5,0.25,0.0,(float)n*PI/2.0);
			DrawLineFraction(os,x1,y1,x2,y2,0.5,0.5,0.75,0.0,(float)n*PI/2.0);
		}
		else if ((pri == back && sec == east) || (pri == east && sec == front) ||
                         (pri == front && sec == west) || (pri == west && sec == back))
		{
			DrawLineFraction(os,x1,y1,x2,y2,0.5,0.0,0.5,1,(float)n*PI/2.0);
			DrawLineFraction(os,x1,y1,x2,y2,0.5,0.5,1.0,0.0,(float)n*PI/2.0);
		}
		else
		{
			DrawLineFraction(os,x1,y1,x2,y2,0.5,0.0,0.5,1,(float)n*PI/2.0);
			DrawLineFraction(os,x1,y1,x2,y2,0.5,0.5,0.0,0.0,(float)n*PI/2.0);
		}
	}
	else if (pri == north)
	{
	    int n = DirToN(sec);
		
		DrawLineFraction(os,x1,y1,x2,y2,0.5,0.5,0.25,0.0,(float)n*PI/2.0);
		DrawLineFraction(os,x1,y1,x2,y2,0.5,0.5,0.75,0.0,(float)n*PI/2.0);
	}
	else if (pri == south)
	{
	    int n = DirToN(sec);
	
		DrawLineFraction(os,x1,y1,x2,y2,0.25,0.5,0.5,0.0,(float)n*PI/2.0);
		DrawLineFraction(os,x1,y1,x2,y2,0.75,0.5,0.5,0.0,(float)n*PI/2.0);
	}
	
	DrawCircleFraction(os,x1,y1,x2,y2,0.5,0.5,0.1,0,filled);			
}

void DrawRotate(std::ostream &os, float x1, float y1, float x2, float y2, Direction pri, Direction sec)
{
	if (pri != north && pri != south)
	{
		int n = DirToN(pri);
		
		DrawLineFraction(os,x1,y1,x2,y2,0.5,0.2,0.5,1,(float)n*PI/2.0);
		
		DrawEllipseFraction(os,x1,y1,x2,y2,0.5,0.1,0.5,0.1,(float)n*PI/2.0,false);
		DrawLineFraction(os,x1,y1,x2,y2,0.5,0.2,0.4,0.1,(float)n*PI/2.0);
		DrawLineFraction(os,x1,y1,x2,y2,0.5,0.2,0.4,0.3,(float)n*PI/2.0);
	}
	else if (pri == north)
	{
		DrawCircleFraction(os,x1,y1,x2,y2,0.5,0.5,0.1,0,true);		
		DrawCircleFraction(os,x1,y1,x2,y2,0.5,0.5,0.4,0,false);		
		DrawLineFraction(os,x1,y1,x2,y2,0.5,0.1,0.6,0.0,0.0);
		DrawLineFraction(os,x1,y1,x2,y2,0.5,0.1,0.6,0.2,0.0);
	}
	else if (pri == south)
	{
		DrawCircleFraction(os,x1,y1,x2,y2,0.5,0.5,0.1,0,false);		
		DrawCircleFraction(os,x1,y1,x2,y2,0.5,0.5,0.4,0,false);		
		DrawLineFraction(os,x1,y1,x2,y2,0.5,0.1,0.4,0.0,0.0);
		DrawLineFraction(os,x1,y1,x2,y2,0.5,0.1,0.4,0.2,0.0);
	}
}

void DrawPart(std::ostream &os, int x, int y, int z, Part p, Direction pri, Direction sec)
{
	float xorigin = 50;
	float yorigin = 50;
	float scale = 0.5;
#if 1
	float layersepy = 4.5;
	float layersepx = 2;
#else
	float layersepy = 10;
	float layersepx = 2;	
#endif
	float x1 = xorigin+((float)x)*scale - ((float)y)*layersepx;
	float y1 = yorigin+((float)z)*scale - ((float)y)*layersepy;
	float x2 = xorigin+((float)(x+1))*scale - ((float)y)*layersepx;
	float y2 = yorigin+((float)(z+1))*scale - ((float)y)*layersepy;	

	DrawLine(os,x1,y1,x2,y1);
	DrawLine(os,x1,y2,x2,y2);
	DrawLine(os,x1,y1,x1,y2);
	DrawLine(os,x2,y1,x2,y2);
	
	switch(p)
	{
	case delta:
		DrawWire(os,x1,y1,x2,y2,pri,sec);
		break;
	case nor:
		DrawNor(os,x1,y1,x2,y2,pri,sec);
		break;
	case slide:
		DrawSli(os,x1,y1,x2,y2,pri,sec);
		break;
	case fuse:
		DrawFusUfs(os,x1,y1,x2,y2,pri,sec,true);
		break;
	case unfuse:
		DrawFusUfs(os,x1,y1,x2,y2,pri,sec,false);
		break;
	case rotate:
		DrawRotate(os,x1,y1,x2,y2,pri,sec);
		break;
	default:
		break;
	}
}

void CBasicObject::Draw(std::ostream &os, int xo, int yo, int zo)
{
  unsigned ni = cells.back().back().size();
  unsigned nj = cells.size();
  unsigned nk = cells.back().size();

  {for(unsigned j = 0; j<nj; j++)
    for(unsigned k = 0; k<nk; k++)
      for(unsigned i = 0; i<ni; i++)
      {
        if (cells[j][k][i].part != nul)
		  DrawPart(os,i+xo,j+yo,k+zo,cells[j][k][i].part,cells[j][k][i].or1,cells[j][k][i].or2);
      }
  }
}

void CCompoundObject::Draw(std::ostream &os, int xo, int yo, int zo)
{
  for(std::vector<CObjectPosition>::iterator i = objects.begin(); i != objects.end(); i++)
  {
    (*i).object->Draw(os,xo+(*i).xpos,yo+(*i).ypos,zo+(*i).zpos);
  }
}

CObject *CCompoundObject::GetObject(const char *n)
{
  for(std::vector<CObjectPosition>::iterator i = objects.begin(); i!=objects.end(); i++)
  {
    if (!strcmp((*i).name.c_str(),n))
    {
      return (*i).object;
    }
  }

  return NULL;

}

CPartCell *CCompoundObject::GetPartCell(int x, int y, int z)
{
  CPartCell *rval = NULL;

  for(std::vector<CObjectPosition>::iterator i = objects.begin(); i != objects.end(); i++)
  {
	int xs = (*i).object->XSize();
	int ys = (*i).object->YSize();
	int zs = (*i).object->ZSize();

	int xm = (*i).xpos + (*i).object->XMin();
	int ym = (*i).ypos + (*i).object->YMin();
	int zm = (*i).zpos + (*i).object->ZMin();

    if ( xm <= x && xm + xs > x &&
         ym <= y && ym + ys > y &&
         zm <= z && zm + zs > z )
    {
      rval = (*i).object->GetPartCell(x - (*i).xpos, y - (*i).ypos, z - (*i).zpos );
      if (rval && rval->part != nul)
      {
        return rval;
      }  
    }    
  }

  return rval;
}

CTerminal *CCompoundObject::GetTerminal(const char *n, int &x, int &y, int &z)
{
  if (terminalalias.find(n) != terminalalias.end())
  {
    n = terminalalias[n].c_str(); 
  }

  for(std::vector<CObjectPosition>::iterator i = objects.begin(); i!=objects.end(); i++)
  {
    if (!strncmp((*i).name.c_str(),n,(*i).name.size()) && n[(*i).name.size()]=='.')
    {
      x += (*i).xpos;
      y += (*i).ypos;
      z += (*i).zpos;

      return (*i).object->GetTerminal(n+(*i).name.size()+1,x,y,z);
    }
  }

  return NULL;
}

void CCompoundObject::Terminal(const char *a, const char *n)
{
  terminalalias[a] = n;
}

void CCompoundObject::AddObject(CObject *o, const char *n, int x, int y, int z)
{
  objects.push_back(CObjectPosition(o,n,x,y,z));
}

bool CCompoundObject::AddObject(CObject *o, const char *n, const char *t, const char *targt)
{
  CTerminal *thist,*thatt;
  int thisx=0,thisy=0,thisz=0;
  int thatx=0,thaty=0,thatz=0;

  if (thist = GetTerminal(targt,thisx,thisy,thisz))
  {
    if (thatt = o->GetTerminal(t,thatx,thaty,thatz))
    {
      if (thist->face == OppositeDir(thatt->face))
      {
        AddObject(o,n,thisx-thatx-DirectionToX(thatt->face),
                      thisy-thaty-DirectionToY(thatt->face),
                      thisz-thatz-DirectionToZ(thatt->face));

        return true;
      }
      else
      {
        std::cout << "Terminal orientation mismatch for:" << std::endl;
        std::cout << "  " << t << ":" << (int)thatt->face << std::endl;
        std::cout << "  " << targt << ":" << (int)thist->face << std::endl;
        std::cout << "(" << n << ")" << std::endl;
        return false;        
      }
    }
    else
    {
      std::cout << "Terminal not found: " << t << std::endl;
      std::cout << "(" << n << ")" << std::endl;
      return false;
    }
  }
  else
  {
    std::cout << "Terminal not found: " << targt << std::endl;
    std::cout << "(" << n << ")" << std::endl;
    return false;
  }
}

void CCompoundObject::Rotate(Direction a, int &xnew, int &ynew, int &znew, int x, int y, int z)
{
  int xnewr,ynewr,znewr;
  int xd,yd,zd;

  xnew = ynew = znew = 0;
  for(std::vector<CObjectPosition>::iterator i = objects.begin(); i != objects.end(); i++)
  {
    (*i).object->Rotate(a,xnewr,ynewr,znewr,x,y,z);
    xd = (*i).xpos - x;
    yd = (*i).ypos - y;
    zd = (*i).zpos - z;

    switch(a)
    {
      case north:
        (*i).xpos = x - zd;
        (*i).zpos = z + xd;
        break;
      case south:
        (*i).xpos = x + zd;
        (*i).zpos = z - xd;
        break;
      case east:
        (*i).ypos = y + zd;
        (*i).zpos = z - yd;
        break;
      case west:
        (*i).ypos = y - zd;
        (*i).zpos = z + yd;
        break;
      case front:
        (*i).xpos = x + yd;
        (*i).ypos = y - xd;
        break;
      case back:
        (*i).xpos = x - yd;
        (*i).ypos = y + xd;
        break;
    }

    (*i).xpos -= xnewr;
    (*i).ypos -= ynewr;
    (*i).zpos -= znewr;

  }
}

int CCompoundObject::XSize(void)
{
  int min,max;
  bool first = true;

  for(std::vector<CObjectPosition>::iterator i = objects.begin(); i != objects.end(); i++)
  {
    int mintest = i->xpos + i->object->XMin();
    int maxtest = mintest + i->object->XSize() - 1;

    if (first || mintest < min)
    {
      min = mintest;
    }

    if (first || maxtest > max)
    {
      max = maxtest;
    }

	first = false;
  }

  return max-min+1;
}

int CCompoundObject::YSize(void)
{
  int min,max;
  bool first = true;

  for(std::vector<CObjectPosition>::iterator i = objects.begin(); i != objects.end(); i++)
  {
    int mintest = i->ypos + i->object->YMin();
    int maxtest = mintest + i->object->YSize() - 1;

    if (first || mintest < min)
    {
      min = mintest;
    }

    if (first || maxtest > max)
    {
      max = maxtest;
    }

	first = false;
  }

  return max-min+1;
}

int CCompoundObject::ZSize(void)
{
  int min,max;
  bool first = true;

  for(std::vector<CObjectPosition>::iterator i = objects.begin(); i != objects.end(); i++)
  {
    int mintest = i->zpos + i->object->ZMin();
    int maxtest = mintest + i->object->ZSize() - 1;

    if (first || mintest < min)
    {
      min = mintest;
    }

    if (first || maxtest > max)
    {
      max = maxtest;
    }

	first = false;
  }

  return max-min+1;
}

int CCompoundObject::XMin(void)
{
  int min;
  bool first = true;

  for(std::vector<CObjectPosition>::iterator i = objects.begin(); i != objects.end(); i++)
  {
    int mintest = i->xpos + i->object->XMin();

    if (first || mintest < min)
    {
      min = mintest;
    }

    first = false;
  }

  return min;
}

int CCompoundObject::YMin(void)
{
  int min;
  bool first = true;

  for(std::vector<CObjectPosition>::iterator i = objects.begin(); i != objects.end(); i++)
  {
    int mintest = i->ypos + i->object->YMin();

    if (first || mintest < min)
    {
      min = mintest;
    }

    first = false;
  }

  return min;
}

int CCompoundObject::ZMin(void)
{
  int min;
  bool first = true;

  for(std::vector<CObjectPosition>::iterator i = objects.begin(); i != objects.end(); i++)
  {
    int mintest = i->zpos + i->object->ZMin();

    if (first || mintest < min)
    {
      min = mintest;
    }

    first = false;
  }

  return min;
}

void CCompoundObject::Mirror(Direction a)
{
  unsigned xd,yd,zd;

  xd = XSize();
  yd = YSize();
  zd = ZSize();

  for(std::vector<CObjectPosition>::iterator i = objects.begin(); i != objects.end(); i++)
  {
    (*i).object->Mirror(a);

    if (a==east || a==west)
    {
      (*i).xpos = xd - (*i).xpos - (*i).object->XSize();
    }

    if (a==north || a==south)
    {
      (*i).ypos = yd - (*i).ypos - (*i).object->YSize();
    }

    if (a==front || a==back)
    {
      (*i).zpos = zd - (*i).zpos - (*i).object->ZSize();
    }
  }
}

void CCompoundObject::Output(std::ostream &osg, std::ostream &ost, int x, int y, int z)
{
  for(std::vector<CObjectPosition>::iterator i = objects.begin(); i != objects.end(); i++)
  {
    ost << "object " << (*i).name << std::endl;
    ost << "{" << std::endl;

    osg << "record identifier " << (*i).name << std::endl;
    (*i).object->Output(osg,ost,x+(*i).xpos,y+(*i).ypos,z+(*i).zpos);
    osg << "record up" << std::endl;

    ost << "}" << std::endl;
  }
}

void CCompoundObject::ResolveConnections(CObject *topLevel, int x, int y, int z)
{
  for(std::vector<CObjectPosition>::iterator i = objects.begin(); i != objects.end(); i++)
  {
    i->object->ResolveConnections(topLevel,x+i->xpos,y+i->ypos,z+i->zpos);
  }
}
