// cubeview.cpp : implementation of the CCBlocks3DView class
//
// This is a part of the Microsoft Foundation Classes C++ library.
// Copyright (C) 1992-1997 Microsoft Corporation
// All rights reserved.
//
// This source code is only intended as a supplement to the
// Microsoft Foundation Classes Reference and related
// electronic documentation provided with the library.
// See these sources for detailed information regarding the
// Microsoft Foundation Classes product.

#include "stdafx.h"

#include <fstream>
#include <math.h>

#include "MainFrm.h"

#include "CBlocks3D.h"

#include "CBlocks3Ddoc.h"
#include "CBlocks3Dview.h"
#include "SettingsDialog.h"
#include "AnimationDialog.h"

#include "gl\gl.h"
#include "gl\glu.h"

#define PI 3.14159265358979323846264

#ifdef _DEBUG
#undef THIS_FILE
static char BASED_CODE THIS_FILE[] = __FILE__;
#endif

unsigned char threeto8[8] = 
{
	0, 0111>>1, 0222>>1, 0333>>1, 0444>>1, 0555>>1, 0666>>1, 0377
};

unsigned char twoto8[4] = 
{
	0, 0x55, 0xaa, 0xff
};

unsigned char oneto8[2] = 
{
	0, 255
};

static int defaultOverride[13] = 
{
	0, 3, 24, 27, 64, 67, 88, 173, 181, 236, 247, 164, 91
};

static PALETTEENTRY defaultPalEntry[20] = 
{
	{ 0,   0,   0,    0 },
	{ 0x80,0,   0,    0 },
	{ 0,   0x80,0,    0 },
	{ 0x80,0x80,0,    0 },
	{ 0,   0,   0x80, 0 },
	{ 0x80,0,   0x80, 0 },
	{ 0,   0x80,0x80, 0 },
	{ 0xC0,0xC0,0xC0, 0 },

	{ 192, 220, 192,  0 },
	{ 166, 202, 240,  0 },
	{ 255, 251, 240,  0 },
	{ 160, 160, 164,  0 },

	{ 0x80,0x80,0x80, 0 },
	{ 0xFF,0,   0,    0 },
	{ 0,   0xFF,0,    0 },
	{ 0xFF,0xFF,0,    0 },
	{ 0,   0,   0xFF, 0 },
	{ 0xFF,0,   0xFF, 0 },
	{ 0,   0xFF,0xFF, 0 },
	{ 0xFF,0xFF,0xFF, 0 }
};

/////////////////////////////////////////////////////////////////////////////
// CCBlocks3DView

IMPLEMENT_DYNCREATE(CCBlocks3DView, CView)

BEGIN_MESSAGE_MAP(CCBlocks3DView, CView)
	//{{AFX_MSG_MAP(CCBlocks3DView)
	ON_WM_CREATE()
	ON_WM_DESTROY()
	ON_WM_SIZE()
	ON_WM_ERASEBKGND()
	ON_WM_KEYDOWN()
	ON_WM_KEYUP()
	ON_COMMAND(ID_RUN_SETTINGS, OnRunSettings)
	ON_UPDATE_COMMAND_UI(ID_RUN_SETTINGS, OnUpdateRunSettings)
	ON_COMMAND(ID_WRITE_EPS, OnWriteEps)
	ON_COMMAND(ID_TOOLS_SAVEBITMAP, OnToolsSavebitmap)
	ON_COMMAND(ID_TOOLS_ANIMATION, OnToolsAnimation)
	//}}AFX_MSG_MAP
	ON_COMMAND(ID_FILE_SAVETERMINALS, &CCBlocks3DView::OnFileSaveTerminals)
	ON_COMMAND(ID_VIEW_CENTREOFMASS,OnViewCentreOfMass)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CCBlocks3DView construction/destruction

CCBlocks3DView::CCBlocks3DView()
{
	m_pDC = NULL;
	m_pOldPalette = NULL;
	m_play = FALSE;

	xp = 0;
	yp = 0;
	zp = 5.0;

	alt = 0;
	az = 0;

	animSteps = 0;
	animIndex = 0;

}

CCBlocks3DView::~CCBlocks3DView()
{
}

/////////////////////////////////////////////////////////////////////////////
// CCBlocks3DView drawing

void CCBlocks3DView::OnDraw(CDC* pDC)
{
    DrawScene();
	if (GetDocument()->goingslow)
		GetDocument()->stepthread = 0;
}

/////////////////////////////////////////////////////////////////////////////
// CCBlocks3DView diagnostics

#ifdef _DEBUG
void CCBlocks3DView::AssertValid() const
{
	CView::AssertValid();
}

void CCBlocks3DView::Dump(CDumpContext& dc) const
{
	CView::Dump(dc);
}

CCBlocks3DDoc* CCBlocks3DView::GetDocument() // non-debug version is inline
{
	return STATIC_DOWNCAST(CCBlocks3DDoc, m_pDocument);
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CCBlocks3DView message handlers


BOOL CCBlocks3DView::PreCreateWindow(CREATESTRUCT& cs) 
{
    // An OpenGL window must be created with the following flags and must not
    // include CS_PARENTDC for the class style. Refer to SetPixelFormat
    // documentation in the "Comments" section for further information.
    cs.style |= WS_CLIPSIBLINGS | WS_CLIPCHILDREN;
	
	return CView::PreCreateWindow(cs);
}

int CCBlocks3DView::OnCreate(LPCREATESTRUCT lpCreateStruct) 
{
	if (CView::OnCreate(lpCreateStruct) == -1)
		return -1;

	Init();	// initialize OpenGL
	
	return 0;
}

void CCBlocks3DView::OnDestroy() 
{
	HGLRC	hrc;

	KillTimer(1);

	hrc = ::wglGetCurrentContext();

    ::wglMakeCurrent(NULL,  NULL);

    if (hrc)
        ::wglDeleteContext(hrc);

    if (m_pOldPalette)
        m_pDC->SelectPalette(m_pOldPalette, FALSE);

    if (m_pDC)
        delete m_pDC;

	CView::OnDestroy();
}

void CCBlocks3DView::OnSize(UINT nType, int cx, int cy) 
{
	CView::OnSize(nType, cx, cy);
	
    if(cy > 0)
    {    
        glViewport(0, 0, cx, cy);

        if((m_oldRect.right > cx) || (m_oldRect.bottom > cy))
            RedrawWindow();

        m_oldRect.right = cx;
        m_oldRect.bottom = cy;

        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        gluPerspective(45.0f, (GLdouble)cx/cy, 1.0f, 500.0f);
        glMatrixMode(GL_MODELVIEW);
    }
}

/////////////////////////////////////////////////////////////////////////////
// GL helper functions

void CCBlocks3DView::Init()
{
    PIXELFORMATDESCRIPTOR pfd;
    int         n;
	HGLRC		hrc;
	GLfloat		fMaxObjSize, fAspect;
	GLfloat		fNearPlane, fFarPlane;

    m_pDC = new CClientDC(this);

    ASSERT(m_pDC != NULL);

    if (!bSetupPixelFormat())
        return;

    n = ::GetPixelFormat(m_pDC->GetSafeHdc());
    ::DescribePixelFormat(m_pDC->GetSafeHdc(), n, sizeof(pfd), &pfd);

    CreateRGBPalette();

    hrc = wglCreateContext(m_pDC->GetSafeHdc());
    wglMakeCurrent(m_pDC->GetSafeHdc(), hrc);

    GetClientRect(&m_oldRect);
    glClearDepth(1.0f);
    glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	glShadeModel(GL_SMOOTH);

	//glEnable(GL_LINE_SMOOTH);
	glLineWidth(1.0);

	if (m_oldRect.bottom)
		fAspect = (GLfloat)m_oldRect.right/m_oldRect.bottom;
	else	// don't divide by zero, not that we should ever run into that...
		fAspect = 1.0f;
	fNearPlane = 1.0f;
	fFarPlane = 500.0f;
	fMaxObjSize = 3.0f;
	m_fRadius = 13.0f;

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
//    glOrtho(-3.0,3.0,-3.0,3.0,fNearPlane,fFarPlane);
    gluPerspective(45.0f, fAspect, fNearPlane, fFarPlane);
    glMatrixMode(GL_MODELVIEW);
}

BOOL CCBlocks3DView::bSetupPixelFormat()
{
    static PIXELFORMATDESCRIPTOR pfd = 
	{
        sizeof(PIXELFORMATDESCRIPTOR),  // size of this pfd
        1,                              // version number
        PFD_DRAW_TO_WINDOW |            // support window
          PFD_SUPPORT_OPENGL |          // support OpenGL
          PFD_DOUBLEBUFFER,             // double buffered
        PFD_TYPE_RGBA,                  // RGBA type
        24,                             // 24-bit color depth
        0, 0, 0, 0, 0, 0,               // color bits ignored
        0,                              // no alpha buffer
        0,                              // shift bit ignored
        0,                              // no accumulation buffer
        0, 0, 0, 0,                     // accum bits ignored
        32,                             // 32-bit z-buffer
        0,                              // no stencil buffer
        0,                              // no auxiliary buffer
        PFD_MAIN_PLANE,                 // main layer
        0,                              // reserved
        0, 0, 0                         // layer masks ignored
    };
    int pixelformat;

    if ( (pixelformat = ChoosePixelFormat(m_pDC->GetSafeHdc(), &pfd)) == 0 )
    {
        MessageBox("ChoosePixelFormat failed");
        return FALSE;
    }

    if (SetPixelFormat(m_pDC->GetSafeHdc(), pixelformat, &pfd) == FALSE)
    {
        MessageBox("SetPixelFormat failed");
        return FALSE;
    }

    return TRUE;
}

unsigned char CCBlocks3DView::ComponentFromIndex(int i, UINT nbits, UINT shift)
{
    unsigned char val;

    val = (unsigned char) (i >> shift);
    switch (nbits) 
	{

    case 1:
        val &= 0x1;
        return oneto8[val];
    case 2:
        val &= 0x3;
        return twoto8[val];
    case 3:
        val &= 0x7;
        return threeto8[val];

    default:
        return 0;
    }
}


void CCBlocks3DView::CreateRGBPalette()
{
    PIXELFORMATDESCRIPTOR pfd;
    LOGPALETTE *pPal;
    int n, i;
 
    n = ::GetPixelFormat(m_pDC->GetSafeHdc());
    ::DescribePixelFormat(m_pDC->GetSafeHdc(), n, sizeof(pfd), &pfd);

    if (pfd.dwFlags & PFD_NEED_PALETTE)
    {
        n = 1 << pfd.cColorBits;
        pPal = (PLOGPALETTE) new char[sizeof(LOGPALETTE) + n * sizeof(PALETTEENTRY)];

        ASSERT(pPal != NULL);

        pPal->palVersion = 0x300;
        pPal->palNumEntries = n;
        for (i=0; i<n; i++)
        {
            pPal->palPalEntry[i].peRed =
                    ComponentFromIndex(i, pfd.cRedBits, pfd.cRedShift);
            pPal->palPalEntry[i].peGreen =
                    ComponentFromIndex(i, pfd.cGreenBits, pfd.cGreenShift);
            pPal->palPalEntry[i].peBlue =
                    ComponentFromIndex(i, pfd.cBlueBits, pfd.cBlueShift);
            pPal->palPalEntry[i].peFlags = 0;
        }

        /* fix up the palette to include the default GDI palette */
        if ((pfd.cColorBits == 8)                           &&
            (pfd.cRedBits   == 3) && (pfd.cRedShift   == 0) &&
            (pfd.cGreenBits == 3) && (pfd.cGreenShift == 3) &&
            (pfd.cBlueBits  == 2) && (pfd.cBlueShift  == 6)
           )
        {
			for (i = 1 ; i <= 12 ; i++)
                pPal->palPalEntry[defaultOverride[i]] = defaultPalEntry[i];
        }

        m_cPalette.CreatePalette(pPal);
        delete pPal;

        m_pOldPalette = m_pDC->SelectPalette(&m_cPalette, FALSE);
        m_pDC->RealizePalette();
    }
}

void CCBlocks3DView::DrawScene(void)
{
    static BOOL     bBusy = FALSE;
	bool writeAnim = false;

	CCBlocks3DDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);


    if(bBusy) 
		return;
    bBusy = TRUE;

    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glPushMatrix();
		
		glRotatef(az,0,1,0);
		glRotatef(alt,cos(2*PI*az/360.0),0,sin(2*PI*az/360.0));

		glTranslatef(-xp,-yp,-zp);
	
		bool didsuspend = false;

		if (pDoc->suspendthread == 0)
		{
			pDoc->suspendthread = 1;
			while(pDoc->ack == 0)
				Sleep(0);

			didsuspend = true;
		}

		pDoc->U.Draw();

		if (animSteps != 0)
		{
			CUniverse *puni = &(((CCBlocks3DDoc *)(GetDocument()))->U);
			if (animLast + animSteps <= puni->iters && animLast != puni->iters)
			{
				writeAnim = true;
				animLast = puni->iters;
			}
		}


		if (didsuspend)
		{
			pDoc->suspendthread = 0;
			while(pDoc->ack == 1 && pDoc->suspendthread == 0)  // check for suspendthread == 0 here in case the 'run until' limit has been reached
				Sleep(0);
		}

	glPopMatrix();

    glFinish();
    SwapBuffers(wglGetCurrentDC());


    bBusy = FALSE;

	UpdateIterations();

	if (writeAnim)
	{
		CString filename;
			
		filename.Format("%s\\%s%.6d.bmp",animFolder,animPrefix,animIndex);

		WriteViewToDIB(filename);

		animIndex++;
	}
}

BOOL CCBlocks3DView::OnEraseBkgnd(CDC* pDC) 
{
	return TRUE;
}

void CCBlocks3DView::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	// TODO: Add your message handler code here and/or call default
	CCBlocks3DDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);
	
	switch(nChar)
	{
	case 'P':
		az += 10.0;
		break;
	case 'O':
		az -= 10.0;
		break;
	case 'Q':
		alt += 10.0;
		break;
	case 'A':
		alt -= 10.0;
		break;
	case ' ':
		xp += sin(2*PI*az/360)*cos(2*PI*alt/360);
		zp -= cos(2*PI*az/360)*cos(2*PI*alt/360);
		yp -= sin(2*PI*alt/360);
		break;
	case 'B':
		xp -= sin(2*PI*az/360)*cos(2*PI*alt/360);
		zp += cos(2*PI*az/360)*cos(2*PI*alt/360);
		yp += sin(2*PI*alt/360);
		break;
	case 'K':
		pDoc->U.press = true;
		break;
	}

	CView::OnKeyDown(nChar, nRepCnt, nFlags);
}

void CCBlocks3DView::OnKeyUp(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	// TODO: Add your message handler code here and/or call default
	
	CView::OnKeyUp(nChar, nRepCnt, nFlags);
}

void CCBlocks3DView::UpdateIterations(void)
{
	CString str;
	CMainFrame *pFrame = (CMainFrame *)AfxGetApp()->m_pMainWnd;
	CStatusBar *pStatus = &(pFrame->m_wndStatusBar);

	CUniverse *puni = &(((CCBlocks3DDoc *)(GetDocument()))->U);
	if (pStatus)
	{
		str.Format("%d %d %d %d %d %d",puni->activeCount,puni->totalDrawn,puni->iters,puni->lastmovex,puni->lastmovey,puni->lastmovez);
		pStatus->SetPaneText(1,str);
	}
}

void CCBlocks3DView::OnRunSettings() 
{
	CSettingsDialog sd(GetDocument()->uv);

	sd.m_iterations = GetDocument()->rununtil;
	sd.m_alt = alt;
	sd.m_az = az;
	sd.m_x = xp;
	sd.m_y = yp;
	sd.m_z = zp;
	sd.m_description = description;

	if (sd.DoModal() == IDOK)
	{
		GetDocument()->rununtil = sd.m_iterations;
	
		alt = sd.m_alt;
		az = sd.m_az;
		xp = sd.m_x;
		yp = sd.m_y;
		zp = sd.m_z;
		description = sd.m_description;
	}
}

void CCBlocks3DView::OnUpdateRunSettings(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(GetDocument()->suspendthread && !GetDocument()->goingslow);		
}

// This is just like DrawScene, except that it calls a function to do a postscript output rather than graphical output
void CCBlocks3DView::OnWriteEps() 
{
	CCBlocks3DDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);

	std::ofstream os ("c:\\dump.eps");

	glPushMatrix();
		
		glRotatef(az,0,1,0);
		glRotatef(alt,cos(2*PI*az/360.0),0,sin(2*PI*az/360.0));

		glTranslatef(-xp,-yp,-zp);
	
		bool didsuspend = false;

		if (pDoc->suspendthread == 0)
		{
			pDoc->suspendthread = 1;
			while(pDoc->ack == 0)
				Sleep(0);

			didsuspend = true;
		}

		pDoc->U.WriteEPS(os);

		if (didsuspend)
		{
			pDoc->suspendthread = 0;
			while(pDoc->ack == 1 && pDoc->suspendthread == 0)  // check for suspendthread == 0 here in case the 'run until' limit has been reached
				Sleep(0);
		}

	glPopMatrix();

    glFinish();
}

// WriteDIB		- Writes a DIB to file
// Returns		- TRUE on success
// szFile		- Name of file to write to
// hDIB			- Handle of the DIB
BOOL WriteDIB( LPCTSTR szFile, HANDLE hDIB)
{
	BITMAPFILEHEADER	hdr;
	LPBITMAPINFOHEADER	lpbi;

	if (!hDIB)
		return FALSE;

	CFile file;
	if( !file.Open( szFile, CFile::modeWrite|CFile::modeCreate) )
		return FALSE;

	lpbi = (LPBITMAPINFOHEADER)hDIB;

	int nColors = 1 << lpbi->biBitCount;

	// Fill in the fields of the file header 
	hdr.bfType		= ((WORD) ('M' << 8) | 'B');	// is always "BM"
	hdr.bfSize		= GlobalSize (hDIB) + sizeof( hdr );
	hdr.bfReserved1 	= 0;
	hdr.bfReserved2 	= 0;
	hdr.bfOffBits		= (DWORD) (sizeof( hdr ) + lpbi->biSize +
						nColors * sizeof(RGBQUAD));

	// Write the file header 
	file.Write( &hdr, sizeof(hdr) );

	// Write the DIB header and the bits 
	file.Write( lpbi, GlobalSize(hDIB) );

	return TRUE;
}


// DDBToDIB		- Creates a DIB from a DDB
// bitmap		- Device dependent bitmap
// dwCompression	- Type of compression - see BITMAPINFOHEADER
// pPal			- Logical palette
static HANDLE DDBToDIB( CBitmap& bitmap, DWORD dwCompression, CPalette* pPal )
{
	BITMAP			bm;
	BITMAPINFOHEADER	bi;
	LPBITMAPINFOHEADER 	lpbi;
	DWORD			dwLen;
	HANDLE			hDIB;
	HANDLE			handle;
	HDC 			hDC;
	HPALETTE		hPal;


	ASSERT( bitmap.GetSafeHandle() );

	// The function has no arg for bitfields
	if( dwCompression == BI_BITFIELDS )
		return NULL;

	// If a palette has not been supplied use defaul palette
	hPal = (HPALETTE) pPal->GetSafeHandle();
	if (hPal==NULL)
		hPal = (HPALETTE) GetStockObject(DEFAULT_PALETTE);

	// Get bitmap information
	bitmap.GetObject(sizeof(bm),(LPSTR)&bm);

	// Initialize the bitmapinfoheader
	bi.biSize		= sizeof(BITMAPINFOHEADER);
	bi.biWidth		= bm.bmWidth;
	bi.biHeight 		= bm.bmHeight;
	bi.biPlanes 		= 1;
	bi.biBitCount		= bm.bmPlanes * bm.bmBitsPixel;
	bi.biCompression	= dwCompression;
	bi.biSizeImage		= 0;
	bi.biXPelsPerMeter	= 0;
	bi.biYPelsPerMeter	= 0;
	bi.biClrUsed		= 0;
	bi.biClrImportant	= 0;

	// Compute the size of the  infoheader and the color table
	int nColors = (1 << bi.biBitCount);
	if( nColors > 256 )
		nColors = 0;
	dwLen  = bi.biSize + nColors * sizeof(RGBQUAD);

	// We need a device context to get the DIB from
	hDC = GetDC(NULL);
	hPal = SelectPalette(hDC,hPal,FALSE);
	RealizePalette(hDC);

	// Allocate enough memory to hold bitmapinfoheader and color table
	hDIB = GlobalAlloc(GMEM_FIXED,dwLen);

	if (!hDIB){
		SelectPalette(hDC,hPal,FALSE);
		ReleaseDC(NULL,hDC);
		return NULL;
	}

	lpbi = (LPBITMAPINFOHEADER)hDIB;

	*lpbi = bi;

	// Call GetDIBits with a NULL lpBits param, so the device driver 
	// will calculate the biSizeImage field 
	GetDIBits(hDC, (HBITMAP)bitmap.GetSafeHandle(), 0L, (DWORD)bi.biHeight,
			(LPBYTE)NULL, (LPBITMAPINFO)lpbi, (DWORD)DIB_RGB_COLORS);

	bi = *lpbi;

	// If the driver did not fill in the biSizeImage field, then compute it
	// Each scan line of the image is aligned on a DWORD (32bit) boundary
	if (bi.biSizeImage == 0){
		bi.biSizeImage = ((((bi.biWidth * bi.biBitCount) + 31) & ~31) / 8)
						* bi.biHeight;

		// If a compression scheme is used the result may infact be larger
		// Increase the size to account for this.
		if (dwCompression != BI_RGB)
			bi.biSizeImage = (bi.biSizeImage * 3) / 2;
	}

	// Realloc the buffer so that it can hold all the bits
	dwLen += bi.biSizeImage;
	if (handle = GlobalReAlloc(hDIB, dwLen, GMEM_MOVEABLE))
		hDIB = handle;
	else{
		GlobalFree(hDIB);

		// Reselect the original palette
		SelectPalette(hDC,hPal,FALSE);
		ReleaseDC(NULL,hDC);
		return NULL;
	}

	// Get the bitmap bits
	lpbi = (LPBITMAPINFOHEADER)hDIB;

	// FINALLY get the DIB
	BOOL bGotBits = GetDIBits( hDC, (HBITMAP)bitmap.GetSafeHandle(),
				0L,				// Start scan line
				(DWORD)bi.biHeight,		// # of scan lines
				(LPBYTE)lpbi 			// address for bitmap bits
				+ (bi.biSize + nColors * sizeof(RGBQUAD)),
				(LPBITMAPINFO)lpbi,		// address of bitmapinfo
				(DWORD)DIB_RGB_COLORS);		// Use RGB for color table

	if( !bGotBits )
	{
		GlobalFree(hDIB);

		SelectPalette(hDC,hPal,FALSE);
		ReleaseDC(NULL,hDC);
		return NULL;
	}

	SelectPalette(hDC,hPal,FALSE);
	ReleaseDC(NULL,hDC);
	return hDIB;
}




BOOL CCBlocks3DView::WriteViewToDIB( LPCTSTR szFile )
{
	CBitmap 	bitmap;
	CDC 		memDC;
	CRect		rect;

	memDC.CreateCompatibleDC(m_pDC);

	GetClientRect(rect);

	bitmap.CreateCompatibleBitmap(m_pDC, rect.Width(),rect.Height() );

	CBitmap* pOldBitmap = memDC.SelectObject(&bitmap);
	memDC.BitBlt(0, 0, rect.Width(),rect.Height(), m_pDC, 0, 0, SRCCOPY);

	// Create logical palette if device support a palette
	CPalette pal;
	if( m_pDC->GetDeviceCaps(RASTERCAPS) & RC_PALETTE )
	{
		UINT nSize = sizeof(LOGPALETTE) + (sizeof(PALETTEENTRY) * 256);
		LOGPALETTE *pLP = (LOGPALETTE *) new BYTE[nSize];
		pLP->palVersion = 0x300;

		pLP->palNumEntries =
			GetSystemPaletteEntries( *m_pDC, 0, 255, pLP->palPalEntry );

		// Create the palette
		pal.CreatePalette( pLP );

		delete[] pLP;
	}

	memDC.SelectObject(pOldBitmap);

	// Convert the bitmap to a DIB
	HANDLE hDIB = DDBToDIB( bitmap, BI_RGB, &pal );

	if( hDIB == NULL )
		return FALSE;

	// Write it to file
	WriteDIB( szFile, hDIB );

	// Free the memory allocated by DDBToDIB for the DIB
	GlobalFree( hDIB );
	return TRUE;
}

void CCBlocks3DView::OnToolsSavebitmap() 
{
	CFileDialog fdg( false,"bmp",NULL,OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT,"Bitmap Files (*.bmp)|*.bmp||");
	
	if (fdg.DoModal() == IDOK)
	{
		WriteViewToDIB((LPCTSTR)(fdg.GetPathName()));	
	}
}

void CCBlocks3DView::OnToolsAnimation() 
{
	CAnimationDialog adg;

	if (adg.DoModal() == IDOK)
	{
		animFolder = adg.m_folder;
		animPrefix = adg.m_prefix;
		animSteps = atoi(adg.m_steps);
		animIndex = 0;
		animLast = -1;
	}
}

void CCBlocks3DView::OnFileSaveTerminals()
{
	CFileDialog fdg( false,"trm",NULL,OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT,"Terminal Files (*.trm)|*.trm||");
	
	if (fdg.DoModal() == IDOK)
	{
		CCBlocks3DDoc* pDoc = GetDocument();

		std::ofstream os(fdg.GetPathName());

		if (os)
		{
			pDoc->trm->WriteTerminals(os);
		}
	}
}

void CCBlocks3DView::OnViewCentreOfMass()
{
	double x = 0, y = 0, z = 0;
	CCBlocks3DDoc *pDoc = GetDocument();

	if (pDoc)
	{
		pDoc->U.CentreOfMass(x,y,z);

		this->xp = x;
		this->yp = y;
		this->zp = z+50;
		this->alt = 0;
		this->az = 0;
	}
}