#ifndef BLOCK_H
#define BLOCK_H

#include <fstream>

class CUniverseCell;

template<class T, class U, class V>
    struct triple {
    typedef T first_type;
    typedef U second_type;
	typedef V third_type;
    T first;
    U second;
	V third;
    triple();
    triple(const T& x, const U& y, const V& z)
		{first = x; second = y; third = z;}
    template<class W, class X, class Y>
        triple(const triple<W, X, Y>& tr)
		{first = tr.first; second = tr.second; third = tr.third;}
    };

template<class W, class X, class Y>
bool operator<(const triple<W, X, Y> &a,const triple<W, X, Y> &b)
{
	if (a.first < b.first)
		return true;
	else if (a.first > b.first)
		return false;
	else if (a.second < b.second)
		return true;
	else if (a.second > b.second)
		return false;
	else return a.third < b.third;
}

typedef enum {north=0,east,south,west,front,back} Direction;

typedef enum {northmask = 1, eastmask = 2, southmask = 4, westmask = 8, frontmask = 16, backmask = 32};


#ifndef BLOCK_CPP
extern int XDirArray[6];
extern int YDirArray[6];
extern int ZDirArray[6];
extern int OppositeArray[6];
#else
int XDirArray[6] = 
{
	0,1,0,-1,0,0,
};

int YDirArray[6] =
{
	1,0,-1,0,0,0,
};

int ZDirArray[6] =
{
	0,0,0,0,1,-1,
};

int OppositeArray[6] =
{
	south,west,north,east,back,front,
};
#endif

#define DirectionToX(p) (XDirArray[(int)(p)])
#define DirectionToY(p) (YDirArray[(int)(p)])
#define DirectionToZ(p) (ZDirArray[(int)(p)])
#define OppositeDir(p)  ((Direction)OppositeArray[(int)(p)])
/*
int DirectionToX(Direction p);
int DirectionToY(Direction p);
int DirectionToZ(Direction p);
Direction OppositeDir(Direction p);
*/
Direction GetFrontPos(Direction np, Direction ep);

#define MOVE_NORTH 0x00000001
#define MOVE_EAST  0x00000002
#define MOVE_SOUTH 0x00000004
#define MOVE_WEST  0x00000008
#define MOVE_FRONT 0x00000010
#define MOVE_BACK  0x00000020
#define MOVE_BAD   0x80000000

class CRegion;

class CMove
{
public:
	CMove(Direction d) 
		: dir(d), cancel(false)
	{
		dx = DirectionToX(d);
		dy = DirectionToY(d);
		dz = DirectionToZ(d);
	}

	bool SameDirection(CMove *m)
		{return dx == m->dx && dy == m->dy && dz == m->dz;}
	
	Direction dir;
	int dx,dy,dz;
	bool cancel;
	CRegion *region;
};



#define TEMP_BLOCK ((CBlock *)1)

class CBlock;
class CStimulus;

typedef bool (*ProcessFunc)(CBlock *, CUniverseCell *);


class CBlock
{
public:
	typedef enum {empty = 0,
				  delta,
				  nor,
				  push,
				  slide,
				  rotate,
				  fuse,
				  unfuse,
				  keypress,
				  signal,
				  slideon,
				  numtypes} Type;


	CBlock(int xp, int yp, int zp, Type t);
	~CBlock(void);

	bool Process(CUniverseCell *U);
	__forceinline bool GetInput(Direction d);
	__forceinline void SetOutput(Direction d, bool v);
	void ApplyStimulus(Direction d);
	void DoMove(void);
	void RotateAboutAxis(Direction d);
	bool DoRotate(void);
	bool Update(CUniverseCell *U);
	// Will return bits for each connected neighbour
	unsigned Move(CMove *m, CBlock *mb);
	void Rotate(Direction d);

	void ResolveConnections(CUniverseCell *U);
	void SetNeighbours(CUniverseCell *U);
	//void UnsetNeighbours(CUniverseCell *U);

	bool Draw(void);
	void WriteEPS(std::ofstream &os);

	void SetRegion(CRegion *r) {region = r;}
public:
	// The orientation of a block can be specified by saying which direction
	// the face that is north normal is facing, and which direction the face 
	// that is east normal is facing.
	// The two directions must obviously be adjacent, but apart from that
	// are unrestricted
	Direction northpos,eastpos,frontpos;
	Type tp;
	int x,y,z;
	unsigned char buffer;
	unsigned char output;
	CBlock *neighbour[6];
	unsigned char connect;

	CMove *move;
	CBlock *moveinto;

	// The contents of BlockMove are set during 'DoMove' and
	// then copied during update
	struct
	{
		Direction northpos,eastpos,frontpos;
		Type tp;
		unsigned char buffer;
		unsigned char connect;
		CRegion *region;
		CStimulus *stim;
		// TODO - For the moment, no attempt is made to resolve any conflicts that happen when
		// rotating a block (i.e. if two agents try to rotate a block at the same time)
	} blockmove;

	bool torotate;
	Direction rotateaxis;
	CRegion *region;
	CStimulus *stim;

	//ProcessFunc pf;
};

#endif