#include "..\stdafx.h"

#include <assert.h>
#include "Universe.h"

#define MOD(x,y) (((x)+1000000*(y))%(y))

#define DIV(x,y) (((x)+((x)<0))/(y) - ((x)<0))

#define MAX_UNSIGNED 4294967295



CUniverseCell::CUniverseCell(int x, int y, int z, CUniverse *p)
{
	xpos = x;
	ypos = y;
	zpos = z;
	parent = p;

	listbits = 0;

	parent->InsertIntoAttentionNext(this);
	parent->InsertIntoAttentionNow(this);
	parent->InsertIntoDoUpdate(this);
	parent->InsertIntoDoMove(this);

	for(unsigned i = 0; i<CELL_X; i++)
		for(unsigned j = 0; j<CELL_Y; j++)
			for(unsigned k = 0; k<CELL_Z; k++)
				blocks[i][j][k] = new CBlock(x+i,y+j,z+k,CBlock::empty);

	ResolveConnections();
}

CUniverseCell::~CUniverseCell()
{
	for(unsigned i = 0; i<CELL_X; i++)
		for(unsigned j = 0; j<CELL_Y; j++)
			for(unsigned k = 0; k<CELL_Z; k++)
				delete blocks[i][j][k];
}

unsigned CUniverseCell::Process(void)
{
	parent->InsertIntoDoUpdate(this);

	unsigned rval = 0;

	for(unsigned i = 0; i<CELL_X; i++)
		for(unsigned j = 0; j<CELL_Y; j++)
			for(unsigned k = 0; k<CELL_Z; k++)
				if (blocks[i][j][k]->Process(this))
				{
					bool n = (j==CELL_Y-1);
					bool s = (j==0);
					bool e = (i==CELL_X-1);
					bool w = (i==0);
					bool f = (k==CELL_Z-1);
					bool b = (k==0);

					rval |= 
						(n?PROC_N_MASK:0) |
						(s?PROC_S_MASK:0) |
						(e?PROC_E_MASK:0) |
						(w?PROC_W_MASK:0) |
						(f?PROC_F_MASK:0) |
						(b?PROC_B_MASK:0);
				}

	listbits &= ~ATTENTIONNOW_MASK;

	return rval;
}

bool CUniverseCell::DoMove(void)
{
	bool rval = false;

	for(unsigned i = 0; i<CELL_X; i++)
		for(unsigned j = 0; j<CELL_Y; j++)
			for(unsigned k = 0; k<CELL_Z; k++)
			{
				blocks[i][j][k]->DoMove();
				if (blocks[i][j][k]->DoRotate())
					rval = true;
			}

	listbits &= ~DOMOVE_MASK;
	return rval;
}

bool CUniverseCell::Update(void)
{
	bool rval = false;

	for(unsigned i = 0; i<CELL_X; i++)
		for(unsigned j = 0; j<CELL_Y; j++)
			for(unsigned k = 0; k<CELL_Z; k++)
				if (blocks[i][j][k]->Update(this))
					rval = true;

	listbits &= ~DOUPDATE_MASK;

	return rval;
}


void CUniverseCell::AddBlock(CBlock *b)
{
	int i = MOD(b->x,CELL_X);
	int j = MOD(b->y,CELL_Y);
	int k = MOD(b->z,CELL_Z);

	if (blocks[i][j][k]->tp != CBlock::empty)
	{
		char str[256];

		sprintf(str,"Error: UniverseCell at %d %d %d, block %d %d %d "
			    "contains part type %d, replacing with type %d. Regions: %s %s",
				xpos,ypos,zpos,i,j,k,blocks[i][j][k]->tp,b->tp,
				(blocks[i][j][k]->region?(blocks[i][j][k]->region->GetID().c_str()):"?"),
				(b->region?(b->region->GetID().c_str()):"?"));
		AfxMessageBox(str);
	}

	delete blocks[i][j][k];
	blocks[i][j][k] = b;

	b->SetNeighbours(this);
}

// assume that x, y and z are correct
CBlock *CUniverseCell::GetBlock(int x, int y, int z)
{
	assert(x>=xpos && x<xpos + CELL_X &&
		y>=ypos && y<ypos + CELL_Y &&
		z>=zpos && z<zpos + CELL_Z);

	parent->InsertIntoDoMove(this);
	parent->InsertIntoDoUpdate(this);
	parent->InsertIntoAttentionNext(this);

	return blocks[MOD(x,CELL_X)][MOD(y,CELL_Y)][MOD(z,CELL_Z)];
}

bool CUniverseCell::GetSignalValue(int x, int y, int z, Direction d)
{
	if (x>=xpos && x<xpos + CELL_X &&
		y>=ypos && y<ypos + CELL_Y &&
		z>=zpos && z<zpos + CELL_Z)
	{
		return blocks[MOD(x,CELL_X)][MOD(y,CELL_Y)][MOD(z,CELL_Z)]->output & (1<<d);
	}
	else
	{
		return parent->GetSignalValue(x,y,z,d);
	}
}

// This function will try to move the cell at x,y,z
unsigned CUniverseCell::Move(int x, int y, int z, CMove *m)
{
	parent->InsertIntoDoMove(this);
	parent->InsertIntoDoUpdate(this);
	parent->InsertIntoAttentionNext(this);

	if (x>=xpos && x<xpos + CELL_X &&
		y>=ypos && y<ypos + CELL_Y &&
		z>=zpos && z<zpos + CELL_Z)
	{
		int mbx = x-m->dx;
		int mby = y-m->dy;
		int mbz = z-m->dz;
		CBlock *mb;

		if (mbx>=xpos && mbx<xpos + CELL_X &&
			mby>=ypos && mby<ypos + CELL_Y &&
			mbz>=zpos && mbz<zpos + CELL_Z)
		{
			mb = blocks[MOD(mbx,CELL_X)][MOD(mby,CELL_Y)][MOD(mbz,CELL_Z)];
		}
		else
		{
			mb = parent->GetBlock(mbx,mby,mbz);
		}

		return blocks[MOD(x,CELL_X)][MOD(y,CELL_Y)][MOD(z,CELL_Z)]->Move(m,mb);
	}
	else
	{
		return parent->Move(x,y,z,m);
	}
}

void CUniverseCell::Save(std::ofstream &os)
{
	os << xpos << " " << ypos << " " << zpos << std::endl;
	os << CELL_X << " " << CELL_Y << " " << CELL_Z << " " << std::endl;

	for(unsigned j = 0; j<CELL_Y; j++)
		for(unsigned k = 0; k<CELL_Z; k++)
			for(unsigned i = 0; i<CELL_X; i++)
			{
				os << (int)(blocks[i][j][k]->tp) << std::endl;
				os << (int)(blocks[i][j][k]->northpos) << std::endl;
				os << (int)(blocks[i][j][k]->eastpos) << std::endl;

				{for(unsigned l = 0; l<6; l++)
				{
					os << (int)!!(blocks[i][j][k]->output & (1<<l)) << std::endl;
				}}
				{for(unsigned l = 0; l<6; l++)
				{
					os << (int)(blocks[i][j][k]->connect & (1<<l)) << std::endl;
				}}
			}
}

void CUniverseCell::ResolveConnections(void)
{
	for(unsigned i = 0; i<CELL_X; i++)
		for(unsigned j = 0; j<CELL_Y; j++)
			for(unsigned k = 0; k<CELL_Z; k++)
				blocks[i][j][k]->ResolveConnections(this);
}

void CUniverseCell::SetNeighbours(void)
{
	for(unsigned i = 0; i<CELL_X; i++)
		for(unsigned j = 0; j<CELL_Y; j++)
			for(unsigned k = 0; k<CELL_Z; k++)
				blocks[i][j][k]->SetNeighbours(this);
}

bool CUniverseCell::Unused(void)
{
	bool used = false;

	for(unsigned i = 0; i<CELL_X; i++)
		for(unsigned j = 0; j<CELL_Y; j++)
			for(unsigned k = 0; k<CELL_Z; k++)
				if (blocks[i][j][k]->tp != CBlock::empty)
					used = true;

	return !used;
}

void CUniverseCell::FindOrigin(std::vector<triple<int, int, int> > &candidates)
{
	for(unsigned i = 0; i<CELL_X; i++)
		for(unsigned j = 0; j<CELL_Y; j++)
			for(unsigned k = 0; k<CELL_Z; k++)
			{
				if (blocks[i][j][k]->tp == CBlock::rotate &&
					blocks[i][j][k]->northpos == north)
				{
					candidates.push_back(triple<int,int,int>(blocks[i][j][k]->x,
															 blocks[i][j][k]->y,
															 blocks[i][j][k]->z));
				}
			}
}

void CUniverseCell::CentreOfMass(double &x, double &y, double &z, unsigned &n)
{
	for(unsigned i = 0; i<CELL_X; i++)
		for(unsigned j = 0; j<CELL_Y; j++)
			for(unsigned k = 0; k<CELL_Z; k++)
			{
				if (blocks[i][j][k]->tp != CBlock::empty)
				{
					x += (double)xpos + (double)i;
					y += (double)ypos + (double)j;
					z += (double)zpos + (double)k;
					n++;
				}	
			}

}

CUniverse::CUniverse(void) : press(false)
{
#if 0
	CBlock *nb;

	nb = new CBlock(0,0,0,CBlock::nor);

	AddBlock(nb);

	nb = new CBlock(0,1,0,CBlock::fuse);

	AddBlock(nb);

	nb = new CBlock(0,2,0,CBlock::delta);

	AddBlock(nb);

	nb = new CBlock(1,2,0,CBlock::delta);

	AddBlock(nb);

	nb = new CBlock(1,1,0,CBlock::slide);

	AddBlock(nb);

	nb = new CBlock(1,0,0,CBlock::delta);

	AddBlock(nb);

	nb = new CBlock(1,-1,0,CBlock::nor);

	AddBlock(nb);
#endif

	iters = 0;
	totalDrawn = 0;
	lastmovex = lastmovey = lastmovez = 0;

	attentionNow = &l1;
	attentionNext = &l2;

	moveView = false;
	moveViewDir = Direction::north;
}

CUniverse::~CUniverse()
{
	{for(std::list<CRegion *>::iterator i = regions.begin(); i!=regions.end(); i++)
	{
		delete (*i);
	}}
}

void CUniverse::AddBlock(CBlock *b)
{
	int mx = DIV(b->x,CELL_X);
	int my = DIV(b->y,CELL_Y);
	int mz = DIV(b->z,CELL_Z);

	CUniverseCell *uc = cells.Find(mx,my,mz);

	if (uc)
	{
		uc->AddBlock(b);
		MarkCells(mx,my,mz,PROC_ALL_MASK,false);
	}
	else
	{
		uc = new CUniverseCell(mx*CELL_X,my*CELL_Y,mz*CELL_Z,this);
		cells.Add(mx,my,mz,uc);
		uc->AddBlock(b);

		MarkCells(mx,my,mz,PROC_ALL_MASK,false);
	}
}

CBlock *CUniverse::GetBlock(int x, int y, int z)
{
	int mx = DIV(x,CELL_X);
	int my = DIV(y,CELL_Y);
	int mz = DIV(z,CELL_Z);

	CUniverseCell *uc = cells.Find(mx,my,mz);

	if (uc)
	{
		return uc->GetBlock(x,y,z);
	}
	else
	{
		uc = new CUniverseCell(mx*CELL_X,my*CELL_Y,mz*CELL_Z,this);
		cells.Add(mx,my,mz,uc);
		return uc->GetBlock(x,y,z);
	}
}

CBlock *CUniverse::GetBlockNoAdd(int x, int y, int z)
{
	int mx = DIV(x,CELL_X);
	int my = DIV(y,CELL_Y);
	int mz = DIV(z,CELL_Z);

	CUniverseCell *uc = cells.Find(mx,my,mz);

	if (uc)
	{
		return uc->GetBlock(x,y,z);
	}
	else
	{
		return NULL;
	}
}

bool CUniverse::GetSignalValue(int x, int y, int z, Direction d)
{
	int mx = DIV(x,CELL_X);
	int my = DIV(y,CELL_Y);
	int mz = DIV(z,CELL_Z);

	CUniverseCell *uc = cells.Find(mx,my,mz);

	if (uc)
	{
		uc->GetSignalValue(x,y,z,d);
	}
	else
	{
		return false;
	}
}

void CUniverse::ApplyStimulus(int x, int y, int z, Direction f)
{
	x += DirectionToX(f);
	y += DirectionToY(f);
	z += DirectionToZ(f);

	int mx = DIV(x,CELL_X);
	int my = DIV(y,CELL_Y);
	int mz = DIV(z,CELL_Z);

	GetBlock(x,y,z)->ApplyStimulus(OppositeDir(f));

	MarkCells(mx,my,mz,PROC_ALL_MASK,false);
}

void CUniverse::MarkCells(int x, int y, int z, unsigned procmask, bool nextOrNow)
{
	for(int i = -1; i<=1; i++)
		for(int j = ((i==0)?-1:0); j<=((i==0)?1:0); j++)
			for(int k = ((i==0 && j==0)?-1:0); k<=((i==0 && j==0)?1:0); k++)
			{
					bool dont = false;

					dont = dont || (j==-1 && !(procmask & PROC_S_MASK));
					dont = dont || (j==1 && !(procmask & PROC_N_MASK));
					dont = dont || (i==-1 && !(procmask & PROC_W_MASK));
					dont = dont || (i==1 && !(procmask & PROC_E_MASK));
					dont = dont || (k==-1 && !(procmask & PROC_B_MASK));
					dont = dont || (k==1 && !(procmask & PROC_F_MASK));
						
					if (!dont)
					{
						CUniverseCell *uc = cells.Find(x+i,y+j,z+k);

						if (uc)
						{
							if (nextOrNow)
							{
								InsertIntoAttentionNext(uc);
							}
							else
							{
								InsertIntoAttentionNow(uc);
							}
						}
					}
			}
}

void CUniverse::Process(void)
{
	unsigned tmpActiveCount = 0;

	for(std::list<CUniverseCell *>::iterator i = attentionNow->begin(); i!=attentionNow->end(); i++)
	{
		unsigned procmask;

		if ((procmask = (*i)->Process()))
		{
			MarkCells(DIV((*i)->xpos,CELL_X),DIV((*i)->ypos,CELL_Y),DIV((*i)->zpos,CELL_Z),procmask);
		}

		tmpActiveCount++;
	}

	attentionNow->clear();

	activeCount = tmpActiveCount;
}

void CUniverse::DoMove(void)
{
	for(std::list<CUniverseCell *>::iterator i = doMove.begin(); i!=doMove.end(); i++)
	{
		if ((*i)->DoMove())
		{
			MarkCells(DIV((*i)->xpos,CELL_X),DIV((*i)->ypos,CELL_Y),DIV((*i)->zpos,CELL_Z),PROC_ALL_MASK);
		}
	}

	doMove.clear();
	assert(doMove.size() == 0);
}

void CUniverse::Update(void)
{
	for(std::list<CUniverseCell *>::iterator i = doUpdate.begin(); i!=doUpdate.end(); i++)
	{
		if ((*i)->Update())
		{
			MarkCells(DIV((*i)->xpos,CELL_X),DIV((*i)->ypos,CELL_Y),DIV((*i)->zpos,CELL_Z),PROC_ALL_MASK);
		}
	}

	doUpdate.clear();
}

void CUniverse::Clean(void)
{
	while(moves.size())
	{
		delete moves.front();

		moves.pop_front();
	}

	for(std::list<CUniverseCell *>::iterator i = attentionNext->begin(); i!=attentionNext->end(); i++)
	{
		(*i)->listbits |= ATTENTIONNOW_MASK;
		(*i)->listbits &= ~ATTENTIONNEXT_MASK;
	}

	std::list<CUniverseCell *> *tmp = attentionNow;
	attentionNow = attentionNext;
	attentionNext = tmp;

	press = false;
	iters++;
#if 1
	if (iters % 10000 == 0)
	{
		// Delete any empty universecells

		cells.Clean();
	}
#endif
}

// This function will try to move the cell at x,y,z
unsigned CUniverse::Move(int x, int y, int z, CMove *m)
{
	int mx = DIV(x,CELL_X);
	int my = DIV(y,CELL_Y);
	int mz = DIV(z,CELL_Z);

	CUniverseCell *uc = cells.Find(mx,my,mz);

	if (uc)
	{
		return uc->Move(x,y,z,m);
	}
	else
	{
		uc = new CUniverseCell(mx*CELL_X,my*CELL_Y,mz*CELL_Z,this);
		cells.Add(mx,my,mz,uc);
		return uc->Move(x,y,z,m);
	}
}

// Call this function to try to move the block at x,y,z in the
// specified direction. This function will work out moves for
// all connected blocks, or blocks that are in the way in the
// direction of the move.
void CUniverse::MakeMove(int x, int y, int z, Direction d)
{
	if (GetBlock(x,y,z)->tp == CBlock::empty)
		return;

	lastmovex = x;
	lastmovey = y;
	lastmovez = z;

	int mx,my,mz;
	unsigned mval;
	unsigned mask;
	CMove *m = new CMove(d);
	moves.push_back(m);
	std::list< triple<int,int,int> > todo;

	todo.push_back(triple<int,int,int>(x,y,z));

	while(todo.size())
	{
		mx = todo.front().first;
		my = todo.front().second;
		mz = todo.front().third;

		mval =  Move(mx,my,mz,m);

		if (mval & MOVE_BAD)
		{
			m->cancel = true;
			break;
		}

		mask = 0x00000001;
		for(Direction i = north; i<=back; i=(Direction)(i+1))
		{
			if (mval & mask)
			{
				todo.push_back(triple<int,int,int>(
					mx+DirectionToX(i),
					my+DirectionToY(i),
					mz+DirectionToZ(i)
					));
			}

			mask <<= 1;
		}

		todo.pop_front();
	}
}

void CUniverse::Rotate(int x, int y, int z, Direction d)
{
	CBlock *b = GetBlock(x,y,z);
	if (b->tp == CBlock::empty)
		return;

	b->Rotate(d);
}

void CUniverse::Connect(int x, int y, int z, Direction d, bool c)
{
	CBlock *b1 = GetBlock(x,y,z);
	CBlock *b2 = GetBlock(x+DirectionToX(d),y+DirectionToY(d),z+DirectionToZ(d));

	if (b1->tp != CBlock::empty && b2->tp != CBlock::empty)
	{
		if (c)
		{
			b1->connect |= (1<<d);
			b2->connect |= (1<<OppositeDir(d));
		}
		else
		{
			b1->connect &= ~(1<<d);
			b2->connect &= ~(1<<OppositeDir(d));
		}
		
	}
}

void CUniverse::InsertIntoAttentionNext(CUniverseCell *c)
{
	if (!(c->listbits & ATTENTIONNEXT_MASK))
	{
		attentionNext->push_back(c);
		c->listbits |= ATTENTIONNEXT_MASK;
	}
}

void CUniverse::InsertIntoAttentionNow(CUniverseCell *c)
{
	if (!(c->listbits & ATTENTIONNOW_MASK))
	{
		attentionNow->push_back(c);
		c->listbits |= ATTENTIONNOW_MASK;
	}
}

void CUniverse::InsertIntoDoMove(CUniverseCell *c)
{
	if (!(c->listbits & DOMOVE_MASK))
	{
#if 0
		// Test to see whether it is already in there
		for(std::list<CUniverseCell *>::iterator i = doMove.begin(); i != doMove.end(); i++)
		{
			if ((*i) == c)
			{
				CUniverseCell *r = (*i);
				assert(0);
			}
		}
#endif

		doMove.push_back(c);
		c->listbits |= DOMOVE_MASK;
	}
}

void CUniverse::InsertIntoDoUpdate(CUniverseCell *c)
{
	if (!(c->listbits & DOUPDATE_MASK))
	{
		doUpdate.push_back(c);
		c->listbits |= DOUPDATE_MASK;
	}
}

void CUniverse::Load(std::ifstream &is)
{
	static char token[128];
	int xd,yd,zd;
	int xp,yp,zp;
	int tp;
	int np,ep;
	int tmp;
	bool readregion;
	CRegion *currentregion = NULL;

	while(is)
	{
	  is >> token;

	  // If the token read is 'record', this is a new style file format and one of several
	  // types of record will be read. Otherwise it is an old style file format and the token
	  // must be the x coordintate of a region.

	  if (!strcmp(token,"record"))
	  {
		  is >> token;

		  if (!strcmp(token,"region"))
		  {
			  is >> xp;
			  readregion = true;
		  }
		  else if (!strcmp(token,"identifier"))
		  {
			  is >> token;
			  currentregion = new CRegion(token,currentregion);

			  regions.push_back(currentregion);
			  readregion = false;
		  }
		  else if (!strcmp(token,"up"))
		  {
			  if (currentregion)
				currentregion = currentregion->GetParent();

			  readregion = false;
		  }
	  }
	  else
	  {
		  xp = atoi(token);
		  readregion = true;
	  }

	  if (readregion)
	  {
		  is >> yp >> zp;
		  is >> xd >> yd >> zd;

		  if (is)
		  {
			  for(unsigned i = 0; i<yd; i++)
			  {
				for(unsigned j = 0; j<zd; j++)
				{
					for(unsigned k = 0; k<xd; k++)
					{
						is >> tp;
						is >> np >> ep;

						CBlock *b=NULL;

						if (tp != CBlock::empty)
						{
						  b = new CBlock(k+xp,i+yp,j+zp,(CBlock::Type)tp);
						  b->northpos = (Direction)np;
						  b->eastpos = (Direction)ep;
						  b->frontpos = GetFrontPos((Direction)np,(Direction)ep);
						}

						if (b) b->output = 0;

						{for(unsigned l = 0; l<6; l++)
						{
						  is >> tmp;
						  if (b && tmp) b->output |= (1<<l);
						}}
						{for(unsigned l = 0; l<6; l++)
						{
						  is >> tmp;
						  if (b && tmp && tp != CBlock::empty)
						  {
 							b->connect |= (1<<l);
						  }
						  else
						  {
						    if (b) b->connect &= ~(1<<l);
						  }
						}}
						
						if(b)
						{
							if (currentregion)
								b->SetRegion(currentregion);
							AddBlock(b);
						}
					}
				}
			  }
		  }
		}
	}

	ResolveConnections();
	SetNeighbours();
}

void CUniverse::Save(std::ofstream &os)
{
	cells.Save(os);
}

void CUniverse::ResolveConnections(void)
{
	cells.ResolveConnections();
}

void CUniverse::SetNeighbours(void)
{
	cells.SetNeighbours();
}

void CUniverse::WriteWordToMemoryBlock(int x, int y, int z, unsigned pos, unsigned word)
{
	bool inverted = !(pos < 7);

	for(int i=0; i<pos ;i++)
	{
		if (i < 7)
			z--;

		if (i>=7 && i<31+7)
			y--;

		if (i>=31+7 && i<31+7+7)
			z++;

		for(int j=0; j<15; j++)
		{
			if (i==31+7+7 + j*14)
				y++;

			if (i>=31+7+7+1 + j*14 && i<31+7+7+1+6 + j*14)
				z--;

			if (i==31+7+7+1+6 + j*14)
				y++;

			if (i>=31+7+7+1+6+1 + j*14 && i<31+7+7+1+6+1+6 + j*14)
				z++;
		}
	}

	int dhgl=0;

	for(int xo = 0; xo<6; xo++)
	{
		bool set = !!(word & (1<<xo));

		if (inverted)
			set = !set;

		CBlock *b = GetBlockNoAdd(x+xo,y,z);

		if (b && b->tp == CBlock::nor)
		{
			if (set)
			{
				b->output = (1<<b->northpos);
			}
			else
			{
				b->output = 0;
			}
		}
		else if (b && b->tp == CBlock::delta)
		{
			if (set)
			{
				b->output = ~(1<<OppositeDir(b->northpos));
			}
			else
			{
				b->output = 0;
			}
		}
		else if (b)
		{
			assert(0);
		}
	}
}

void CUniverse::WriteToMemory(int x, int y, int z, int offset, std::vector<unsigned> &data)
{
	int blocks = ((data.size()-1) / 256)+1;

	for(unsigned i = 0; i<blocks; i++)
	{
		for(unsigned ix = 0; ix < 256; ix ++)
		{
			unsigned dataindex;

			unsigned word = 0;

			for(unsigned j = 0; j<6; j++)
			{
				dataindex = i*256+(512-ix+offset+i*6+j)%256;

				if (dataindex < data.size())
				{
					word |= data[dataindex]&(1<<j);
				}
			}

			WriteWordToMemoryBlock(x+i*6,y,z,ix,word);
		}
	}
}

void CUniverse::FindOrigin(int &x, int &y, int &z)
{
	std::vector<triple<int,int,int> > candidates;

	cells.FindOrigin(candidates);

	bool first = true;

	for(std::vector<triple<int,int,int> >::iterator i = candidates.begin(); i != candidates.end(); i++)
	{
		if (first)
		{
			x = i->first;
			y = i->second;
			z = i->third;
		}
		else
		{
			if (i->third > z)
			{
				x = i->first;
				y = i->second;
				z = i->third;
			}
			else if (i->third == z)
			{
				if (i->second < y)
				{
					x = i->first;
					y = i->second;
					z = i->third;
				}
			}
		}

		first = false;
	}
}

void CUniverse::CentreOfMass(double &x, double &y, double &z)
{
	cells.CentreOfMass(x,y,z);
}