// SettingsDialog.cpp : implementation file
//

#include "stdafx.h"
#include "CBlocks3D.h"
#include "SettingsDialog.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CSettingsDialog dialog
CSettingsDialog::CSettingsDialog(CUniverseViews *u, CView *pView)
{
	m_pView = pView;
	uv = u;
}

CSettingsDialog::CSettingsDialog(CUniverseViews *u, CWnd* pParent /*=NULL*/)
	: CDialog(CSettingsDialog::IDD, pParent)
{
	m_pView = NULL;
	uv = u;

	//{{AFX_DATA_INIT(CSettingsDialog)
	m_iterations = 0;
	m_alt = 0.0f;
	m_az = 0.0f;
	m_description = _T("");
	m_views = _T("");
	m_x = 0.0f;
	m_y = 0.0f;
	m_z = 0.0f;
	//}}AFX_DATA_INIT
}

BOOL CSettingsDialog::Create()
{
	return CDialog::Create(CSettingsDialog::IDD);
}

void CSettingsDialog::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CSettingsDialog)
	DDX_Text(pDX, IDC_ITERATIONS, m_iterations);
	DDV_MinMaxUInt(pDX, m_iterations, 0, 1000000000);
	DDX_Text(pDX, IDC_ALT, m_alt);
	DDX_Text(pDX, IDC_AZ, m_az);
	DDX_Text(pDX, IDC_DESCRIPTION, m_description);
	DDX_LBString(pDX, IDC_VIEWS, m_views);
	DDX_Text(pDX, IDC_X, m_x);
	DDX_Text(pDX, IDC_Y, m_y);
	DDX_Text(pDX, IDC_Z, m_z);
	//}}AFX_DATA_MAP

	if (!pDX->m_bSaveAndValidate)
	{
		CListBox *lb = (CListBox *)GetDlgItem(IDC_VIEWS);
		CString tmp;

		if (uv)
		{	
			for(std::vector<CUniverseView>::iterator i = uv->universeviews.begin(); i!=uv->universeviews.end(); i++)
			{
				tmp.Format("x=%d y=%d z=%d alt=%d az=%d %s",
					(int)i->x,(int)i->y,(int)i->z,(int)i->alt,(int)i->az,i->description.c_str());
				lb->InsertString(-1,tmp);
			}

			CButton *ab = (CButton *)GetDlgItem(IDC_ADD);
			CButton *db = (CButton *)GetDlgItem(IDC_DELETE);

			// disable delete until something is selected
			ab->EnableWindow(TRUE);
			db->EnableWindow(FALSE);
		}
		else
		{
			CButton *ab = (CButton *)GetDlgItem(IDC_ADD);
			CButton *db = (CButton *)GetDlgItem(IDC_DELETE);

			// disable delete until something is selected
			ab->EnableWindow(FALSE);
			db->EnableWindow(FALSE);
		}
	}
}


BEGIN_MESSAGE_MAP(CSettingsDialog, CDialog)
	//{{AFX_MSG_MAP(CSettingsDialog)
	ON_BN_CLICKED(IDC_ADD, OnAdd)
	ON_BN_CLICKED(IDC_DELETE, OnDelete)
	ON_LBN_SELCHANGE(IDC_VIEWS, OnSelchangeViews)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CSettingsDialog message handlers

void CSettingsDialog::OnCancel() 
{
	// TODO: Add extra cleanup here
	if (m_pView)
	{
		m_pView->PostMessage(WM_GOODBYE,IDCANCEL);
	}
	else
	{
		CDialog::OnCancel();
	}
}

void CSettingsDialog::OnOK() 
{
	// TODO: Add extra validation here
	if (m_pView)
	{
		UpdateData(TRUE);
		m_pView->PostMessage(WM_GOODBYE,IDOK);
	}
	else
	{
		CDialog::OnOK();
	}
}

void CSettingsDialog::OnAdd() 
{
	if (!uv)
		return;

	CEdit *ex = (CEdit *)GetDlgItem(IDC_X);
	CEdit *ey = (CEdit *)GetDlgItem(IDC_Y);
	CEdit *ez = (CEdit *)GetDlgItem(IDC_Z);
	CEdit *ealt = (CEdit *)GetDlgItem(IDC_ALT);
	CEdit *eaz = (CEdit *)GetDlgItem(IDC_AZ);
	CEdit *description = (CEdit *)GetDlgItem(IDC_DESCRIPTION);

	CString xstr,ystr,zstr,altstr,azstr,descstr;
	ex->GetWindowText(xstr);
	ey->GetWindowText(ystr);
	ez->GetWindowText(zstr);
	ealt->GetWindowText(altstr);
	eaz->GetWindowText(azstr);
	description->GetWindowText(descstr);

	descstr.Trim();

	float x,y,z,alt,az;
	x = atof((LPCTSTR)xstr);
	y = atof((LPCTSTR)ystr);
	z = atof((LPCTSTR)zstr);
	alt = atof((LPCTSTR)altstr);
	az = atof((LPCTSTR)azstr);

	uv->universeviews.push_back(CUniverseView(std::string((LPCTSTR)descstr),x,y,z,alt,az));

	CListBox *lb = (CListBox *)GetDlgItem(IDC_VIEWS);
	CString tmp;

	tmp.Format("x=%d y=%d z=%d alt=%d az=%d %s",
		(int)x,(int)y,(int)z,(int)alt,(int)az,(LPCTSTR)descstr);
	lb->InsertString(-1,tmp);

	uv->Write();
}

void CSettingsDialog::OnDelete() 
{
	CListBox *lb = (CListBox *)GetDlgItem(IDC_VIEWS);

	if (lb)
	{
		int s =	lb->GetCurSel();

		if (s != LB_ERR)
		{
			lb->DeleteString(s);

			// TODO - delete the item from the data structure too and then write it back
			// to the file

			// disable the delete control because now nothing will be selected
			CButton *db = (CButton *)GetDlgItem(IDC_DELETE);
			db->EnableWindow(FALSE);		
		}
	}
}


void CSettingsDialog::OnSelchangeViews() 
{
	// TODO: Add your control notification handler code here
	CListBox *lb = (CListBox *)GetDlgItem(IDC_VIEWS);
	
	if (lb)
	{
		int s =	lb->GetCurSel();

		if (s != LB_ERR)
		{
			CEdit *ex = (CEdit *)GetDlgItem(IDC_X);
			CEdit *ey = (CEdit *)GetDlgItem(IDC_Y);
			CEdit *ez = (CEdit *)GetDlgItem(IDC_Z);
			CEdit *ealt = (CEdit *)GetDlgItem(IDC_ALT);
			CEdit *eaz = (CEdit *)GetDlgItem(IDC_AZ);
			CEdit *description = (CEdit *)GetDlgItem(IDC_DESCRIPTION);

			// parse the current selection and put it in the boxes;
			CString d;
			int p1,p2;

			lb->GetText(s,d);

			p1 = d.Find("x=");
			p2 = d.Find(" ",p1);
			if (p1!=-1 && p2!=-1)
			{
				ex->SetWindowText(d.Mid(p1+2,p2-p1-2));
			}
			p1 = d.Find("y=");
			p2 = d.Find(" ",p1);
			if (p1!=-1 && p2!=-1)
			{
				ey->SetWindowText(d.Mid(p1+2,p2-p1-2));
			}
			p1 = d.Find("z=");
			p2 = d.Find(" ",p1);
			if (p1!=-1 && p2!=-1)
			{
				ez->SetWindowText(d.Mid(p1+2,p2-p1-2));
			}
			p1 = d.Find("alt=");
			p2 = d.Find(" ",p1);
			if (p1!=-1 && p2!=-1)
			{
				ealt->SetWindowText(d.Mid(p1+4,p2-p1-4));
			}
			p1 = d.Find("az=");
			p2 = d.Find(" ",p1);
			if (p1!=-1 && p2!=-1)
			{
				eaz->SetWindowText(d.Mid(p1+3,p2-p1-2));

				description->SetWindowText(d.Mid(p2+2));
			}


			CButton *db = (CButton *)GetDlgItem(IDC_DELETE);
			// enable delete now that something is selected
			db->EnableWindow(TRUE);		
		}
	}
}
