#include <vector>
#include <string>

#include "monitor.h"


CMonitorEntry::CMonitorEntry(std::string n)
{
	name = n;
}

void CMonitorEntry::AddTrigger(CTerminal *t)
{
	if (!t) return;

	triggers.push_back(t);
}

void CMonitorEntry::AddAdditional(CTerminal *t)
{
	if (!t) return;

	additional.push_back(t);
}

CMonitor::CMonitor(std::istream &is, CTerminals *terms)
{
	std::string logfilename;
	std::string token;

	if (!terms) return;

	is >> logfilename;

	logfile.open(logfilename.c_str());

	while(is)
	{
		is >> token;

		if (is)
		{
			entries.push_back(CMonitorEntry(token));
			bool ontriggers = true;

			is >> token;

			while(token != "end" && token != "END")
			{
				if (token == ":")
				{
					ontriggers = false;
				}
				else if (ontriggers)
				{
					entries.back().AddTrigger(terms->Lookup(token));
				}
				else
				{
					entries.back().AddAdditional(terms->Lookup(token));
				}

				is >> token;
			}
		}
	}
}
	

void CMonitor::MonitorCheck(CUniverse *u)
{
	if (!u) return;

	for(std::vector<CMonitorEntry>::iterator i = entries.begin(); i != entries.end(); i++)
	{
		bool dooutput = false;

		{for(std::vector<CTerminal *>::iterator j = i->triggers.begin(); j != i->triggers.end() && !dooutput; j++)
		{
			dooutput = u->GetSignalValue((*j)->xpos,(*j)->ypos,(*j)->zpos,(*j)->face);
		}}

		if (dooutput)
		{
			logfile << i->name << " ";

			{for(std::vector<CTerminal *>::iterator j = i->triggers.begin(); j != i->triggers.end(); j++)
			{
				logfile << 
					u->GetSignalValue((*j)->xpos,(*j)->ypos,(*j)->zpos,(*j)->face) ? "1" : "0";
			}}

			logfile << " ";

			{for(std::vector<CTerminal *>::iterator j = i->additional.begin(); j != i->additional.end(); j++)
			{
				logfile <<
					u->GetSignalValue((*j)->xpos,(*j)->ypos,(*j)->zpos,(*j)->face) ? "1" : "0";
			}}

			logfile << std::endl;
		}

	}
}

