#ifndef CMonitor_h
#define CMonitor_h

// Classes for monitoring the state of blocks.
// These will output times of all non-zero signals appearing on blocks
//
// Format of a .mon file that specifies what to monitor is:
//
// <logfile>
// "Name" <signal1> <signal2> ... <signal n> : <signal n+1> ... END
// ...
//
// This tells the monitor to output its monitorings to logfile
// signals 1 to n will be monitored and will also be trigger signals
// so that monitoring will be done whenever one of these is non-zero
// signals n+1 onwards will be output along with the trigger signals, but
// are not trigger signals.

#include <vector>
#include <string>
#include <fstream>

#include "terminal.h"
#include "../Engine/Universe.h"

class CMonitorEntry
{
public:
	CMonitorEntry(std::string n);

	void AddTrigger(CTerminal *t);
	void AddAdditional(CTerminal *t);
	std::string name;

	std::vector<CTerminal *> triggers;
	std::vector<CTerminal *> additional;
	
};

class CMonitor
{
public:
	CMonitor(std::istream &is, CTerminals *terms);
	
	void MonitorCheck(CUniverse *u);

public:
	std::ofstream logfile;
	std::vector<CMonitorEntry> entries;
};

#endif