#ifndef CMonitor_h
#define CMonitor_h

// Classes for monitoring the state of blocks.
// These will output times of all non-zero signals appearing on blocks
//
// Format of a .mon file that specifies what to monitor is:
//
// <logfile>
// "Name" <signal1> <signal2> ... <signal n> : <signal n+1> ...
// ...
//
// This tells the monitor to output its monitorings to logfile
// signals 1 to n will be monitored and will also be trigger signals
// so that monitoring will be done whenever one of these is non-zero
// signals n+1 onwards will be output along with the trigger signals, but
// are not trigger signals.

#include <vector>
#include <string>

#include "terminal.h"


class CMonitorEntry
{
public:
	CMonitorEntry(std::string n);

	void AddTrigger(CBlock *t);
	void AddAdditional(CBlock *t);
	std::string name;

	std::vector<CBlock *> triggers;
	std::vector<CBlock *> additional;
	
}

class CMonitor
{
public:
	CMonitor(std::istream &is, CTerminals *terms);
	
	void AddEntry(CMonitorEntry &e);
	void MonitorCheck(void);

public:
	std::ostream logfile;
	std::vector<CMonitorEntry> entries;
}

#endif